#include "trace.h"
#include "classify.h"
#include "role_restr.h"


bool Concept::operator==(const Concept& C) const
/*
    Two concepts are equal iff they have 
     
     . the same parents
     . the same role restrictions
     . the same tests
 */
{
    if (this == &C)
        return TRUE;

    /*
        Attention... compare name only if both are classified.
                                          ^^^^
        (both: not just *any one*, because we may use this routine to
         see if an unclassified concept is equal to a classified one.
         If they are, we delete the unclassified concept and return the
         classified one -- not done in this routine)

        A subtle bug before.
     */
    if (classified() && C.classified() && name() == C.name())
        return TRUE;

    /*
        if any one is primitive, then always not equal (except both
        have the same name, which is checked above).
     */
    if (primitive() || C.primitive())
        return FALSE;

    /*
        Here, both are real concepts.  In any kind of concept comparisons,
        we do not use the parent information because the parent info will
        be changed during classification and classification calls this
        procedure to determine how to classify.  It would be recursive to
        make use of parent info here.  Fortunately, it is sufficient to 
        use role restrictions and other restrictions to do the comparisons.

        Notice that we do need to make sure RRs are inherited.  This is
        done in the normalization function.  Therefore we assume the 
        concepts to be compared have been normalized.
     */

    /*
        As a second thought, the above is wrong.  We do need to compare
        the parents.  Consider this:

                              Thing
                          /  /    \  \     <--------- all primitive
                        P1  P2    Q1  Q2
                          \/        \/
                          C1        C2

        C1 will be equal to C2 if we don't compare parents!
     */

    if (_parents != C._parents)
        return FALSE;

    /*
        Done checking parents.  Checking parents alone is not enough 
        because we may have such concepts:

             Thing
               |
               | primitive            In this diagram, C1 and C2 have
               |                      the same parent C, but they are
               C                      not the same concept... so need
              / \                     to check restrictions.
             /   C2 has red color
            /
         C1 has blue color
     */

    const Set<RoleRestriction*>& rr_list1 = role_restrictions();
    const Set<RoleRestriction*>& rr_list2 = C.role_restrictions();

    if (rr_list1.size() != rr_list2.size())
        return FALSE;

    /* 
        For each RoleRestriction, find the corresponding RoleRestriction in
        C and compare them (rr1 and rr2) to see if they are equal.
     */

    extern RoleRestriction* name_in(
   	       const Set<RoleRestriction*>& source,
	       const RoleRestriction* rr
	   );

    for (Set<RoleRestriction*>::const_iterator i = rr_list1.begin();
        i != rr_list1.end(); i++) {

        RoleRestriction* rr1 = *i;     // any RR in 'this'

        /*
            Here, compare them.
         */
        RoleRestriction* rr2 = name_in(rr_list2, rr1); // corresponding RR in C
	if (!rr2)
	    return FALSE;

        if (*rr1 != *rr2)
            return FALSE;
    }

    /* 
	Now make sure the one-ofs are the same.
     */

    const Set<Individual*>& oo_list1 = one_ofs();
    const Set<Individual*>& oo_list2 = C.one_ofs();
    if (!set_equal(oo_list1, oo_list2, individual_comp))
	return FALSE;

    /*
	Now make sure each tests in this is also in C.
	I think no need to compare all inherited tests since we know
	the parents are same if we get to this point.
     */

    if (_tests != C._tests)
	return FALSE;

    return TRUE;
}


bool Concept::operator<=(const Concept& C) const
/*
    At first, I defined this operator as op< (properly subsume).  Now 
    I think it is too difficult to write.  "subsume" (ie. op<=) will
    be a lot simpler.  In the following discussion, C' is 'this' and
    C is the incoming concept.  We wish to determine whether C' <= C.

    There are four cases, divided into two similar divisions:

    (1) C' is primitive, C is not primitive

        (define-prim C' (fills color blue))                 Thing
                                                            /  |  \ prim
        (define-con C1 (fills color blue))            prim /   |   P
                                                          /    |    \
        (define-prim P thing)                           C' --> C1 <--C2
        (define-con C2 (and P (fills color blue)))      :      :     :
                                                       blue   blue  blue
        In the diagram, we can see that C' <= C1
        but C1 is not <= C2, because C2 is a P and
        there is no proof that C' is also a P.  (In fact C2 <= C1 too,
        although this fact is not useful in our discussion see (2)).

    (2) C' is not primitive, C is not primitive

	See the above diagram, C1 is not <= C2 but C2 <= C1.  We have
	the same rule for (1) and (2), namely,
	 _________________________________________________________ 
	|                                                         |
	|  If C is not primitive, C' <= C if and only if          |
	|  all RR of C >= all RR of C' AND all parents of C >= C' |
	|_________________________________________________________|

	What if all the RR are inherited?  Can we change the above rule to

	   all RR of C >= all RR of C'  (?)

        The answer is no due to primitive ancestor concepts do not 
	necessarily have RRs.  Thus, not inherited.

	But it is true that only told restrictions need to be satisfied
	if we use the above (recursive) rule.

    (3) C' is primitive, C is primitive
	
	The rule above does not work when C is primitive, e.g.,

	   Thing          Following the above rule, C' <= C which is 
	    / \           clearly incorrect.
      prim /   \ prim
	  /     \
         C'      C

    (4) C' is not primitive, C is primitive

	   Thing              
	    /|\               Since primitive concepts specify only
           / | \  <-- all     necessary but not sufficient conditions,
	  /  |  \     prim    it is not 'sufficient' for any concept to
        C1   C   C2           get classified under a primitive concept
	  \     /             unless it is explicitly stated so.
	   \   /
	    \ /               Therefore, we evolve the following rule.
	     C'
	 ___________________________________________ 
	|                                           |
	| If C is primitive, C' <= C if and only if |
	| ancestors of C' contains C                |
	|___________________________________________|

    (5) There is an exception.  When C is ClassicThing, the above rule
	(4) doesn't work ... e.g., 

	(define-concept C' (all options "-r"))

	The ancestors of C' does not contain C.  So, we add a special
	rule that says if C is ClassicThing and HostThing is not an
	ancestor of C', then return TRUE.
        
 */
{
#if defined(Debug) 
cout << "Concept::<=" << endl;
#endif
    Concept* pC = (Concept*)&C;

    if (trace.on(Trace::subsumption))
	cout << name() << " <= " << C.name() << "?" << endl;

    if (this == &C) {
	if (trace.on(Trace::subsumption))
	    cout << "    yes, same object." << endl;
        return TRUE;
    }

    /*
	To achieve efficiency, if both concepts are classified, simply
	check the taxonomy.  Therefore, when re-classifying a concept,
	_classified flag needs to be set to false... done in 
	Concept::classify().
     */
    if (classified() && C.classified()) {
	if (trace.on(Trace::subsumption))
	    cout << "    both are classified... " << endl;
	Set<Concept*> anc = ancestors();
	int tmp = anc.find(pC) != anc.end();

	if (trace.on(Trace::subsumption))
	    if (tmp)
		cout << "    yes, rhs is an ancestor of lhs.\n";
	    else
		cout << "    no, rhs is not an ancestor of lhs.\n";
	return tmp;
    }

    /*
	If C is a primitive concept... cases (3), (4) and (5)
     */

    if (C.primitive()) {
	if (trace.on(Trace::subsumption))
	    cout << "    rhs is a primitve concept...\n";
	const Set<Concept*>& anc = ancestors();
	if (&C == ClassicThing && this != Thing && object_base.initialized()) {
	    /*
		The exception case, see (5).  The reason we need to see
		if the object base has been initialized is because we are
		using 'HostThing' below.  We must ensure it has been init.
	     */
	    int tmp = anc.find(HostThing) == anc.end();
	    if (trace.on(Trace::subsumption)) 
		if (tmp) 
		    cout << "    yes, rhs is a ClassicThing " << 
			"& lhs is not a HostThing.\n";
		else
		    cout << "    no, rhs is a ClassicThing " <<
			"& lhs is a HostThing.\n";
	    return tmp;
	} else {
	    int tmp = anc.find(pC) != anc.end();
	    if (trace.on(Trace::subsumption))
		if (tmp)
		    cout << "    yes, rhs is an ancestor of lhs.\n";
		else
		    cout << "    no, rhs is not an ancestor of lhs.\n";
	    return tmp;
	}
    }
     
    /*
	Here, C is NOT a primitive concept... cases (1) and (2)
     */

    if (trace.on(Trace::subsumption))
	cout << "    checking each RR of rhs...\n";

    extern RoleRestriction* name_in(
   	       const Set<RoleRestriction*>& source,
	       const RoleRestriction* rr
	   );

    trace.suspend();
    Set<RoleRestriction*> rr_list1 = role_restrictions();
    Set<RoleRestriction*> rr_list2 = C.role_restrictions();
    trace.resume();

    for (Set<RoleRestriction*>::iterator i = rr_list2.begin(); 
	i != rr_list2.end(); i++) {

	RoleRestriction* rr1 = *i;  // any RR in C

        if (trace.on(Trace::subsumption))
	    cout << "        role " << rr1->name() << endl;

	/*
	    Here, find the RR in 'this' that has the same 'role'.
	 */
	// corresponding RR in 'this'
	RoleRestriction* rr2 = name_in(rr_list1, rr1);   

	if (!rr2) {
	    if (trace.on(Trace::subsumption))
		cout << "        not found in lhs ... false.\n";
	    return FALSE;
	}

	if (!(*rr1 >= *rr2)) {
	    if (trace.on(Trace::subsumption))
		cout << "        does not >= same role in lhs ... false.\n";
	    return FALSE;
	}

	if (trace.on(Trace::subsumption))
	    cout << "        ok" << endl;
    }

    /*
	All parents of C must subsume 'this'.
     */

    if (trace.on(Trace::subsumption))
	cout << "    checking all parents of rhs.\n";

    for (Set<Concept*>::iterator p = C._parents.begin();
	p != C._parents.end(); p++)
	if (!(*this <= **p)) {
	    if (trace.on(Trace::subsumption)) {
		cout << "    " << (*p)->name() << " does not >= lhs.\n";
		cout << name() << " <= " << C.name() << "? no\n";
	    }
	    return FALSE;
        }

    /*
	Check the one-ofs.  this's one-ofs should be subset of C.
     */

    if (trace.on(Trace::subsumption))
	cout << "    checking all one-ofs.\n";

    const Set<Individual*>& oo_list1 = one_ofs();
    const Set<Individual*>& oo_list2 = C.one_ofs();
    if (oo_list1.empty() && !oo_list2.empty()) {
	if (trace.on(Trace::subsumption))
	    cout << "    No.  lhs is empty, rhs is not." << endl;
	return FALSE;
    }
    if (!oo_list2.empty() && !set_subset(oo_list1, oo_list2, individual_comp)) {
	if (trace.on(Trace::subsumption))
	    cout << "    No.  lhs is not subset of rhs." << endl;
	return FALSE;
    }

    /*
	Also, if this has more tests than C, then this is more restricted.
     */

    if (trace.on(Trace::subsumption))
	cout << "    checking all tests.\n";
     
    if (!(C.tests() <= tests())) {
	if (trace.on(Trace::subsumption))
	    cout << "    No.  rhs is not subset of lhs." << endl;
	return FALSE;
    }

    if (trace.on(Trace::subsumption))
	cout << name() << " <= " << C.name() << "? yes\n";

#if defined(Debug) 
cout << "Concept::<= returns" << endl;
#endif

    return TRUE;
}


bool operator<(const Concept& C, const Concept& D)
{
    return C.operator<=(D) && !C.operator==(D);
}


bool operator>(const Concept& C, const Concept& D)
{
    return D < C;
}


bool operator>=(const Concept& C, const Concept& D)
{
    return D <= C;
}


bool Concept::classify_parents(Concept* C)
{
    Set<Concept*> tmp;
    bool ok = classify_parents_aux(C, tmp);

    /*
	After any new parents are added to C, we need to normalize again,
	consider:

            D = (and B C)
	    E = (and B C (fills ...))

	During classification of E, we add D to E:

	    E = (and B C D (fills ...))

	We need to normalize E to

	    E = (and D (fills ...))
     */

    normalize_parents();

    return ok;
}


bool Concept::classify_parents_aux(Concept* C, Set<Concept*>& duplicates)
/*
    aka Most Specific Subsumers

    Use subsumption relationship to classify this under 'C' or its 
    descendants.  This is an aux function for Concept::classify().

    'C' represents the root of a concept taxonomy, generally it
    should be 'Thing' except when called recursively by classify_parents
    itself.

    This is a function for internal use.  The user should use 
    Concept::classify().
 */
{
    /*
        Find all the parents of this
     */

    /*
	Duplicate check, see comments about duplicates below.
     */

    if (duplicates.find(C) != duplicates.end())
	return TRUE;
    else 
	duplicates.insert(C);

    if (trace.on(Trace::classification))
	cout << "Finding sup " << *this << " <= " << *C << "? ";

    if (*this < *C) {
	if (trace.on(Trace::classification))
	    cout << "yes.\n";
	
        /*
            Here *C is a potential parent because 'this' is
            subsumed under it.
         */
        bool need_to_add = TRUE;

        /* 
            See if any of the children of the potential parent subsumes
            'this'.  If so, the potential parent is at least a *grand-
            parent* of 'this' and does not need to  be added to the 
            *parent* list.
         */

	/*
	    A very interesting bug here.  Notice the following taxonomy:

		B     C
		 \   /
		   D  = (and B C)
		   |
		   E  = (and B D)

	    E < thing?  yes
	 (*)Try B and C
	    E < B?  yes
	    Try D
	    E < D?  yes ... no more children add E to D
	    retract to (*)
	    E < C?  yes
	    Try D          .................................................
	    E < D?  yes ... one more child this time                        :
	    Try E                                                           :
	    E < E?  yes ... no more children add E to D ... duplicate!!!    :
                                                                            :
            Should do a duplicate check here ...............................:

	 */

        for (Set<Concept*>::iterator i = C->_children.begin();
            i != C->_children.end(); i++)
            if (classify_parents_aux(*i, duplicates))
                need_to_add = FALSE;

        if (need_to_add) {
	    if (trace.on(Trace::classification))
		cout << "Adding " << *this << " to " << *C << ".\n";
            add_parent(C);
	/*
	    After some parents are added, we should normalize them again.
	    e.g., 
		  D = (and B C)
		  E = (and B C (fills name ...))

	    D will be added to E in classify_parents_aux:

		  E = (and B C D (fills name ...))

	    Since D is found to be a child of (and B C), E should be
	    normalized to

		  E = (and D (fills name ...))

	    The same is not needed in ClassicIndividual::classify_parents_aux
	    because individuals have separate concept descriptors.  This 
	    means the parents will be added to the _parents set one by one.
	 */
            C->add_child(this);
        }
        return TRUE;
    } else
        if (*C == *this) {
	C->print();
	    if (trace.on(Trace::classification))
		cout << "equal.\n";

            /*
                this is equivalent to C.  This makes this obsolete and
                must not be classified under Thing.  Only an alias is
                created.
             */
            C->add_name(name());
            OB.insert(name(), C->id());
            return FALSE;
        } else {
	    if (trace.on(Trace::classification))
		cout << "no.\n";
            return FALSE;
	}
}


bool Concept::classify_children()
/*
    aka Most General Subsumees
 */
{
    for (Set<Concept*>::iterator i = _parents.begin();
        i != _parents.end(); i++) {
	 
	/*
	    It is necessary to use descendants() in the following.  
	    Previously, children() was used and was buggy.  This is
	    due to multiple inheritance.  Consider this hierarchy:

		Thing          Suppose B is added to the hierarchy,
		 /  \ prim     and B subsumes D.  B.classify_children
		/    C         is called.  Searching only under B's
	       /      \        siblings will determine B does not 
	      B	<----- D       subsume C, but will overlook D < B

	 */
	Set<Concept*> des = (*i)->descendants();

        for (Set<Concept*>::iterator j = des.begin(); j != des.end(); j++) {
            /*
                if this subsumes any descendants, j,  of each parent of this,
                then this is added as a parent of j.  This also means the 
                original parent is deleted from j's parent list.
                **j is used because *j is Concept* and de-reference it is
                a Concept.
             */

	    /*
		Since "this" is obviously a child of its parent(s), 
		it should be skipped from the comparison.
	     */

	    if (*j == this)
		continue;

	    if (trace.on(Trace::classification))
		cout << "Finding sub " << *this << " > " << **j << "? ";

            if (**j < *this) {
		if (trace.on(Trace::classification)) {
		    cout << "yes.\n";
		    cout << "Adding " << **j << " to " << *this << ".\n";
		}
                add_child(*j);
                (*j)->remove_parent(*i);
		(*j)->add_parent(this);
            } else
		if (trace.on(Trace::classification))
		    cout << "no.\n";
	}

            /*
                We do not need to worry about *this == *(*j) because it 
                will be caught by classify_parent.
             */
    }

    return TRUE;
}


void Concept::classify_instances()
/*
    This funcion is called after a concept, C, has been classified 
    into the taxonomy.  Suppose C is classified directly under a 
    parent concept, B, we need to examine each direct instance of
    B to see if they fits under C.  If so, move them.
 */
{
    for (Set<Concept*>::iterator p = _parents.begin(); p != _parents.end(); 
	p++) {

        Set<Individual*> tmp = (*p)->direct_instances();
	for (Set<Individual*>::iterator i = tmp.begin(); i != tmp.end(); i++) {
	    if (trace.on(Trace::classification))
		cout << "Finding ind " << *this << " > " << **i << "? ";
	    if (**i < *this) {
	    /*
		i needs to be moved under "this".  This means 

		. delete p from i's parent
		. add this to i's parent
		. delete i from p's direct instances
		. add i to this' direct instances
	     */
		if (trace.on(Trace::classification))
		    cout << "yes.\n";
		(*i)->remove_parent(*p);
		(*i)->add_parent(this);
		(*p)->remove_instance(*i);
		add_instance(*i);
	    } else
		if (trace.on(Trace::classification))
		    cout << "no.\n";
	}
    }
}


void Concept::classify()
/*
     Classify a named concept.  Assume 'this' has been normalized.
     The procedure is to find the immediate parents and immediate
     children of 'this'.  As well as instances.

     ### need to take care of empty concepts.
     ### e.g., disjoints ... well, probably done in normalization rather
     ### than in classify

     ### need to take care of primitive concepts... maybe in < and ==
 */
{
#if defined(Debug)
cout << "classify()" << endl;
#endif
    assert(name() != "");
    assert(!null());

    /*
	Assume it is not classified.  This information is necessary in
	order to re-classify an already classified concept;  Otherwise,
	the subsumption routines will simply search for C's position in
	the taxonomy, instead of re-classifying it.
     */
    _classified = FALSE;   

    bool ok = classify_parents(Thing);

    /*
        There is a slim chance that classify_parents returns FALSE.
        This is when there is a concept in the taxonomy equivalent to
        'this'.  In this case, classify_parents cannot do anything
        but return.
     */
    if (!ok) 
        return;

    classify_children();

    classify_instances();

    _classified = ok;
#if defined(Debug)
cout << "classify() returns" << endl;
#endif
}


bool operator<(const Individual& I, const Concept& C)
/*
    The operator defined for instance-of (ako) relationship is

	operator<(const Individual&, const Concept&)

    The fact that the Individual can be Host or Classic makes
    this slightly more complicated.  It means there should be
    two routines for each kind.  Both of them are simple in 
    our adaptation (for one, all roles are assumed closed).
 */
{
    /*
	I am down casting here.  OK because we check if it is host() first.
     */
    if (I.host())
	return isa((const HostIndividual&)I, C);
    else
	return isa((const ClassicIndividual&)I, C);
}


bool isa(const HostIndividual& I, const Concept& C)
/*
    "I" must have one and only one parent because it is a host
    individual.  So, simply return I.parent <= C.  And take 
    care of one_of constraints.
 */
{
#if 0
    return (*I.parent()) <= C;
#else
    /*
	This routine turns out to be very very problematic.
	Suppose we want to find out if

	    "a" <= (and string (one-of "a" "b"))

	The parent of "a" is only defined as string not 
	(and string (one-of "a"))... this causes this
	procedure to return FALSE if we use a simplistic
	parent <= C comparison as hinted above.

	The correct comparison should be 

	   (and string (one-of "a")) <= C

	Therefore, we should compare the ancestors of both
	sides.  Then, check one-ofs.
     */


    /*
	I.parent() must be <= all of C's parents.
     */

    const Set<Concept*>& parents = C.parents();
    const Concept* host = I.parent();
    for (Set<Concept*>::const_iterator i = parents.begin(); 
	i != parents.end(); i++) 
	if (!(*host <= **i))
	    return FALSE;

    /* 
	Jesus Christ!!!  Every day there is a fucking G++ problem.
	Every day!!!  Why can't the following compile!?  (error in
	set_in...)

	const Individual* tmp = &I;
    */

    Individual* tmp = (Individual*)&I;
    const Set<Individual*>& oo = C.one_ofs();
    if (!oo.empty() && !set_in(tmp, oo, individual_comp))
	return FALSE;

    return TRUE;
#endif
}


bool isa(const ClassicIndividual& I, const Concept& C)
/*
    "I" is a Classic individual.  We do not allow late assertions
    of role fillers, therefore, there is no close-role command.
    As a result, I < C iff the I.descriptor <= C, except that
    when considering descriptors, the <= operator assumes all
    roles to be closed.
 */
{
    /* 
	I discard const here.  This is done because I is going to be
	updated with test results... but done only once in the very
	beginning.   Should be declared as mutable somewhere.  Don't
	think g++ supports this though.
     */

    return I._descriptor->instance_of(C, (ClassicIndividual*)&I);
}


bool Concept::instance_of(const Concept& C, Individual* I) const
/*
    This is a peculiar function.  Normally, a concept is not considered
    as instance of another concept.  This is used as an auxiliary function
    by operator<(const Individual&, const Concept&).
 
    This function follows very closely to 

	bool Concept::operator<=(const Concept& C) const

    except that 'this' is supposed to be the descriptor of an individual.
    In this case, the RRs for 'this' are considered closed.

    See explanations in the above operator.

    Also, pay attention to the use of "ancestors" below.  Ancestors of
    "this" (the descriptor concepts of I) may be different from anc of
    I because of asserted parents (such as from consequents of rules).
    To be absolutely correct, we need to take union of both because when
    a new individual is being defined, there is nothing in I->_parents.
    After define_individual, this->parents will be added to I->_parents.
    But this routine may get called during define_individual, therefore
    we take unions.
 */
{
#if defined(Debug)
cout << "instance_of()" << endl;
#endif
    if (trace.on(Trace::subsumption))
	cout << name() << " <- " << C.name() << "?" << endl;

    if (this == &C) {
	if (trace.on(Trace::subsumption))
	    cout << "    yes, same object." << endl;
        return TRUE;
    }

#if 0
    /*
	This is not done for classification of individuals because 
	for example:

	  (def-prim P (all options string))
	  (def-con C (and P (all options (one-of "a" "b" "c"))))
  
	  (def-ind i P)
	  (ind-close-role i options)

	This should classify i under C.  But if we use the classified
	results as short cut, then P is compared with C and the 
	classication will stop because P is parent of C!
     */

    /*
	To achieve efficiency, if both concepts are classified, simply
	check the taxonomy.  Therefore, when re-classifying a concept,
	_classified flag needs to be set to false... done in 
	Concept::classify().
     */

    if (classified() && C.classified()) {
	if (trace.on(Trace::subsumption))
	    cout << "    both are classified... " << endl;
	Set<Concept*> anc = ancestors() + I->ancestors();
	/*
	    set_union(anc1.begin(), anc1.end(), anc2.begin(), anc2.end(),
	    back_inserter(anc));
	*/
	int tmp = anc.find(&C) != anc.end();

        if (trace.on(Trace::subsumption))
            if (tmp)
                cout << "    yes, rhs is an ancestor of lhs.\n";
            else
                cout << "    no, rhs is not an ancestor of lhs.\n";
        return tmp;
    }
#endif

    /*
	If C is a primitive concept... cases (3), (4) and (5)
     */

    if (C.primitive()) {
        if (trace.on(Trace::subsumption))
            cout << "    rhs is a primitve concept...\n";
	Set<Concept*> anc = ancestors() + I->ancestors();
	/*
	    set_union(anc1.begin(), anc1.end(), anc2.begin(), anc2.end(),
	    back_inserter(anc));
	*/
 
	if (&C == ClassicThing && this != Thing && object_base.initialized()) {
	    /*
		The exception case, see (5).  The reason we need to see
		if the object base has been initialized is because we are
		using 'HostThing' below.  We must ensure it has been init.
	     */
	    int tmp = anc.find(HostThing) == anc.end();
            if (trace.on(Trace::subsumption))
                if (tmp)
                    cout << "    yes, rhs is a ClassicThing " <<
                        "& lhs is not a HostThing.\n";
                else
                    cout << "    no, rhs is a ClassicThing " <<
                        "& lhs is a HostThing.\n";
            return tmp;
	} else {
	    int tmp = find(anc.begin(), anc.end(), &C) != anc.end();
            if (trace.on(Trace::subsumption))
                if (tmp)
                    cout << "    yes, rhs is an ancestor of lhs.\n";
                else
                    cout << "    no, rhs is not an ancestor of lhs.\n";
            return tmp;
        }
    }
     
    /*
	Here, C is NOT a primitive concept... cases (1) and (2)
     */
      
    if (trace.on(Trace::subsumption))
        cout << "    checking each RR of rhs...\n";

    extern RoleRestriction* name_in(
               const Set<RoleRestriction*>& source,
               const RoleRestriction* rr
           );

    trace.suspend();
    /*
	This is a subtle bug.  rr_list1 should use the role_restrictions
	of I instead of the RR of the descriptor concept.  

	Set<RoleRestriction*> rr_list1 = role_restrictions();
     */
    Set<RoleRestriction*> rr_list1 = 
	((ClassicIndividual*)I)->role_restrictions();
    Set<RoleRestriction*> rr_list2 = C.role_restrictions();
    trace.resume();

    for (Set<RoleRestriction*>::iterator i = rr_list2.begin();
        i != rr_list2.end(); i++) {

        RoleRestriction* rr1 = *i;  // any RR in C

        if (trace.on(Trace::subsumption))
            cout << "        role " << rr1->name() << endl;

	/*
	    Here, find the RR in 'this' that has the same 'role'.
	 */

	RoleRestriction* rr2 = name_in(rr_list1, rr1);  
        if (!rr2) {
            if (trace.on(Trace::subsumption))
                cout << "        not found in lhs ... false.\n";
            return FALSE;
        }

	/*
	    **** Here is the first difference between this function and 
		 Concept::operator<=.  Here we must close the role, rr2, 
		 before we do the comparison.

	    **** Well, no we should not close it... instead, an role
		 should be automatically closed if it is filled with max
		 fillers.
        
	    if (trace.on(Trace::subsumption))
		cout << "        close role in lhs.\n";
	 */

	if (!(*rr1 >= *rr2)) {
            if (trace.on(Trace::subsumption))
                cout << "        does not >= same role in lhs ... false.\n";
	    return FALSE;
        }

        if (trace.on(Trace::subsumption))
            cout << "        ok" << endl;
    }

    /*
	All parents of C must subsume 'this'.
     */

    if (trace.on(Trace::subsumption))
        cout << "    checking all parents of rhs.\n";

    for (Set<Concept*>::iterator p = C._parents.begin();
	p != C._parents.end(); p++)
	if (!this->instance_of(**p, I)) {	// *** another difference.
            if (trace.on(Trace::subsumption)) {
                cout << "    " << (*p)->name() << " does not -> lhs.\n";
                cout << name() << " <- " << C.name() << "? no\n";
            }
	    return FALSE;
        }

    /*
	Check if C has any one-ofs.
     */

    const Set<Individual*>& oo_list1 = C.one_ofs();
    const Set<Individual*>& oo_list2 = one_ofs();
    Set<Individual*> intersection;
    bool contra = FALSE;

    if (oo_list1.empty())
	intersection = oo_list2;
    else
	if (oo_list2.empty())
	    intersection = oo_list1;
	else {
	    set_intersection(oo_list1.begin(), oo_list1.end(),
		oo_list2.begin(), oo_list2.end(),
		inserter(intersection), individual_comp);
	    if (intersection.empty())
		contra = TRUE;
	}

    if (contra || 
       (intersection.size() != 0 && !set_in(I, intersection, individual_comp)))
	return FALSE;

    /*
	Check all classic tests.  Another major difference.
     */

    const Set<Test*>& tests = C.tests();
    for (Set<Test*>::const_iterator t = tests.begin(); t != tests.end(); t++)
	if ((**t)(I) != YES)
	    return FALSE;

    if (trace.on(Trace::subsumption))
        cout << name() << " <- " << C.name() << "? yes\n";

#if defined(Debug)
cout << "instance_of() returns" << endl;
#endif

    return TRUE;
}


void ClassicIndividual::classify()
{
    _classified = FALSE;
    _dirty = TRUE;

#if 1
    /*
	Dirty flag is used to achieve better efficiency.  Normally, when
	a new filler is added through a computed-filler rule, the ind is
	immediately re-classified.  Now it is delayed to solve redundancy
	problem when more than one CF rule are fired during classification.
     */
    while (_dirty) {
	_dirty = FALSE;
	classify_parents(Thing);
    }
#else
    classify_parents(Thing);
#endif

    _classified = TRUE;
}


bool ClassicIndividual::classify_parents(Concept* C)
{
    Set<Concept*> tmp;
    bool ok = classify_parents_aux(C, tmp);

    normalize_parents();

    return ok;
}


void indent(int n)
{
    while (--n)
	cout << "  ";
}


bool ClassicIndividual::classify_parents_aux(Concept* C, 
     Set<Concept*>& duplicates)
/*
    This is an aux function for ClassicIndividual::classify().

    'taxonomy' represents the root of a concept taxonomy, generally it
    should be 'Thing' except when called recursively by classify_parents
    itself.

    This is a function for internal use.  The user should use 
    Concept::classify().
 */
{
    static int level = 0;

    level++;

    /*
        Find all the parents of this
     */

    if (duplicates.find(C) != duplicates.end()) {
    /*
	This duplicate checking seems to get us into trouble. 
	Consider CP-DIRECTORY1, it is a descendant of both RECURSIVE
	and CP-SYNTAX.  When classifying an ind under RECURSIVE->
	CP-DIR1, it is found that the ind is not a CP-DIR1; however, 
	there may be more information when it is reclassified under 
	CP-SYNTAX->CP-DIR1 (consider any filler rules fired by parents
	of CP-SYNTAX, for instance)... so we must not skip it.

	I forgot whether this duplicate checking is for efficiency or
	necessity.  Probably the former.  Will try commenting it out.
     */
     /*
	The above is probably not an issue.  The minor bug is this:
        we should not blindly return TRUE here.  The evaluated concept
	in duplicates could be a non-concept... in that case, we should
	return FALSE.  In fact, we shouldn't return at all.  The 
	solution seems to be clear: insert the concept only if it is 
	a positive concept.
      */
	level--;
	return TRUE;
    } 
    /*
	Delay... see comments above.

    else
	duplicates.insert(C);
    */

    if (trace.on(Trace::classification)) {
	indent(level);
	cout << "Finding par " << *this << " < " << *C << "? ";
    }

    if (*this < *C) {

	if (trace.on(Trace::classification))
	    cout << "yes.\n";

        /*
            Here *C is a potential parent because 'this' is
            subsumed under it.
         */

        bool need_to_add = TRUE;
	duplicates.insert(C);

        /* 
            See if any of the children of the potential parent subsumes
            'this'.  If so, the potential parent is at least a *grand-
            parent* of 'this' and does not need to  be added to the 
            *parent* list.
         */

        for (Set<Concept*>::iterator i = C->_children.begin();
            i != C->_children.end(); i++) {
            if (classify_parents_aux(*i, duplicates))
                need_to_add = FALSE;
	}

	/*
	    It is possible that during a reclassification of I, C has
	    I as one _direct_instance but it should be erased.
	 */

        if (!need_to_add)
	    C->remove_instance(this);

        if (need_to_add && _parents.find(C) == _parents.end()) {
	    if (trace.on(Trace::classification)) {
		indent(level);
		cout << "Adding " << *this << " to " << *C << ".\n";
	    }
            add_parent(C);
            C->add_instance(this);
        }

        level--;
        return TRUE;
    } else {
	if (trace.on(Trace::classification))
	    cout << "no.\n";
	level--;
	return FALSE;
    }
}


bool ClassicIndividual::classify(Concept* C)
/*
    This is called by rules to try to classify an individual under 
    an incidental concept.  It should also get called by the 
    propagation mechanism, which has not been implemented yet.

    Propagation mechanism is used to deduce the property of a filler
    which its value is discovered in a value restriction.  e.g.,

        (define-concept RM (all arguments WRITABLE))
	(define-individual cmd001 (and RM (fills arguments file001)))
                                              _______________/
	Here, classify file001 under WRITABLE 
 */
{
    /*
	Notice that _descriptor, D, of an individual is the original
	defining concept.  To classify I under another concept C
	other than its defining D, we have to modify the _parent set
	directly.... except when D is already there.

	So, the algorithm is to add C to the _parent set and try
	to find contradiction.

	New discovery, after adding C to the _parent set, and there is
	no contradiction, we must not stop there.  The individual need
	be pushed down further into children of C in case there are any.
     */

    /*
	If this individual is already classified under C, simply
	return TRUE.
     */

    if (trace.on(Trace::classification)) 
	cout << "Classifying " << *this << " under " << *C << "... ";

    if (*this < *C) {
	if (trace.on(Trace::classification))
	    cout << "already an ind.\n";
	return TRUE; 
    } else
	if (trace.on(Trace::classification))
	    cout << endl;

    /*
	Add C to the parent set, normalize and see if contradiction
	results.
     */

    add_parent(C);

    normalize();

    if (contradict()) {
	remove_parent(C);
	cout << "Conflict\n";
	return FALSE;
    } else {
	/* 
	    I don't remmeber why I put add_instance here instead of calling
	    classify().  Classify() should be called otherwise for a rule

		FILENAME --> EXISTENCE
 
            the individual will only be classified under EXISTENCE but not
	    pushed down to EXISTENCE's children, which should be done.

	    C->add_instance(this);

	    The problem of using classify_parents is, this individual is 
	    not really considered as classified under C.  The better way
	    is to write another aux function, called assert_parent.

	    ### seems ok for now.
	 */
	classify_parents(C);

	if (trace.on(Trace::classification))
	    cout << "Adding " << *this << " to " << *C << ".\n";

	/*
	    Here we have a dilemma.  Normally, the classify_parents
	    procedure adds I to a C and fires the rules.  But in the
	    above, we add_parent(C) to initiate the classification
	    process.  If C turns out to be the final parent of I, we
	    are responsible for firing the rules associated w/ C.  If
	    C is not the final parent, but some of its children is,
	    then we don't need to worry, because classify_parents 
	    will do rules firing.
	 */

	if (_parents.find(C) != _parents.end())
	    C->add_instance(this);

	return TRUE;
    }
}
