#include "concept.h"
#include "individual.h"
#include "procedure.h"
#include "rule.h"
#include "role_restr.h"

ostream& operator<<(ostream& cout, const Concept& C)
{
    if (C.classified())
	cout << C.name();
    else
	C.pretty_print(cout);

    return cout;
}


int concept_print_aux(const Concept* C)
{
    cout << "    " << *C << endl;
    return 0;
}


int rr_print_aux(const RoleRestriction* rr)
{
    cout << "    " << *rr << endl;
    return 0;
}


void Concept::print(ostream& cout) const
{
    Object::print(cout);

    if (null()) 
	cout << "Notice: null concept!" << endl;

    if (!_names.empty())
	cout << "Names:";
    for (List<String>::const_iterator i = _names.begin(); 
	i != _names.end(); i++)
	cout << " " << *i;
    if (unclassified()) 
	cout << " (unclassified)";
    cout << endl;

    if (!_parents.empty()) {
	cout << "Parents:" << endl;
	for_each(_parents.begin(), _parents.end(), concept_print_aux);

	const Set<Concept*>& tmp = ancestors();
	if (!tmp.empty()) {
	    cout << "Ancestors:" << endl;
	    for_each(tmp.begin(), tmp.end(), concept_print_aux);
	}
    }

    if (!_children.empty()) {
	cout << "Children:" << endl;
	for_each(_children.begin(), _children.end(), concept_print_aux);
     
	const Set<Concept*>& tmp = descendants();
	if (!tmp.empty()) {
	    cout << "Descendants:" << endl;
	    for_each(tmp.begin(), tmp.end(), concept_print_aux);
	}
    }

    if (!_direct_instances.empty()) {
	cout << "Direct Instances:" << endl;
	for (Set<Individual*>::iterator i = _direct_instances.begin();
	    i != _direct_instances.end(); i++)
	    cout << "    " << **i << endl;
    }

    const Set<Individual*>& tmp_instances = instances();
    if (!tmp_instances.empty()) {
	cout << "Instances:" << endl;
	for (Set<Individual*>::iterator i = tmp_instances.begin();
	    i != tmp_instances.end(); i++)
	    cout << "    " << **i << endl;
    }

    /*
	The following is duplicated in ClassicIndividual::print().
	If any changes are done, it is highly probable that the
	same changes need to be done there.
     */

    {
	const Set<RoleRestriction*>& rr_list = told_role_restrictions();
	if (!rr_list.empty()) {
	    cout << "Told Role Restrictions:" << endl;
	    for_each(rr_list.begin(), rr_list.end(), rr_print_aux);
	}
    }

    {
	const Set<RoleRestriction*>& rr_list = inherited_role_restrictions();
	if (!rr_list.empty()) {
	    cout << "Inherited Role Restrictions:" << endl;
	    for_each(rr_list.begin(), rr_list.end(), rr_print_aux);
	}
    }

    {
	const Set<RoleRestriction*>& rr_list = role_restrictions();
	if (!rr_list.empty()) {
	    cout << "Role Restrictions:" << endl;
	    for_each(rr_list.begin(), rr_list.end(), rr_print_aux);
	}
    }

    /*
	End of duplicated features
     */

    {
	const Set<Individual*>& tmp = told_one_ofs();
	if (!tmp.empty()) {
	    cout << "Told One-ofs: " << endl;
	    cout << "    ";
	    for (Set<Individual*>::const_iterator i = tmp.begin(); 
		i != tmp.end(); i++)
		cout << **i << " ";
	    cout << endl;
	}
    }

    {
	const Set<Individual*>& tmp = inherited_one_ofs();
	if (!tmp.empty()) {
	    cout << "Inherited One-ofs: " << endl;
	    cout << "    ";
	    for (Set<Individual*>::const_iterator i = tmp.begin(); 
		i != tmp.end(); i++)
		cout << **i << " ";
	    cout << endl;
	}
    }

    {
	const Set<Individual*>& tmp = one_ofs();
	if (!tmp.empty()) {
	    cout << "One-ofs: " << endl;
	    cout << "    ";
	    for (Set<Individual*>::const_iterator i = tmp.begin(); 
		i != tmp.end(); i++)
		cout << **i << " ";
	    cout << endl;
	}
    }

    {
	const Set<Test*>& tmp = told_tests();
	if (!tmp.empty()) {
	    cout << "Told Tests: " << endl;
	    for (Set<Test*>::const_iterator i = tmp.begin(); 
		i != tmp.end(); i++)
		cout << "    " << **i << endl;
	}
    }

    {
	const Set<Test*>& tmp = inherited_tests();
	if (!tmp.empty()) {
	    cout << "Inherited Tests: " << endl;
	    for (Set<Test*>::const_iterator i = tmp.begin(); 
		i != tmp.end(); i++)
		cout << "    " << **i << endl;
	}
    }

    {
	const Set<Test*>& tmp = tests();
	if (!tmp.empty()) {
	    cout << "Tests: " << endl;
	    for (Set<Test*>::const_iterator i = tmp.begin(); 
		i != tmp.end(); i++)
		cout << "    " << **i << endl;
	}
    }

    const List<Rule*>& forward = RB.forward_rules((Concept*)this);
    if (!forward.empty()) {
	cout << "Direct Forward Rules: " << endl;
	for (List<Rule*>::const_iterator i = forward.begin(); 
	    i != forward.end(); i++)
	    cout << "    " << **i << endl;
    }

    const List<Rule*>& indirect = RB.indirect_rules((Concept*)this);
    if (!indirect.empty()) {
	cout << "Indirect Forward Rules: " << endl;
	for (List<Rule*>::const_iterator i = indirect.begin(); 
	    i != indirect.end(); i++)
	    cout << "    " << **i << endl;
    }

    const List<Rule*>& backward = RB.backward_rules((Concept*)this);
    if (!backward.empty()) {
	cout << "Backward Rules: " << endl;
	for (List<Rule*>::const_iterator i = backward.begin(); 
	    i != backward.end(); i++)
	    cout << "    " << **i << endl;
    }
}


static bool complex_rr(const RoleRestriction* rr)
/*
    We need to print 'and' in front of the concept description if
    there is only one RR and there are more than one item in it.
 */
{
    int items = 0;

    if (rr->low() != -1) 
	items++;
    if (rr->high() != -1)
	items++;
    if (rr->vr() != 0)
	items++;
    if (!rr->fillers().empty())
	items++;

    return items > 1;
}

/*
    Pretty-print the definition of this concept in functional 
    (Lisp-like) form.
 */
void Concept::pretty_print(ostream& cout) const
{
    if (primitive()) {
	/*
	    Primitive concepts cannot be pretty-print.  e.g.,

	    (define-primitive Ferrari SportsCar)
	    (define-primitive Lamborghini SportsCar)

	    pretty-printing both Ferrari and Lamborghini will
	    simply print the definition of SportsCar, which is
	    somewhat unsatisfactory.  So, we print the names
	    Ferrari, Lamborghini, &c, only.  The user will 
	    need to call print() for complete information on a
	    primitive concept.
	 */
	cout << name();
	return;
    }

    bool need_and = FALSE;
    /*
	Whether we need to print 'and' is really tricky.  
     */
    if (_parents.size() + 
	_told_role_restrictions.size() > 1)
	need_and = TRUE;
    if (!need_and && _told_role_restrictions.size() == 1 && 
	complex_rr((*_told_role_restrictions.begin()).second))
	need_and = TRUE;

    if (need_and)
	cout << "(and";

    for (Set<Concept*>::iterator i = _parents.begin(); i != _parents.end(); i++)
	cout << (need_and? " ": "") << **i;

    const Set<RoleRestriction*>& rr_list = told_role_restrictions();
    for (Set<RoleRestriction*>::iterator rr = rr_list.begin();
	rr != rr_list.end(); rr++) {
	if ((*rr)->vr())
	    cout << " (all " << (*rr)->name() << " " << *(*rr)->vr() << ")";
	if ((*rr)->low() != -1 && (*rr)->low() == (*rr)->high()) {
	    if ((*rr)->vr())
		cout << endl << "    ";
	    cout << " (exactly " << (*rr)->low() << " " << (*rr)->name() << ")";
	} else {
	    if ((*rr)->low() != -1) {
		if ((*rr)->vr())
		    cout << endl << "    ";
		cout << " (at-least " << (*rr)->low() << " " << 
		    (*rr)->name() << ")";
	    }
	    if ((*rr)->high() != -1) {
		if ((*rr)->vr())
		    cout << endl << "    ";
		cout << " (at-most " << (*rr)->high() << " " << 
		    (*rr)->name() << ")";
	    }
	}
	if (!(*rr)->fillers().empty()) {
	    if ((*rr)->vr() || (*rr)->low() != -1 || (*rr)->high() != -1)
		cout << endl << "    ";
	    cout << " (fills " << (*rr)->name();
	    for (List<Individual*>::const_iterator i = (*rr)->fillers().begin();
		i != (*rr)->fillers().end(); i++) {
		cout << " ";
		(*i)->pretty_print(cout);  // see operator<< in role.C
	    }
	    cout << ")";
	}
    }

    const Set<Individual*>& oo_list = one_ofs();
    if (oo_list.size()) {
	cout << " one-of(";
        for (Set<Individual*>::const_iterator i = oo_list.begin();
	    i != oo_list.end(); ) {
	    cout << **i;
	    if (++i != oo_list.end())
		cout << " ";
	}
        cout << ")";
    }

    if (need_and)
	cout << ")";
}


Concept::Concept(const String& name, Concept* concept): Object(name),
    _contradict_one_ofs(FALSE)
/*
    Concept constructor.  Define a named concept.

    A named concept is defined by the following procedure.  First, the
    input concept is copied.  Then it is normalized and classified.
 */
{
    *this = *concept;
    add_name(name);
    normalize();
    classify();
}


Concept::~Concept()
/*
    Delete all references to this concept from the world_map.  Don't
    worry about world yet, ~Object() will take care of that.
 */
{
    List<String>::iterator i;

    for (i = _names.begin(); i != _names.end(); i++)
	assert(object_base.erase(*i));
}


Concept& Concept::operator=(const Concept& concept)
/*
    The assignment operator creates an unnamed concept out of the  
    input concept (which can be named or unnamed).  Briefly all 
    the attributes are copied except the 'names' list.
 */
{
    /*
	_classified is not copied because names are not copied and
	unnamed concepts are unclassified.
     */
    _null = concept._null;
    _contradict_one_ofs = concept._contradict_one_ofs;
    _parents = concept._parents;
    _children = concept._children;
    _told_role_restrictions = concept._told_role_restrictions;
    _role_restrictions = concept._role_restrictions;
    _tests = concept._tests;
    _told_one_ofs = concept._told_one_ofs;
}


void ancestors_aux(const Concept* C, Set<Concept*>& S)
/*
    Recursively find all ancestors of a concept C.
 */
{
    if (C->_parents.empty())
	return;
    
    // S.insert(C->_parents.begin(), C->_parents.end());
    copy(C->_parents.begin(), C->_parents.end(), inserter(S));
    
    for (Set<Concept*>::iterator i = C->_parents.begin();
	i != C->_parents.end(); i++)
	ancestors_aux(*i, S);
}


Set<Concept*> Concept::ancestors() const
{
    Set<Concept*> tmp;
    tmp.insert(Thing); 	         // Thing is ancestors of every concept.
    ancestors_aux(this, tmp);    // Recursively find all ancestors.

    return tmp;
}


void descendants_aux(const Concept* C, Set<Concept*>& S)
/*
    Recursively find all descendants of a concept C.
 */
{
    if (C->_children.empty())
	return;
    
    // copy(C->_children.begin(), C->_children.end(), inserter(S, S.end()));
    // S.insert(C->_children.begin(), C->_children.end());
    copy(C->_children.begin(), C->_children.end(), inserter(S));

#if 1
    for (Set<Concept*>::iterator i = C->_children.begin();
	i != C->_children.end(); i++)
	descendants_aux(*i, S);
#else
    /*
	Just an exercise.  Notice the function descendants_aux must be
	declare as

	    int descendants_aux(const Concept* C, Set<Concept*> S)
	    ^^^                                              ^^^

	for the following to work.
     */
    for_each(C->_children.begin(), C->_children.end(), bind2nd(
	pointer_to_binary_function<const Concept*, Set<Concept*>, int>(
	    descendants_aux), S));
#endif

    return;
}


Set<Concept*> Concept::descendants() const
{
    Set<Concept*> tmp;
    descendants_aux(this, tmp);

    return tmp;
}


void instances_aux(const Concept* C, Set<Individual*>& S)
/*
    Recursively find all (indirect) instances of a concept C.
 */
{
    // S.insert(C->_direct_instances.begin(), C->_direct_instances.end());
    copy(C->_direct_instances.begin(), C->_direct_instances.end(),
        inserter(S));

    if (!C->_children.empty())
	for (Set<Concept*>::iterator i = C->_children.begin();
	    i != C->_children.end(); i++)
	    instances_aux(*i, S);
}


Set<Individual*> Concept::instances() const
{
    Set<Individual*> tmp;
    instances_aux(this, tmp);

    return tmp;
}


static bool in(const Set<RoleRestriction*>& source, const RoleRestriction* rr)
{
    for (Set<RoleRestriction*>::const_iterator i = source.begin();
	i != source.end(); i++)
	if (**i == *rr)
	    return TRUE;

    return FALSE;
}


RoleRestriction* name_in(
    const Set<RoleRestriction*>& source, 
    const RoleRestriction* rr
) {
    for (Set<RoleRestriction*>::const_iterator i = source.begin();
	i != source.end(); i++)
	if ((*i)->name() == rr->name())
	    return *i;

    return 0;
}


const Set<RoleRestriction*>& Concept::role_restrictions() const
{
    return _role_restrictions;
}


Set<RoleRestriction*> Concept::told_role_restrictions() const
{
    Set<RoleRestriction*> tmp;

    /*
	Note: copy() is not used because tmp is of type RoleRestriction*
	while _told_role_restrictions is of type pair<Role*, RR*>.
     */
    for (Map<Role*, RoleRestriction*>::const_iterator i = 
	_told_role_restrictions.begin(); 
	i != _told_role_restrictions.end(); i++)
	tmp.insert((*i).second);

    return tmp;
}


Set<RoleRestriction*> Concept::inherited_role_restrictions() const
{
    Set<RoleRestriction*> result;

    for (Set<Concept*>::iterator p = _parents.begin(); p != _parents.end(); 
	p++) {
	const Set<RoleRestriction*>& tmp1 = (*p)->inherited_role_restrictions();
	// result.insert(tmp1.begin(), tmp1.end());
	copy(tmp1.begin(), tmp1.end(), inserter(result));
	const Set<RoleRestriction*>& tmp2 = (*p)->told_role_restrictions();
	for (Set<RoleRestriction*>::iterator i = tmp2.begin(); 
	    i != tmp2.end(); i++)
	    if (!in(result, *i))
		result.insert(*i);
    }

    return result;
}


const Set<Test*>& Concept::told_tests() const
{
    return _tests;
}


Set<Test*> Concept::inherited_tests() const
{
    Set<Test*> result;

    for (Set<Concept*>::iterator p = _parents.begin(); 
	p != _parents.end(); p++) {
	const Set<Test*>& tmp1 = (*p)->inherited_tests();
	// result.insert(tmp1.begin(), tmp1.end());
	copy(tmp1.begin(), tmp1.end(), inserter(result));
	const Set<Test*>& tmp2 = (*p)->told_tests();
	// result.insert(tmp2.begin(), tmp2.end());
	copy(tmp2.begin(), tmp2.end(), inserter(result));
    }

    return result;
}


Set<Test*> Concept::tests() const
{
    Set<Test*> result = told_tests() + inherited_tests();

    return result;
}


void Concept::add_role_restriction_aux(RoleRestriction* RR)
/*
    Add a new RR into the _role_restrictions list.

    . close the role before adding
    . if closed_RR >= any rr already in the list, return
    . if any rr > closed_RR, remove it from the list.
    . add it!
 */
{
#if 0
    RoleRestriction *closed_RR = new RoleRestriction(*RR);
    closed_RR->close();
    if (!add_closed_role_restriction_aux(closed_RR))
	delete closed_RR;
#else
    /*
	We might not want to close an RR for a concept.  Consider:
	    _told_RR = (all name string)
	we don't want the RR to be
	    ((all name string) (exactly 0 name))
     */
    add_closed_role_restriction_aux(RR);
#endif
}
    

bool Concept::add_closed_role_restriction_aux(RoleRestriction* RR)
/*
    Insert an already closed RR into the _role_restrictions list.

    . if closed_RR >= any rr already in the list, return
    . if any rr > closed_RR, remove it from the list.
    . add it!

    Return TRUE if added, FALSE if not added because of the first
    condition above.
 */
{
    for (Set<RoleRestriction*>::iterator rr = _role_restrictions.begin();
	rr != _role_restrictions.end(); rr++) {
	if (**rr <= *RR)
	    return FALSE;
        if (**rr > *RR)
	    _role_restrictions.erase(rr);
    }

    RoleRestriction* RR2 = name_in(_role_restrictions, RR);
    if (RR2) {
	RoleRestriction* merged_RR = new RoleRestriction(*RR2);
	merge_role_restriction(merged_RR, RR);
	_role_restrictions.erase(RR2);
	/*
	    Close the merged RR if it is filled maximally.
	 */
	if (!merged_RR->closed() && merged_RR->number() == merged_RR->high())
	    merged_RR->close();
	_role_restrictions.insert(merged_RR);
    } else {
	/*
	    Close the merged RR if it is filled maximally.
	 */
	if (!RR->closed() && RR->number() == RR->high()) {
	    RoleRestriction* closed_RR = new RoleRestriction(*RR);
	    closed_RR->close();
	    _role_restrictions.insert(closed_RR);
	} else {
	    _role_restrictions.insert(RR);
	}
    }

    return TRUE;
}


void Concept::add_role_restriction(Role* R, RoleRestriction* RR)
/*
    A new role restriction is added to the told_role_restriction.
    The _role_restrictions of this concept as well as its descendants
    must be updated.
 */
{
    _told_role_restrictions.insert(pair<Role* const, RoleRestriction*>(R, RR));
    add_role_restriction_aux(RR);
}   


void Concept::remove_parent(Concept* C) 
{ 
    _parents.erase(C);
}


void Concept::add_parent(Concept* C) 
{ 
#if defined(Debug)
cout << "Concept::add_parent()" << endl;
#endif
    assert(this != Thing);	// one can never add parents to Thing
    _parents.insert(C); 
    inherit_role_restrictions(C);
#if defined(Debug)
cout << "Concept::add_parent() returns" << endl;
#endif
}


void Concept::inherit_role_restrictions(const Concept* C)
{
    /*
	After C is added to "this"'s parent list, the role_restrictions 
	from C should be copied into "this".
     */

    const Set<RoleRestriction*>& rr_list = C->role_restrictions();

    /*
	Notice each rr in rr_list is closed already by definition.  We can
	call a function similar to add_role_restriction_aux.
     */

    for (Set<RoleRestriction*>::const_iterator rr = rr_list.begin();
        rr != rr_list.end(); rr++)
	add_closed_role_restriction_aux(*rr);

    /*
	Recursively do the same thing for each child.  This is the
	basis for RR inheritance.
     */

    for (Set<Concept*>::iterator child = _children.begin(); 
	child != _children.end(); child++)
	(*child)->inherit_role_restrictions(this);
}


Set<Individual*> Concept::inherited_one_ofs() const
{
    Set<Individual*> result;
    for (Set<Concept*>::const_iterator i = parents().begin();
	i != parents().end(); i++) {
	const Set<Individual*>& tmp = (*i)->one_ofs();
	if (result.empty())
	    result = tmp;
	else {
	    if (!tmp.empty()) {
		Set<Individual*> intersection;
		set_intersection(
		    result.begin(), result.end(), tmp.begin(), tmp.end(), 
		    inserter(intersection), individual_comp
		);
		result = intersection;
	    }
	}
    }

    return result;
}


Set<Individual*> Concept::one_ofs() const
{
    const Set<Individual*>& inherit = inherited_one_ofs();
    const Set<Individual*>& told = told_one_ofs();
    Set<Individual*> result;
    if (inherit.size())
	if (told.empty())
	    result = inherit;
	else
	    set_intersection(
		inherit.begin(), inherit.end(), told.begin(), told.end(),
		inserter(result), individual_comp
	    );
    else
	result = told;

    return result;
}


Concept* define_concept(const String& name, Concept* C)
/*
    If the input concept is already a classifed, named concept,
    then this new name is simply an alias to it.  In this case,
    we add this name to world_map pointing to the input concept
    and add in the concept.names list, this new name.

    If the input concept is not classified, we call "new" to 
    form a new concept then let Concept ctor take over.
 */
{
    if (OB.exists(name)) {
	cout << "Symbol \"" << name << "\" exists already." << endl;
	return 0;
    }

    if (!C || C->null()) {
	cout << "Cannot define null concept." << endl;
	return 0;
    }

    /*
	First, find out whether C is primitive or not.  If C is 
	primitive, we have to call new PrimitiveConcept, if not,
	call new Concept.

	actually, we don't need to call new PrimitiveConcept as a
	second thought.  See comments below.
     */

    if (C->classified()) {
	/*
	    A classifed concept.  This is an exceptional case because if
	    C is a classifed concept, it is already a named concept.
	    Thus, we know for sure the new_concept to be defined should 
	    simply be an alias.  e.g., 

		define_concept("test_integer", Integer);

	    this is not defining a new concept "test_integer", because
	    there is nothing new.  It is simply an alias to Integer.

	    So, the action is simply add a symbolic alias.
	 */
	 C->add_name(name);
	 OB.insert(name, C->id());
	 return C;
    } else {
	/*
	    Make sure we check whether the result to see if it is
	    classified or not.  It can be unclassified after creation
	    due to two possibilities:

	     . it is equivalent to an already classified concept.
	       If this is the case, an alias will be created in world_map
	       as well as the equivalent concept.  This is taken care of
	       in classify().

	     . it is an empty concept
	 */
	Concept* new_concept = new Concept(name, C);
		
	if (!new_concept->classified()) {	// empty concept?
	    delete new_concept;
	    return 0;	
	} else
	    return new_concept;
    }
}


PrimitiveConcept* define_primitive(const String& name, Concept* concept)
{
    if (OB.exists(name)) {
	cout << "Symbol \"" << name << "\" exists already." << endl;
	return 0;
    }

    PrimitiveConcept* new_concept;

    /* 
	concept == 0 is a special case for declaring 'thing'.
     */
    if (!concept)
	if (name == "thing") {
	    new_concept = new PrimitiveConcept(name);
	    new_concept->_primitive_parent = 0;
	} else {
	    cout << "Cannot define null primitive concept." << endl;
	    return 0;
	}
    else 
	if (concept->null()) {
	    cout << "Cannot define null primitive concept." << endl;
	    return 0;
	} else {
	    /*
		Primitive concept is simple: create a new PrimitiveConcept.
	     */
	    new_concept = new PrimitiveConcept(name, concept);
	}

    if (!new_concept->classified()) {	// empty concept?
	delete new_concept;
	return 0;
    } else
	return new_concept;
}


void PrimitiveConcept::print(ostream& cout) const
{
    Concept::print();
    if (primitive_parent()) {
	cout << "Primitive Parent: ";
	cout << *primitive_parent();
	cout << endl;
    }
}


void DisjointPrimitiveConcept::print(ostream& cout) const
{
    PrimitiveConcept::print(cout);
    cout << "Partition" << endl;
    cout << "    index: " << group() << endl;
    cout << "    concept: " << *primitive_parent() << endl;
    cout << "    members: "; 
    const List<Concept*>& l = members();
    for (List<Concept*>::const_iterator i = l.begin(); i != l.end(); ) {
	cout << (*i)->name();
	i++;
	if (i != l.end())
	    cout << " ";
	else
	    cout << endl;
    }
}


List<Concept*> DisjointPrimitiveConcept::members() const
/*
    Return members of the same group of this concept.  This can be
    found out by looking at all the children concepts of the
    primtive_parent and gathered all that has the same group as this.
 */
{
    List<Concept*> result;
    const Set<Concept*>& concepts = primitive_parent()->children();

    for (Set<Concept*>::iterator i = concepts.begin(); i != concepts.end(); i++)
	if ((*i)->type() == disjoint_primitive_concept &&
	    ((DisjointPrimitiveConcept*)(*i))->group() == group())
	    result.push_back(*i);

    return result;
}


DisjointPrimitiveConcept* define_disjoint_primitive(
    const String& name, Concept* concept, const String& group
)
{
    if (OB.exists(name)) {
	cout << "Symbol \"" << name << "\" exists already." << endl;
	return 0;
    }

    if (!concept || concept->null()) {
	cout << "Cannot define null disjoint primitive concept." << endl;
	return 0;
    }

    DisjointPrimitiveConcept* new_concept;
    /*
	Primitive concept is simple: create a new DisjointPrimitiveConcept.
	That's it.
     */
    new_concept = new DisjointPrimitiveConcept(name, concept, group);

    if (!new_concept->classified()) { 	// empty concept?
	delete new_concept;
	return 0;
    } else
	return new_concept;
}


Concept* concept(const String& name)
{
    Object* obj = object(name);
    if (obj && (obj->type() == Object::concept ||
		obj->type() == Object::primitive_concept ||
		obj->type() == Object::disjoint_primitive_concept))
	return (Concept*)obj;
    else
	return 0;
}
