#if !defined(CONCEPT_H)
#define CONCEPT_H

#include "object.h"
#include "rule_base.h"

class Concept: public Object {

/*
    A concept is an object that has the following attributes:

      . a list of symbolic names
      . a list of parents
      . a list of children
      . a list of direct instances
      . a list of rules 
      . a list of back-chaining rules
      . a list of tests
      . a list of one-of restrictions
      . a map of filler and role restrictions
 */

    friend Concept* define_concept(const String& name, Concept* concept);
    friend PrimitiveConcept* define_primitive(const String&, Concept*);
    friend DisjointPrimitiveConcept* define_disjoint_primitive(
               const String& name,
               Concept* C,
               const String& group
           );
    friend class PrimitiveConcept;
    friend class ClassicIndividual;  // the ctor needs to do new Concept

    /*
	All concept expression functions are friends because they are
	like define_concept() except they create unnamed concepts. 
	Each one also has an accompanied expression filler that does
	the insertion job... but does not call new Concept.
     */

    friend Concept* and(const List<Concept*>&);
    friend Concept* all(Role*, Concept*);
    friend Concept* at_least(int, Role*);
    friend Concept* at_most(int, Role*);
    friend Concept* exactly(int, Role*);
    friend Concept* between(int, int, Role*);
    friend Concept* fills(Role*, const List<Individual*>&);
    friend Concept* one_of(const List<Individual*>&);
    friend Concept* test_c(const String&, const ObjectVector&);
    friend Concept* test_c(char*, ...);
    friend class RoleRestriction;
    friend void merge_role_restriction(RoleRestriction*, 
	       const RoleRestriction*);

protected:

    /*
        Since we want to keep track of allocation and deallocation of
        Concepts, we protect the constructors so that they can be
        accessed only by the friend functions (either define_concept
        or one of the concept expression functions).  They will always
        use "new" to allocate memory.  These are placed into either
        world or nether_world and will be deallocated by the ObjectBase
        destructor.

        Correction: the latest decision indicates only world is needed.
        See comments for ObjectBase.
     */

    /*
        A single named concept... right now, I can only think of
        "thing" that is allowed to use it.

        A "thing" can be distinguished by 1. the name "thing" and 
        2. the absence of parents.  All other concepts has at least
        "thing" as parents.

        On a second thought, PrimitiveConcept and DisjointPrimitiveConcept
        may call this constructor to define a primitive concept, then
        add parents later on.
     */
    Concept(const String& name): Object(name), _classified(TRUE), _null(FALSE),
	_contradict_one_ofs(FALSE)
    {
        add_name(name);
    }

    /*
        Define a named object.  A named concept is classified into
        the thing hierarchy.  The input concept is assumed to be an
        unclassified, unnamed concept.  The define_concept function
        will take care of defining a new name for an already classified
        concept.
     */
    Concept(const String& name, Concept*);

    /*
        Define an unnamed concept.  This is called only by the concept
	expression functions, such as "all", "fills", "at-most" etc.
     */
    Concept(): _classified(FALSE), _null(TRUE), _contradict_one_ofs(FALSE) {}

public:

    /*
        Copy constructor is not allowed. 
     */
    Concept(const Concept& concept) 
    {
        cout << "Concepts cannot be duplicated.  Use op=" << endl;
        assert(0); 
    }

   ~Concept();

    /*
        The assignment operator creates an unnamed concept out of the
        input concept (which can be named or unnamed).  Briefly all 
        the attributes are copied except the 'names' list.
     */
    Concept& operator=(const Concept& concept);

    /*
	Subsumption relationships.  > and < denotes proper subsumption
	relationships.

	Only == and < are defined.  All other operators ( !=, <=, >=, >)
	can still be used because of the precense of some STL templates.
	(except >= and <=, see below)

	Later, this turns out to be a major pain i.t.a.  In g++ 2.6.2 
	operators defined as template functions take higher precedence
	than operators defined as member functions (certainly a bug!).
	I discovered this when trying out subset relationships for Set.

	I have sent a bug report to bug-g++@prep.ai.mit.edu.  In the
	meantime, I must comment out the template definitions in the
	g++ stl.h header.  Also, see bug.C for a simple demo of this
	bug.

	------------

	As a second thought, it is much easier to define <= than to 
	define < ... I'll do that.
     */

    bool operator==(const Concept&) const;

    bool operator<=(const Concept&) const;

    /*
	The way stl defines >= and <= does not work with subsumption
	relationships.  The followings are workarounds... might be 
	more efficient if rewritten specifically.

	x >= y was defined as !(x < y) ... clearly not correct for
	subsumption or subset as {a} >= {b} ? will return true.

	I have removed <= and put in >= (to define x >= y as y <= x)
     */

#if 0
    /*
	I am taking this out of inline, seems like gcc 2.7.2 has a
	lot of problem doing inline template function resolution.
	Same experience when working with role_restr.h.  
     */
    friend bool operator<(const Concept& C, const Concept& D)
    {
    /*
	gcc 2.7.2 bug?  The following does not call Concept::op<=
	Instead, the op<= in function.h is called.

	return C <= D && C != D;
     */
	return C.operator<=(D) && C != D;
    }
#endif

    friend bool operator<(const Concept& C, const Concept& D);

    friend bool operator>(const Concept& C, const Concept& D);

    friend bool operator>=(const Concept& C, const Concept& D);

    bool instance_of(const Concept&, Individual* I) const;

    virtual Type type() const { return concept; }

    virtual String type_name() const { return "concept"; }

    virtual String name() const 
    { 
	if (_names.begin() != _names.end())
	    return *_names.begin(); 
	else
	    return "(unnamed)";
    }

    bool primitive() const 
    {
	return type() == primitive_concept || 
	       type() == disjoint_primitive_concept;
    }

    virtual void print(ostream& out = cout) const;

    virtual void pretty_print(ostream& out = cout) const;

    friend ostream& operator<<(ostream& out, const Concept&);

    bool null() const { return _null; }

    bool classified() const { return _classified; }

    bool unclassified() const { return !classified(); }

    const Set<Concept*>& parents() const
    {
	return _parents;
    }

    const Set<Concept*>& children() const
    {
	return _children;
    }

    Set<Concept*> descendants() const;

    Set<Concept*> ancestors() const;

    const Set<RoleRestriction*>& role_restrictions() const;

    Set<RoleRestriction*> told_role_restrictions() const;

    Set<RoleRestriction*> inherited_role_restrictions() const;

    Set<Test*> tests() const;

    const Set<Test*>& told_tests() const;

    Set<Test*> inherited_tests() const;

    const Set<Individual*>& direct_instances() const
    {
	return _direct_instances;
    }

    const Set<Individual*>& direct_individuals() const 
    { 
	return direct_instances(); 
    }

    Set<Individual*> instances() const;

    Set<Individual*> individuals() const { return instances(); }

    const Set<Individual*>& told_one_ofs() const 
    {
	return _told_one_ofs;
    }

    Set<Individual*> inherited_one_ofs() const;

    Set<Individual*> one_ofs() const;

private:

    void add_parent(Concept* C);

    void remove_parent(Concept* C);

    void add_child(Concept* C) { _children.insert(C); }

    void add_role_restriction(Role* R, RoleRestriction* RR);

    void add_role_restriction_aux(RoleRestriction* RR);

    bool add_closed_role_restriction_aux(RoleRestriction* RR);

    void inherit_role_restrictions(const Concept*);

    void add_one_of(Individual* I) { _told_one_ofs.insert(I); }

    void add_test(Test* t) { _tests.insert(t); }

    void add_name(const String& name) { _names.push_back(name); }

    void add_instance(Individual* I) 
    {
	_direct_instances.insert(I);

	/*
	    we need to fire all the rules that have "this" or 
	    this' ancestors as antecedents.
	 */

	RB.fire(this, I);
    }

    void remove_instance(Individual* I) { _direct_instances.erase(I); }

    /*
	primitive_parent is undefined for regular concepts.
     */
    virtual const Concept* primitive_parent() const { return 0; }

    /*
	I intentionally use _aux as a suffix instead of utilizing
	overloading here because these _aux functions are not supposed
	to be used by other users.
     */
    friend void descendants_aux(const Concept*, Set<Concept*>&);

    friend void ancestors_aux(const Concept*, Set<Concept*>&);

    friend void instances_aux(const Concept*, Set<Individual*>&);

    void classify();

    void normalize();

    bool contradict();

    void canon();

    void canon_parents();

    void canon_one_ofs();

    void merge_parents(Concept*);

    void normalize_parents();

    void normalize_children();

    bool contradict_parents();

    bool contradict_role_restrictions();

    bool classify_parents(Concept*);

    bool classify_parents_aux(Concept*, Set<Concept*>&);

    bool classify_children();

    void classify_instances();

    /*
        Remember to update operator=() when these are altered.
     */

    bool _classified;

    bool _null;

    List<String> _names;

    Set<Concept*> _parents;

    Set<Concept*> _children;

    Set<Test*> _tests;

    Set<Individual*> _told_one_ofs;

    /*
	Notice that if _told_one_ofs is empty, we do not know whether
	there is no supplied one_ofs, or the one_ofs inherited from 
	the parents contradict.  e.g.,
	   
	    (def-prim C1 (one-of "a" "b"))
	    (def-prim C2 (one-of "c" "d"))
	    (def-con CONTRADICT (and C1 C2))

	Therefore, we need a boolean flag.
     */

    bool _contradict_one_ofs;

    Map<Role*, RoleRestriction*> _told_role_restrictions;

    /*
	The following variable will be updated whenver there is a change
	in "_parents" & "_told_role_restrictions".  Also, these role
	restrictions are closed.
     */

    Set<RoleRestriction*> _role_restrictions;

    Set<Individual*> _direct_instances;
};


/*
    A primitive concept has a primitive parent.  E.g., COMMAND isa
    THING.  Then, the primitive parent of COMMAND is THING.  Similarly,
    if WINE isa THING, then the primitive parent of WINE is also THING.

    A primitive concept only carries necessary conditions.  Therefore,
    if we assert an individual to be a COMMAND, we will know what 
    necessary restrictions that individual has.  These are found in 
    the main body of the PrimitiveConcept object and are inherited from
    the primitive parent and its ancestors.
 */
class PrimitiveConcept: public Concept {

    friend PrimitiveConcept* define_primitive(const String&, Concept*);

protected:

    /*
        Basically, only "thing" is allowed here if a new 
        PrimitiveConcept(name) is not immediately followed by a
        parent.insert(...) statement.  
     */
    PrimitiveConcept(const String& name): Concept(name) {}

    /*
        Define a primitive concept under a concept.
	Since a concept is registered into Objectbase, therefore
	it must be created by calling 'new'.  We enforce this by 
	making the ctor protected and is accessible only through 
	define_primitive().
     */
    PrimitiveConcept(const String& name, Concept* C): 
	Concept(name), _primitive_parent(C) 
    {
	add_parent(C);
	normalize();
	classify();
	if (!C->classified())
	    C->add_child(this);
    }

public:

    virtual Type type() const { return primitive_concept; }

    virtual String type_name() const { return "primitive concept"; }

    virtual void print(ostream& out = cout) const;

    const Concept* primitive_parent() const { return _primitive_parent; }

private:

    Concept* _primitive_parent;	
};


class DisjointPrimitiveConcept: public PrimitiveConcept {

    friend DisjointPrimitiveConcept* define_disjoint_primitive(
               const String& name,
               Concept* C,
               const String& group
           );

protected:

    /*
        Define a disjoint primitive concept.
	Since a concept is registered into Objectbase, therefore
	it must be created by calling 'new'.  We enforce this by 
	making the ctor protected and is accessible only through 
	define_disjoint_primitive().
     */
    DisjointPrimitiveConcept(const String& name, Concept* C, const String& grp):
        PrimitiveConcept(name, C), _group(grp) {}

public:

    virtual Type type() const { return disjoint_primitive_concept; }

    virtual String type_name() const { return "disjoint primitive concept"; }

    virtual void print(ostream& out = cout) const;

    const String& group() const { return _group; }

    bool same_group(const DisjointPrimitiveConcept* C) const 
    /* 
	Another concept C is in the same disjoint group as this one if they
	have the same primitive parent and the same group.
     */ 
    {
	return primitive_parent() == C->primitive_parent() && 
	       group() == C->group();
    }

    /*
	Return members of the same group of this concept.
     */
    List<Concept*> members() const;

private:

    String _group;
};


extern Concept*& Thing;
extern Concept*& ClassicThing;
extern Concept*& HostThing;
extern Concept*& NumberThing;
extern Concept*& StringThing;
extern Concept*& Integer;
extern Concept*& Float;

extern Concept* concept(const String& name);

#endif
