#include <stdio.h>
#include <unistd.h>
#include "role_restr.h"
#include "procedure.h"
#include "expression.h"
#include "file.h"


/*
    All the functions that accept ... arguments have a _l counterpart.
    e.g., and(...) will have an and_l(List<Concept*>) that takes
    List<Concept*> as argument.  

    As a second thought, simply use overloading.

    Also, using "..." is not the best way because the ... is supposed
    to be Object*, Object* etc.  But the virtual function table is 
    lost.  For example:

	test_c("email?", name);
	     _____________/
    This name here is a role, and we should be allowed to do this:

	arg[0]->print() to examine "name".  But we can't after name
	is passed through "...".
 */


Concept* and(Concept* C, ...)
/*
    Create a conjunctive concept.  Each argument should be a Concept*,
    the last Concept* must be 0 to denote end of list.
 */
{
    va_list ap;
    List<Concept*> concepts;

    va_start(ap, C);
    while (C) {
	concepts.push_back(C);
	C = va_arg(ap, Concept*);
    }
    va_end(ap);

    return and(concepts);
}


Concept* and(const List<Concept*>& concepts)
/*
    Actually create an unnamed conjunctive concept by calling new Concept
    and add the list of concepts into its parent list.
 */
{
    Concept* conjunction = new Concept;

    for (List<Concept*>::const_iterator i = concepts.begin();
	i != concepts.end(); i++)
	conjunction->add_parent(*i);

    conjunction->normalize();

    /*
	There is a very special case here.  After normalization, 
	there is only one parent left and is a classified concept,
	and there is no further role restrictions and there is no
	further tests

	e.g., (and integer number thing) will become (and integer)

	In this case, return the classified concept.
     */

    if (conjunction->_parents.size() == 1 
	     && 
       (*conjunction->_parents.begin())->classified()
	     &&
        conjunction->_told_role_restrictions.empty()
	     &&
        conjunction->_told_one_ofs.empty()
	     &&
        conjunction->_tests.empty()) {
	Concept* tmp = *(conjunction->_parents.begin());
	delete conjunction;
	return tmp;
    } else
	return conjunction;
}


Concept* all(Role* R, Concept* C)
/*
    Return a concept that has a role R and the value restriction of
    it is the concept C.

    C has to be a normalized concept.  I don't think any concept can
    be unnormalized though because a concept is created by any of the
    concept expressions, each has called normalize() before return.
 */
{
    assert(R);
    assert(C);
    if (!OB.exists(R->name())) {
	cout << "Role \"" << R->name() << "\" does not exist." << endl;
	return 0;
    }
    if (C->null()) {
	cout << "Cannot use null concept as value restriction." << endl;
	return 0;
    }

    RoleRestriction* RR = new RoleRestriction(R);
    RR->_vr = C;

    if (R->attribute())
	RR->_nr = pair<int, int>(0, 1);

    Concept* new_concept = new Concept;
    new_concept->add_role_restriction(R, RR);

    /*
	Single value restriction does not need to be normalized.
	But we'll call it just to set the null flag.
     */

    new_concept->normalize();

    return new_concept;
}


Concept* at_least(int low, Role* R)
/*
    Return a concept that has a role R and the number restriction of
    it is [low..?]
 */
{
    return between(low, -1, R);
}


Concept* at_most(int high, Role* R)
/*
    Return a concept that has a role R and the number restriction of
    it is [?..high]
 */
{
    return between(-1, high, R);
}


Concept* exactly(int n, Role* R)
/*
    Return a concept that has a role R and the number restriction of
    it is [n..n]
 */
{
    return between(n, n, R);
}


Concept* between(int low, int high, Role* R)
/*
    Return a concept that has a role R and the number restriction of
    it is [low..high]
 */
{
    assert(R);
    if (!OB.exists(R->name())) {
	cout << "Role \"" << R->name() << "\" does not exist." << endl;
	return 0;
    }

    RoleRestriction* RR = new RoleRestriction(R);
    RR->_nr = pair<int, int>(low, high);
    Concept* new_concept = new Concept;
    new_concept->add_role_restriction(R, RR);

    /*
	Single value restriction does not need to be normalized.
	But we'll call it just to set the null flag.
     */

    new_concept->normalize();

    return new_concept;
}


Concept* fills(Role* R, ...)
/*
    Create a fills concept.  Each argument should be an Individual*,
    the last Individual* must be 0 to denote end of list.
 */
{
    va_list ap;
    List<Individual*> individuals;
    Individual* I;

    va_start(ap, R);
    while (I = va_arg(ap, Individual*))
	individuals.push_back(I);
    va_end(ap);

    return fills(R, individuals);
}


Concept* fills(Role* R, const List<Individual*>& fillers)
/*
    Actually create an unnamed fills concept by calling new Concept
    and add the list of individuals into its filler list.
 */
{
    assert(R);
    if (!OB.exists(R->name())) {
	cout << "Role \"" << R->name() << "\" does not exist." << endl;
	return 0;
    }

    RoleRestriction* RR = new RoleRestriction(R);
    RR->_fillers = fillers;

    Concept* new_concept = new Concept;
    new_concept->add_role_restriction(R, RR);
    new_concept->normalize();

    return new_concept;
}


Concept* one_of(Individual* C, ...)
/*
    Create a one-of concept.  Each argument should be an Individual*,
    the last Individual* must be 0 to denote end of list.
 */
{
    va_list ap;
    List<Individual*> inds;

    va_start(ap, C);
    while (C) {
	inds.push_back(C);
	C = va_arg(ap, Individual*);
    }
    va_end(ap);

    return one_of(inds);
}


Concept* one_of(const List<Individual*>& inds)
{
    Concept* new_concept = new Concept;

    copy(inds.begin(), inds.end(), inserter(new_concept->_told_one_ofs));
    new_concept->normalize();

    return new_concept;
}


Concept* test_c(char* name, ...)
/*
    Create a test concept.  Each argument (except the first) can be a 
    Concept* or an Individual* (Object*, collectively), the last Object* 
    must be a 0 to denote end of list.
 */
{
    va_list ap;
    Object* obj;
    ObjectVector args;

    va_start(ap, name);
    while (obj = va_arg(ap, Object*))
	args.push_back(obj);
    va_end(ap);

    return test_c(name, args);
}


Concept* test_c(const String& name, const ObjectVector& args)
{
    Test* new_test = new Test(name, args);
    Concept* new_concept = new Concept;
    new_concept->add_test(new_test);
    new_concept->normalize();

    return new_concept;
}


ComputedConcept* computed_concept(char* name, ...)
/*
    Create a computed concept.  Each argument (except the first) can be a 
    Concept* or an Individual* (Object*, collectively), the last Object* 
    must be a 0 to denote end of list.
 */
{
    va_list ap;
    Object* obj;
    ObjectVector args;

    va_start(ap, name);
    while (obj = va_arg(ap, Object*))
	args.push_back(obj);
    va_end(ap);

    return computed_concept(name, args);
}


ComputedConcept* computed_concept(const String& name, const ObjectVector& args)
{
    return new ComputedConcept(name, args);
}


ComputedFillers* computed_fillers(char* name, Role* R, ...)
/*
    Create a computed-fillers.  Each argument (except the first) can be a 
    Concept* or an Individual* (Object*, collectively), the last Object* 
    must be a 0 to denote end of list.
 */
{
    va_list ap;
    Object* obj;
    ObjectVector args;

    va_start(ap, R);
    while (obj = va_arg(ap, Object*))
	args.push_back(obj);
    va_end(ap);

    return computed_fillers(name, R, args);
}


ComputedFillers* computed_fillers(const String& name, Role* R, 
    const ObjectVector& args)
{
    return new ComputedFillers(name, R, args);
}


bool evaluate(const String& expression)
/*
    There doesn't seem to be a simpler way to evaluate an expression
    besides writing it to a temp file and using it as input to yacc.
 */
{
    char fname[80];
    tmpnam(fname);

    FILE* f = fopen(fname, "w");
    assert(f);
    fputs(expression, f);
    putc('\n', f);
    fclose(f);
    
    extern void eval_load(const char*);
    eval_load(fname);
    unlink(fname);

    return TRUE;
}
