#include "concept.h"
#include "individual.h"
#include "role_restr.h"


/*
    Caller (define_individual()) should guarantee C is not null and name is
    non-existing.

    _descriptor is a copy of the incoming concept.  This concept may
    then be modified slightly.  E.g., roles may be closed.
 */
ClassicIndividual::ClassicIndividual(const String& name, Concept* C):
    Individual(name), _name(name)
{
    assert(!C->null());

    /*
	In this trimmed-down version, fillers cannot be asserted after
	the creation of an individual.  Therefore, we assume all roles
	are closed.
     */

     /*
	Now the implementation does not actually close the roles of
	descriptors.  It always makes a copy if any role has to be 
	closed.  Therefore, we can do a simple pointer assignment.
	Also, If the incoming concept C is a simple primitive concept,
	such as 

	    C = define_primitive("C", ClassicThing);

	then, the assignment doesn't seem to work.  The resulting 
	descriptor, D, will fail D <= C, probably because they are
	bnth primitive, having the same parents, but names are not
	copied.  In fact, in this case, we must use simple pointer
	assignment to solve the problem.

	_descriptor = new Concept;
	*_descriptor = *C;
     */

#if 0
    if (C->classified() && C->primitive())
	_descriptor = C;
    else {
	_descriptor = new Concept;
	*_descriptor = *C;
    }
#else
    _descriptor = C;
#endif

/*
    The original design was to simply use a _descriptor that has a
    copy of the defining concept to represent the whole individual.
    Since every feature of an ind can be extracted from it, I thought
    this should be enough.  E.g. to define I < C, we can simply use
    I._descriptor <= C.  Unfortunately, it seems like the close-role
    operator requires modification of _descriptor, which is fine for
    the direct parent, but not good for inherited roles.
    
    for (Map<Role*, RoleRestriction*>::iterator rr = 
	_descriptor->_told_role_restrictions.begin()
*/

    _role_restrictions = _descriptor->role_restrictions();

    classify();
}


int individual_rr_print_aux(const RoleRestriction* rr)
{
#if 0
/*
    Declaring a closed_rr causes a new object to be created.  Not efficient
    but more correct.
 */
    RoleRestriction closed_rr = *rr;
    closed_rr.close();
    cout << "    " << closed_rr << endl;
#else
    cout << "    " << *rr << endl;
#endif
    return 0;
}


void ClassicIndividual::print(ostream& cout) const
{
    Object::print(cout);

    cout << "Name: " << name() << endl;

    extern int concept_print_aux(const Concept*);
    extern rr_print_aux(const RoleRestriction*&);

    Set<Concept*> tmp = parents();
    Set<Concept*>::iterator i;
    if (!tmp.empty()) {
	cout << "Parents:" << endl;
	/*
	    BUG!! internal compiler error in g++ 2.6.2.  Notice
	    this same line appears in Concept::print() w/o problem.

	    for_each(tmp.begin(), tmp.end(), concept_print_aux);
	 */
	for (i = tmp.begin(); i != tmp.end(); i++)
	    concept_print_aux(*i);

	tmp = ancestors();
	if (!tmp.empty()) {
	    cout << "Ancestors:" << endl;
	    /*
		BUG!! ditto

		for_each(tmp.begin(), tmp.end(), concept_print_aux);
	     */
	    for (i = tmp.begin(); i != tmp.end(); i++)
		concept_print_aux(*i);
	}
    }

    Set<RoleRestriction*> rr_list;
/*
    rr_list = _descriptor->told_role_restrictions();
    if (!rr_list.empty()) {
	cout << "Told Role Restrictions:" << endl;
	for_each(rr_list.begin(), rr_list.end(), individual_rr_print_aux);
    }

    rr_list = _descriptor->inherited_role_restrictions();
    if (!rr_list.empty()) {
	cout << "Inherited Role Restrictions:" << endl;
	for_each(rr_list.begin(), rr_list.end(), individual_rr_print_aux);
    }
 */

    rr_list = role_restrictions();
    if (!rr_list.empty()) {
	cout << "Role Restrictions:" << endl;
	for_each(rr_list.begin(), rr_list.end(), individual_rr_print_aux);
    }

    if (!_fired_rules.empty()) {
	cout << "Rules Fired:" << endl;
	for (Set<Rule*>::iterator r = _fired_rules.begin();
	    r != _fired_rules.end(); r++) 
	    cout << "    " << **r << endl;
    }

    if (!_test_results.empty()) {
	cout << "Test Results:" << endl;
	for (Map<Test*, TestResult>::const_iterator r = _test_results.begin();
	    r != _test_results.end(); r++) 
	    cout << "    " << *((*r).first) << " = " << (*r).second << endl;
    }
}


Set<Concept*> Individual::ancestors() const
{
    extern void ancestors_aux(const Concept*, Set<Concept*>&);
    Set<Concept*> tmp = _parents;
    for (Set<Concept*>::const_iterator i = _parents.begin(); 
	i != _parents.end(); i++)
        ancestors_aux(*i, tmp);

    return tmp;
}


ClassicIndividual* define_individual(const String& name, Concept* C)
{
    if (OB.exists(name)) {
	cout << "Symbol \"" << name << "\" exists already." << endl;
	return 0;
    }

    if (!C || C->null()) {
	cout << "Cannot instantiate null concept." << endl;
	return 0;
    }

    ClassicIndividual* I = new ClassicIndividual(name, C);

    return I->classified()? I: 0;
}


RoleRestriction* ClassicIndividual::role_restriction(const String& name) const 
{
    const Set<RoleRestriction*>& rr = role_restrictions();
    for (Set<RoleRestriction*>::const_iterator i = rr.begin(); 
	i != rr.end(); i++)
	if ((*i)->name() == name) 
	    return *i;

    return 0;
}


List<Individual*> ClassicIndividual::fillers(const String& name) const
{
    RoleRestriction* rr = role_restriction(name);
    assert(rr);
    return rr->fillers();
}


List<Individual*> ClassicIndividual::fillers(const Role* R) const
{
    RoleRestriction* rr = role_restriction(R);
    assert(rr);
    return rr->fillers();
}


Individual* ClassicIndividual::filler(const String& name) const
{
    RoleRestriction* rr = role_restriction(name);
    assert(rr);
    assert(rr->role()->attribute() || rr->fillers().size() <= 1);
    return rr->fillers().size()? rr->fillers().front(): 0;
}


Individual* ClassicIndividual::filler(const Role* R) const
{
    RoleRestriction* rr = role_restriction(R);
    assert(rr);
    assert(rr->role()->attribute() || rr->fillers().size() <= 1);
    return rr->fillers().size()? rr->fillers().front(): 0;
}


unsigned ClassicIndividual::number(const Role* R) const
{
    RoleRestriction* rr = role_restriction(R);
    assert(rr);
    return rr->number();
}


unsigned ClassicIndividual::number(const String& name) const
{
    RoleRestriction* rr = role_restriction(name);
    assert(rr);
    return rr->number();
}


bool ClassicIndividual::add_fillers(Role* R, const List<Individual*>& F)
{
    if (role_restriction(R)) {
	/*
	    If the new fillers F are subsets of the original fillers,
	    then no need to do anything.
	 */
	const List<Individual*>& f = fillers(R);
	for (List<Individual*>::const_iterator i = F.begin(); i != F.end(); 
	    i++) {
	    bool found = FALSE;
	    for (List<Individual*>::const_iterator j = f.begin(); 
		j != f.end(); j++)
		if (**j == **i) {
		    found = TRUE;
		    break;
		}
	    if (!found)
		goto add;
	}
	/*
	    If we are here, then all F are found in f.
	 */
	return TRUE;
    }
    
add:
    RoleRestriction* RR = new RoleRestriction(R);
    RR->_fillers = F;
    _descriptor->add_role_restriction(R, RR);
    add_role_restriction_aux(RR);
    normalize();
    dirty();

    return TRUE;
}


Individual* individual(const String& name)
{
    Object* obj = object(name);
    if (obj && obj->type() == Object::individual) 
	return (Individual*)obj;
    else
	return 0;
}


void ClassicIndividual::add_parent(Concept* C)
{ 
    Individual::add_parent(C); 
    inherit_role_restrictions(C);
}


void ClassicIndividual::remove_parent(Concept* C)
{
    Individual::remove_parent(C); 
}


void ClassicIndividual::inherit_role_restrictions(const Concept* C)
{
    /*
        After C is added to "this"'s parent list, the role_restrictions
        from C should be copied into "this".
     */

    const Set<RoleRestriction*>& rr_list = C->role_restrictions();
     
    /*
        Notice each rr in rr_list is closed already by definition.  We can
        call a function similar to add_role_restriction_aux.
     */
      
    for (Set<RoleRestriction*>::const_iterator rr = rr_list.begin();
        rr != rr_list.end(); rr++)
        add_closed_role_restriction_aux(*rr);
}


void ClassicIndividual::add_role_restriction_aux(RoleRestriction* RR)
/*  
    Add a new RR into the _role_restrictions list.

    . close the role before adding
    . if closed_RR >= any rr already in the list, return
    . if any rr > closed_RR, remove it from the list.
    . add it!
 */ 
{
    RoleRestriction *closed_RR = new RoleRestriction(*RR);
    closed_RR->close();
    if (!add_closed_role_restriction_aux(closed_RR))
        delete closed_RR;
}

   
bool ClassicIndividual::add_closed_role_restriction_aux(RoleRestriction* RR)
/*  
    Insert an already closed RR into the _role_restrictions list.

    . if closed_RR >= any rr already in the list, return
    . if any rr > closed_RR, remove it from the list.
    . add it!

    Return TRUE if added, FALSE if not added because of the first
    condition above.
 */
{ 
    for (Set<RoleRestriction*>::iterator rr = _role_restrictions.begin();
        rr != _role_restrictions.end(); rr++) {
        if (*RR >= **rr)
            return FALSE;
        if (**rr > *RR)
            _role_restrictions.erase(rr);
    }
    
    _role_restrictions.insert(RR);

    return TRUE;
}


void ClassicIndividual::close_role(const Role* R)
{
    close_role_aux(R);
    classify();
}


void ClassicIndividual::close_role(const String& role_name)
{
    RoleRestriction* RR = role_restriction(role_name);
    assert(RR);
    close_role_aux(RR->role());
    classify();
}


void ClassicIndividual::close_role_aux(const Role* R)
{
    assert(R);
    RoleRestriction* RR = role_restriction(R);
    assert(RR);

    if (RR->closed())
	return;

    RoleRestriction* new_RR = new RoleRestriction(*RR);
    new_RR->close();
    _role_restrictions.erase(RR);
    _role_restrictions.insert(new_RR);
}


int individual_comp(const Individual* a, const Individual* b)
/*
    A comparer for set functions is supposed to return 0 for equal and
    non-zero for not equal.
 */
{
    return (*a != *b);
}


static void buggy4132()
/*
    Need to call list<Individual*>::operator= once because 
    list<Individual*> is a parent of List<Individual*> and
    for some reason operator= is not inherited properly.
 */
{
    list<Individual*> a;

    a = a;
}
