#if !defined(INDIVIDUAL_H)
#define INDIVIDUAL_H

#include "role_restr.h"
#include "rule.h"

class Individual: public Object {

/*
    There are two kinds of individuals.  A host individual is simply a 
    number or a string.  A Classic individual has a full description
    ala a concept.  They are implemented as subclasses.  

    Individuals are registered into the ObjectBase, so that they can be
    destroyed automatically when the system exits.

    It is not obvious why HostIndividuals also require to be registered
    into the ObjectBase.  The reason is when HostIndividuals are used as
    fillers, the user will need to use "new" to create a HostIndividual
    pointer, even for integers or strings, as a result, it is error-prone
    to let the user do delete instead of the system.
 */

public:
    
    enum SubType {
	integer_individual,
	float_individual,
	string_individual,
	classic_individual
    };

    Individual(SubType type): _subtype(type) {}

    Individual(const String& name): Object(name),
	_subtype(classic_individual) {}

    virtual ~Individual() {}
    
    virtual Object::Type type() const { return Object::individual; }

    virtual String type_name() const { return "individual"; }

    SubType subtype() const { return _subtype; }

    virtual void pretty_print(ostream& out = cout) const = 0;

    virtual void print(ostream& out = cout) const = 0;

    friend ostream& operator<<(ostream& out, const Individual& I) 
    {
	I.pretty_print(out);
	return out;
    }

    bool host() const { return !classic(); } 

    bool classic() const { return subtype() == classic_individual; }

    bool classified() const { return _classified; }

    Set<Concept*> parents() const
    {
	return _parents;
    }

    Set<Concept*> ancestors() const;

    virtual bool operator==(const Individual&) const = 0;

    /*
	Here is a long struggle about what to do with casting an
	Individual into its host value.  It is conceivable that any
	individual can be casted into int, float or String.  If it
	is not a corresponding host individual, then an error should
	occur.  If it is, the value is returned.  This is a very
	convenient offering to the programmer.  Making the cast
	operators virtual is a possible method suggested to me by
	Ka Cheong Poon (poon@cs.wisc.edu).  I was using a union of
	int_ptr, float_ptr and String_ptr before this.
     */

    /*
        The default action of each is error because only HostIndividual
	can be casted into host values.
     */ 

    virtual operator int() const    { assert(0); }

    virtual operator float() const  { assert(0); }

    virtual operator String() const { assert(0); }

protected:

    friend void Concept::classify_instances();

    virtual void add_parent(Concept* C) { _parents.insert(C); }

    virtual void remove_parent(Concept* C) { _parents.erase(C); }

    friend bool SimpleRule::fire(Individual*);
    friend bool ConceptRule::fire(Individual*);
    friend bool FillersRule::fire(Individual*);
    friend void Rule::init();

    void add_rule(Rule* R) { _fired_rules.insert(R); }

protected:

    /* only used directly by HostIndividual::parent() */

    Set<Concept*> _parents;

    bool _classified;

    Set<Rule*> _fired_rules;

    friend class Test;

    Map<Test*, TestResult> _test_results;

    SubType _subtype;
};


class HostIndividual: public Individual {

public:

    HostIndividual(SubType type): Individual(type) { _classified = FALSE; }

    Concept* parent() const { return *_parents.begin(); }

    void pretty_print(ostream& out = cout) const { out << "Unknown"; }

    void print(ostream& out = cout) const { out << "Unknown" << endl; }

    bool operator==(const Individual& i) const { return this == &i; }
};


class IntegerIndividual: public HostIndividual {

public:
  
    IntegerIndividual(int value): 
	HostIndividual(integer_individual), _value(value) 
    {
	add_parent(Integer);
    }

    void pretty_print(ostream& out = cout) const { out << _value; }

    void print(ostream& out = cout) const { pretty_print(out); out << endl; }

    int value() const { return _value; }

    int operator*() const { return _value; }

    operator int() const { return _value; }

    bool operator==(const Individual& i) const
    {
	Set<Concept*> P = parents();
	if (P.find(Integer) == P.end())
	    return FALSE;
	return value() == ((const IntegerIndividual&)i).value();
    }

    bool operator==(int i) const
    {
	return value() == i;
    }

    bool operator>(int i) const
    {
	return value() > i;
    }

    bool operator<(int i) const
    {
	return value() < i;
    }

    bool operator!=(int i) const
    {
	return value() != i;
    }

    bool operator>=(int i) const
    {
	return value() >= i;
    }

    bool operator<=(int i) const
    {
	return value() <= i;
    }

private:

    int _value;
};


class FloatIndividual: public HostIndividual {

public:
  
    FloatIndividual(float value): 
	HostIndividual(float_individual), _value(value) 
    {
	add_parent(Float);
    }

    void pretty_print(ostream& out = cout) const { out.form("%.2f", _value); }

    void print(ostream& out = cout) const { pretty_print(out); out << endl; }

    float value() const { return _value; }

    float operator*() const { return _value; }

    operator float() const { return _value; }

    bool operator==(const Individual& i) const
    {
	Set<Concept*> P = parents();
	if (P.find(Float) == P.end())
	    return FALSE;
	return value() == ((const FloatIndividual&)i).value();
    }

    bool operator==(float i) const
    {
	return value() == i;
    }

    bool operator>(float i) const
    {
	return value() > i;
    }

    bool operator<(float i) const
    {
	return value() < i;
    }

    bool operator!=(float i) const
    {
	return value() != i;
    }

    bool operator>=(float i) const
    {
	return value() >= i;
    }

    bool operator<=(float i) const
    {
	return value() <= i;
    }

private:

    float _value;
};


class StringIndividual: public HostIndividual {

public:
  
    StringIndividual(const String& value): 
	HostIndividual(string_individual), _value(value) 
    {
	add_parent(StringThing);
    }

    void pretty_print(ostream& out = cout) const 
    {
	out << '"' << _value << '"'; 
    }

    void print(ostream& out = cout) const { pretty_print(out); out << endl; }

    String value() const { return _value; }

    String operator*() const { return _value; }

    operator String() const { return _value; }

    bool operator==(const Individual& i) const
    {
	Set<Concept*> P = parents();
	if (P.find(StringThing) == P.end())
	    return FALSE;
	return value() == ((const StringIndividual&)i).value();
    }

    bool operator==(const String& i) const
    {
	return value() == i;
    }

    bool operator>(const String& i) const
    {
	return value() > i;
    }

    bool operator<(const String& i) const
    {
	return value() < i;
    }

    bool operator!=(const String& i) const
    {
	return value() != i;
    }

    bool operator>=(const String& i) const
    {
	return value() >= i;
    }

    bool operator<=(const String& i) const
    {
	return value() <= i;
    }

private:

    String _value;
};


/*
    These simple inline functions are provided for easy access to the
    HostIndividuals.  With these, instead of doing

	fills(R,
	      new StringIndividual("abc"),
	      new IntegerIndividual(17),
	      ...
	      0);

    the user may simply do

        fills(R,
	      ind("abc"),
	      ind(17),
	      ...
	      0);

    Notice that the user does not need to use delete operator on the
    allocated individuals (although 'new' is used below to allocate
    them) because the individual is registered into the ObjectBase 
    and the ObjectBase destructor will do so.  However, if the user
    wishes to do that, that will cause no harm either.  See comments
    in class Individual.
 */

inline StringIndividual* ind(const String& s) 
{
    return new StringIndividual(s);
}

inline IntegerIndividual* ind(int s) 
{
    return new IntegerIndividual(s);
}

inline FloatIndividual* ind(float s) 
{
    return new FloatIndividual(s);
}


class ClassicIndividual: public Individual {

/*
    This is a more important class than the above "host individuals".
    A ClassicIndividual has a description of roles and tests.  It has
    a name, and, like a named concept, is classified under the Thing
    taxonomy.  It is also registered into the ObjectBase (therefore
    it is also derived from Object.

    The original design was to simply use a _descriptor that has a 
    copy of the defining concept to represent the whole individual.
    Since every feature of an ind can be extracted from it, I thought
    this should be enough.  E.g. to define I < C, we can simply use
    I._descriptor <= C.  Unfortunately, it seems like the close-role
    operator requires modification of _descriptor, which is fine for
    the direct parent, but not good for inherited roles.

    There are two possible solutions.  1. define a list of RR in an
    individual and refine I < C to compare these RRs with C's RRs.
    2. since close-role is imperative for each individual (this is
    true in our system, but not true in the real CLASSIC), we may
    simply modify I < C to take advantage of this knowledge without
    implementing a map.

    Taking solution #2, we need a boolean flag in each concept object
    that is set to true when that concept is actually an individual
    descriptor.
 */

    friend ClassicIndividual* define_individual(const String&, Concept*);
    friend bool SimpleRule::fire(Individual*);
    friend bool ConceptRule::fire(Individual*);
    friend bool FillersRule::fire(Individual*);

public:

    virtual ~ClassicIndividual() 
    {
	assert(OB.erase(_name));
    }
    
    /*
	Since both of the parent classes have type() and type_name()
        defined... we must redefine them here again to resolve the
	ambiguity when it arises.
     */

    Type type() const { return individual; }

    String type_name() const { return "individual"; }

    void pretty_print(ostream& out = cout) const 
    {
	out << name();
    }

    void print(ostream& out = cout) const;

    String name() const { return _name; }

    bool operator==(const Individual& i) const 
    {
	return (this == &i || name() == i.name());
    }

    Set<RoleRestriction*> role_restrictions() const
    {
	return _role_restrictions;
    }

    RoleRestriction* role_restriction(const String& name) const;

    RoleRestriction* role_restriction(const Role* R) const 
    {
	return role_restriction(R->name());
    }

    void inherit_role_restrictions(const Concept*);

    List<Individual*> fillers(const String& name) const;

    List<Individual*> fillers(const Role*) const;

    Individual* filler(const String& attribute_name) const;

    Individual* filler(const Role* attribute) const;

    unsigned number(const Role* role) const;

    unsigned number(const String& role_name) const;

    bool classify(Concept*); 

    void close_role(const String& role_name);

    void close_role(const Role* role);

    void close_role_aux(const Role* role);

protected:

    bool add_fillers(Role*, const List<Individual*>&);

    void add_parent(Concept* C);

    void remove_parent(Concept* C);

    void add_role_restriction_aux(RoleRestriction*);

    bool add_closed_role_restriction_aux(RoleRestriction*);

    /*
	Since an individual is registered into Objectbase, therefore
	it must be created by calling 'new'.  We enforce this by 
	making the ctor protected and is accessible only through 
	define_individual().
     */

    ClassicIndividual(const String& name, Concept* C);

    void classify();

    bool classify_parents(Concept*);

    bool classify_parents_aux(Concept*, Set<Concept*>&);

    void normalize();

    void normalize_parents();

    bool contradict();

    bool contradict_parents();

    bool contradict_role_restrictions();

    bool contradict_tests();

    void dirty()
    {
	_dirty = TRUE;
    }

private:

    String _name;

    Concept* _descriptor;

    Set<RoleRestriction*> _role_restrictions;

    /* isa requires access to _descriptor.  */

    friend bool isa(const ClassicIndividual& I, const Concept& C);

    bool _dirty;
};


#if 0

/*
    The following set operators are very misleading.  It hides
    the fact that the comparison is not done element by element
    (an element is an Individual*), rather it does so by first
    de-referencing the pointer.  It is replaced by a set_subset
    function defined in stl_.h in the same spirit as stl's own
    set_ functions.

    Set <= Set is still allowed.  But it is done by pure element
    comparisons.  
 */

inline bool operator<=(const Set<Individual*>& s1, const Set<Individual*>& s2)
{
    Set<Individual*>::iterator i, j;
    for (i = s1.begin(); i != s1.end(); i++) {
	for (j = s2.begin(); j != s2.end(); j++)
	    if (**i == **j)
		break;
	if (j == s2.end())
	    return FALSE;
    }
    return TRUE;
}

/*
    Comparing sets of individuals.
 */

inline bool operator==(const Set<Individual*>& s1, const Set<Individual*>& s2)
{
    return s1 <= s2 && s1.size() == s2.size();
}


inline bool operator<(const Set<Individual*>& s1, const Set<Individual*>& s2)
{
    return s1 <= s2 && s1.size() < s2.size();
}

#endif

extern Individual* individual(const String& name);

extern int individual_comp(const Individual*, const Individual* b);

#endif
