#include "concept.h"
#include "individual.h"
#include "role_restr.h"


void Concept::normalize()
/*
    A concept needs to be normalized.  The following functions are 
    performed individually:

      . normalize parents to eliminate indirect ancestors
      . normalize children to eliiminate indirect descendants

    It also decides whether this concept is null or not (contradiction)
 */
{
#if defined(Debug)
cout << "normalize()" << endl;
#endif
    canon();

    normalize_parents();
    normalize_children();

    contradict();
#if defined(Debug)
cout << "normalize() returns" << endl;
#endif
}


void merge_role_restriction(RoleRestriction* this_rr, 
     const RoleRestriction* that_rr)
/*
    Essentially merging two rr's means we have something like this:
	
	(all R C1)
	(all R C2)

    so we have to merge the above to  become

	(all R (and C1 C2))

    Each rr is of the following form:

	number restriction: low..high (-1 means don't care)
	value restriction:  a concept
	fillers:            a list of individuals
 */
{
/*
    To merge the two nr's [this_low..this_high] and [that_low..that_high]:

	if that_low > this_low then this_low = that_low
	if that_high < this_high then this_high = that_high
	however, ignore if -1
 */
    int& this_low = this_rr->_nr.first;
    int& this_high = this_rr->_nr.second;
    int that_low = that_rr->_nr.first;
    int that_high = that_rr->_nr.second;

    if (that_low != -1 && that_low > this_low)
        this_low = that_low;
    if ((this_high == -1) || (that_high != -1 && that_high < this_high))
	this_high = that_high;

/*
    To merge the two fillers... simply copy.

    (I do not understand whether this is my misunderstanding of STL 
    iterators or a bug in g++ or a bug in stl.h... I have almost
    exactly the same line as below in role.C, the compiler will
    complain until I use const_iterator.  On the other hand, this
    same line here must use iterator.
 */
    for (List<Individual*>::const_iterator i = that_rr->_fillers.begin();
	i != that_rr->_fillers.end(); i++)
	this_rr->add_filler(*i);

/*
    To merge the vr's... i.e., if there is a vr

	if that_vr < this_vr then this_vr = that_vr.
	if this_vr <= that_vr then ignore.
	if the concepts are not comparable... merge them.

    then, normalize the new combined concept.
 */
    if (that_rr->_vr) {
	Concept*& this_vr = this_rr->_vr;
	Concept* that_vr = that_rr->_vr;

	if (*that_vr < *this_vr) {
	/*
	    that_vr is more specialized and should be used as this's vr.

	    We should first delete this's vr.  However, do not delete if
	    it is classified, because it would be something like integer
	    or string or other in-used concepts.  Actually, to be very
	    economical, we don't need to delete if that_vr is also 
	    unclassified... but let's keep it very simple.

	    Then we assign that_vr to this_vr.  Notice that we need to 
	    allocate memory if that_vr is not classified.  See comments
	    in RoleRestriction's copy constructor in role.C.
	 */
	    if (this_vr->unclassified())
		delete this_vr;
	    if (that_vr->classified())
		this_vr = that_vr;
	    else {
		this_vr = new Concept;
		*this_vr = *that_vr;
	    }
	} else {
	    if (!(*this_vr <= *that_vr)) {
	    /*
		Here we have this situation:  (all R C1) (all R C2), so
		merge C1 and C2 to get (all R (and C1 C2)).  There is a
		small catch, need to allocate memory if C1 is classified,
		because now it is not classified having added C2.  Again,
		see comments in RoleRestrictions's copy constructor in
		role.C.
	     */
		if (this_vr->classified()) {
		    Concept* tmp = this_vr;
		    this_vr = new Concept;
		    this_vr->add_parent(tmp);
		}
		this_vr->merge_parents(that_vr);
	    }
	}
	this_vr->normalize();
    }
}


void Concept::merge_parents(Concept* u)
/*
    Given an unclassified concept [e.g., (and test1 test2 (fills ...))
    or (fills ...) ... ] expand it and merge the components into this.

    This is probably better named as merge because what it does is to
    merge the concept descriptions of u into this.  It also recursively
    calls itself if it turns out any description it is merging needs to
    be elaborated further.  This constitutes the canonization process.

    ... NB: I have changed the name from expand_parents to merge_parents.
 */
{
    /*
	u is supposedly unclassified, but when called by merge_role_
	restriction(), it could be classified.  If so, simply add it 
	to the parent and return.  
     */

    if (u->classified()) {
	add_parent(u);
	return;
    }

    /*
	First copy the parents.  Call merge_parents() recursively
	so that the parent list will only consist of named concepts.
     */

    for (Set<Concept*>::iterator p = u->_parents.begin();
	p != u->_parents.end(); p++) 
	if ((*p)->classified()) {
	    add_parent(*p);
	}
	else
	    merge_parents(*p);

    /*
	Second, merge role restrictions.  

	For each RoleRestriction RR in u, find if there is a
	corresponding RoleRestriction in 'this'.  If not, simply
	create one and copy RR over.  If there is one, then do
	the merge.
     */

    Map<Role*, RoleRestriction*>::iterator this_rr;
    for (Map<Role*, RoleRestriction*>::iterator rr = 
	u-> _told_role_restrictions.begin(); 
	rr != u->_told_role_restrictions.end(); rr++) {

        this_rr = _told_role_restrictions.find((*rr).first);
	if (this_rr == _told_role_restrictions.end()) {
	/*
	    This role does not appear in the RR list of 'this' ... we need
	    to insert a new RR then.  Need a copy.  *(*rr).second looks
	    complicated, but it's not.  rr is an iterator, *rr is thus 
	    the pair <Role*,RoleRestriction*>.  Its second is a 
	    RoleRestriction*.  De-referencing it is the RoleRestriction.
	 */

    /*
	Trying to avoid object creation by directly inserting the pointer.
	This should work all the time because whenever a concept or ind is
	defined like this:

	(def-ind i (and RR1 RR2 ...))

	each RR is freshly created and is not used elsewhere
     */
#if 0
	    Role* role = (*rr).first;
	    RoleRestriction* new_rr = new RoleRestriction(*(*rr).second);
	    _told_role_restrictions.insert(pair<Role*, RoleRestriction*>(
		role, new_rr));
#else
	    _told_role_restrictions.insert(*rr);
#endif
	} else 
	    merge_role_restriction((*this_rr).second, (*rr).second);
    }

    /*
	Third, merge tests.  This is relatively easy... since there is 
	no need to do any subsumption calculation.
     */

    copy(u->_tests.begin(), u->_tests.end(), inserter(_tests));

    /*
	Also, merge the one_ofs.  One-of is merged by finding the 
	intersection, unless one is empty.  
     */

    if (!_contradict_one_ofs) {
	const Set<Individual*>& one_ofs = u->_told_one_ofs;
	const Set<Individual*>& told = _told_one_ofs;
	if (!one_ofs.empty()) {
	    if (told.empty())
		_told_one_ofs = one_ofs;
	    else {
		Set<Individual*> intersection;
		set_intersection(one_ofs.begin(), one_ofs.end(), 
		    told.begin(), told.end(), 
		    inserter(intersection), individual_comp);
		_told_one_ofs = intersection;
		if (intersection.empty())
		    _contradict_one_ofs = TRUE;
	    }
	}
    }
}


void Concept::canon_parents()
/*
    If there is any unclassifed concept in the parent list, expand
    it.  Although this is a recursive procedure (merge_parents is),
    it should not be heavy as all the parent concepts should
    themselves be canonicalized before adding.  Therefore, at most,
    there will be

       (and C1 (and C2 C3 C4))

    but not

       (and C1 (and C2 (and C3 C4)))
 */
{
    Set<Concept*> tmp = parents();

    for (Set<Concept*>::iterator p = tmp.begin(); p != tmp.end(); p++)
	if ((*p)->unclassified()) {
	    merge_parents(*p);
	/*
	    _parents.erase(remove(_parents.begin(), _parents.end(), *p),
		_parents.end());
	 */
	    _parents.erase(*p);
	}
}


void Concept::canon_one_ofs()
/*
    (def-con C (one-of "a" "b" "c"))

    For the above concept, we need to infer the parent string.
    This can be done by adding all the direct parents of each
    one-of individual into the parent list.
    _________________________________________________________________

    Correction, the above is not a good idea.   Consider two 
    individuals (one-of i1 i2) where i1 is a C1 and i2 is a C2.
    What if C1 and C2 are disjoint primitive concepts of P?
    Then we have conflict.  I think it is more correct to put 
    common ancestor there.
 */
{
#if 0
    const Set<Individual*>& tmp = one_ofs();
    for (Set<Individual*>::const_iterator i = tmp.begin();
	i != tmp.end(); i++) {
	const Set<Concept*>& parents = (*i)->parents();
	for (Set<Concept*>::const_iterator p = parents.begin();
	    p != parents.end(); p++)
	    merge_parents(*p);
    }
#else
    const Set<Individual*>& oo = one_ofs();
    if (oo.size()) {
	List<Individual*> tmp;
	copy(oo.begin(), oo.end(), back_inserter(tmp));

	Concept* common_ancestor(const List<Individual*>&);
	Concept* A = common_ancestor(tmp);
	merge_parents(A);
    }
#endif
}


void Concept::canon()
/*
    Expand the concept to a canonical form.  After canon(),
     
      . there should be no unnamed concept in the parent list.
 */
{
    canon_parents();
    canon_one_ofs();
}


void Concept::normalize_parents()
/*
    Normalize the parent list.  The following procedures are performed
    here:

      . keep only the direct parents.  Eliminate all concepts that are 
	indirect ancestors.
 */
{
    /*
        The algorithm to keep the direct parents is unsophisticated, simply 
	check each parent P, if P subsumes any concept C in the parents list, 
	erase P.
     */
    
    Set<Concept*> tmp = parents();
    for (Set<Concept*>::iterator P = tmp.begin(); P != tmp.end(); P++)
        for (Set<Concept*>::iterator C = _parents.begin();
            C != _parents.end(); C++) {
            /*
                So, if P subsumes any C in the parent list, P is 
                obviously an ancester of 'this'.  So, remove it 
                and break (test another P).

                **P is used because *P is a Concept*, dereference
                that is a Concept.
             */
            if (**C < **P) {
                remove_parent(*P);
                break;
            }
	}
}


void Concept::normalize_children()
/*
    Normalize the children list.  Given a list of concepts, keep only 
    the direct children.  Eliminate all concepts that are indirect
    descendants.
 */
{
    /*
        The algorithm is unsophisticated, simply check each child P,
        if P is subsumed by any concept C in the children list, erase P.
     */

    Set<Concept*> tmp = children();
    for (Set<Concept*>::iterator P = tmp.begin(); P != tmp.end(); P++)
        for (Set<Concept*>::iterator C = _children.begin();
            C != _children.end(); C++) 
            /*
                So, if P is subsumed by any C in the parent list, P is 
                obviously an indirect descendant of 'this'.  So, remove 
                it and break (test another P).

                **P is used because *P is a Concept*, dereference
                that is a Concept.
             */
            if (**P < **C) {
                _children.erase(*P);
                break;
            }
        
}


bool Concept::contradict()
/*
    Evaluate whether this concept contradicts itself.  The following
    inspections are performed:
     
      . contradicting parents
      . contradicting role restrictions
 */
{
#if 0
    if (_contradict_one_ofs) cout << "DEBUG contradict one_ofs\n";
    if (contradict_parents()) cout << "DEBUG contradict parents\n";
    if (contradict_role_restrictions()) cout << "DEBUG contradict RR " << *this << endl;
#endif
    _null = _contradict_one_ofs ||
	    contradict_parents() || 
	    contradict_role_restrictions();

    return _null;
}


bool contradict_concepts(const Set<Concept*>& anc)
{
    /*
        The algorithm to check for disjoint parents is simple:
        for each parent P in the parent list, if P is a disjoint
        primitive concept having group G, then check to see if
        the rest of the parent list has a concept in the same
        group G.

	Alert: The above algorithm is inadequate.  consider:

	      HostThing          The above algorithm can detect
	       / * \             conflict in (and Int Float) as
	      /     \            well as (and Number String). But
	   Number   String       not (and Int String) or 
	     /*\                 (and Float String)
	    /   \
	  Int  Float

	The above algorithm should be extended slightly by testing the
	set of ancestors instead of the set of parents.
     */

    for (Set<Concept*>::const_iterator P = anc.begin(); P != anc.end(); P++)
        if ((*P)->type() == Object::disjoint_primitive_concept)
            for (Set<Concept*>::const_iterator Q = P; ++Q != anc.end(); )
                if ((*Q)->type() == Object::disjoint_primitive_concept) {
		    DisjointPrimitiveConcept* p = (DisjointPrimitiveConcept*)*P;
		    DisjointPrimitiveConcept* q = (DisjointPrimitiveConcept*)*Q;
		    if (p->same_group(q))
			return TRUE;
		}

    return FALSE;
}


bool Concept::contradict_parents()
/*
    See if the (normalized) parent list has contradiction.  It is 
    considered to be contradicting if the following happens:

      . some ancestors are from the same disjoint group
 */
{
    return contradict_concepts(ancestors());
}


bool Concept::contradict_role_restrictions()
/*
    The algorithm to check for contradiction in role restrictions call
    for a few other helper functions (defined in class RoleRestriction):

      . contradict_nr (number restriction)
      . contradict_vr (value restriction)
      . contradict_fillers

    Basically, if any one of the role restrictions has a contradiction
    in any of the above attributes, return TRUE.
 */
{
    for (Map<Role*, RoleRestriction*>::iterator i = 
	_told_role_restrictions.begin(); 
	i != _told_role_restrictions.end(); i++)
	if ((*i).second->contradict()) {
	    return TRUE;
	}

    return FALSE;
}


bool ClassicIndividual::contradict()
{
#if 0
    if (contradict_parents()) cout << "DEBUG contradict parents\n";
    if (contradict_tests()) cout << "DEBUG contradict tests\n";
    if (contradict_role_restrictions()) cout << "DEBUG contradict RR " << *this << endl, print();
#endif
    return contradict_parents() || contradict_role_restrictions() ||
	   contradict_tests();
}


bool ClassicIndividual::contradict_parents()
{
    return contradict_concepts(ancestors());
}


bool ClassicIndividual::contradict_role_restrictions()
{
    Set<RoleRestriction*> rr = role_restrictions();
    for (Set<RoleRestriction*>::iterator i = rr.begin(); i != rr.end(); i++) {
	if ((*i)->contradict())
	    return TRUE;
    }

    return FALSE;
}


bool ClassicIndividual::contradict_tests()
/*
    Check if this individual has some contradicting tests.  The tests
    are taking from each of the parents.
 */
{
    /*
	First, gather a set of tests.
     */

    Set<Test*> tests;
    for (Set<Concept*>::iterator P = _parents.begin(); 
	P != _parents.end(); P++) {
	Set<Test*> tmp = (*P)->tests();
	// tests.insert(tmp.begin(), tmp.end());
	copy(tmp.begin(), tmp.end(), inserter(tests));
    }

    /*
	If any one test returns NO, then there is contradiction.
     */

    for (Set<Test*>::iterator i = tests.begin(); i != tests.end(); i++)
	if ((**i)(this) == NO)
	    return TRUE;

    return FALSE;
}


void ClassicIndividual::normalize()
{
    normalize_parents();
}


void ClassicIndividual::normalize_parents()
/*
    see Concept::normalize_parents() above for doc.
 */
{
    Set<Concept*> tmp = parents();
    for (Set<Concept*>::iterator P = tmp.begin(); P != tmp.end(); P++)
        for (Set<Concept*>::iterator C = _parents.begin();
            C != _parents.end(); C++) 
            if (**C < **P) {
                remove_parent(*P);
                break;
            }
}
