#include "concept.h"
#include "individual.h"
#include "role.h"
#include "procedure.h"
#include "func.h"

Id ObjectBase::insert(Object* object)
/*
    Register an unnamed object into the knowledge base.  Return a
    non-negative integer id.
 */
{
#if defined(Debug)
cout << "ObjectBase::insert(Object*)" << endl;
#endif
    Id id = world.size();

    world.push_back(object);

#if defined(Debug)
cout << "ObjectBase::insert(Object*) returns" << endl;
#endif
    return id;
}


bool ObjectBase::insert(const String& name, Id id)
/*
    Insert <name,id> into the world map.  Make sure there is no duplicates.
 */
{
    if (exists(name)) {
	cout << "duplicate \"" << name << "\" in the knowledge base." << endl;
	find(name)->print();
	return FALSE;
    } else {
	world_map.insert(pair<const String, Id>(name, id));
	return TRUE;
    }
}


bool ObjectBase::erase(Id id)
{
    if (id >= 0 && id < world.size()) {
	world[id] = 0;

    /*
	No need to worry about world map because dtors of the corresponding
	subclasses that make use of it will erase the references.

	No need to call delete on world[id] either, because it is someone
	calling delete on it that triggers this function (~Object() calls
	this method).
     */
	return TRUE;
    } else 
	return FALSE;
}


bool ObjectBase::erase(const String& name)
/*
    Erase name from the world map.
 */
{
    return world_map.erase(name) == 1;
}


Object* ObjectBase::find(const String& name)
/*
    Look into the world_map to find a certain name and return a ptr to it.
    Return 0 if not found.
 */
{
    Map<String, Id>::iterator i;

    i = world_map.find(name);
    if (i == world_map.end()) 
	return 0;
    else
	return world[(*i).second];
}


bool ObjectBase::exists(const String& name)
{
    return find(name) != 0;
}


ObjectBase::ObjectBase()
/*
    Init the whole knowledge base.
 */
{
#ifdef Debug
cout << "ObjectBase::ObjectBase()" << endl;
#endif
    /*
	The taxonmoy is not useful until Thing, ClassicThing and HostThing
	have been setup.  _init is used to indicate this.
     */

    _init = FALSE;

    thing = define_primitive("thing", 0);
    assert(thing);

    classic_thing = define_disjoint_primitive("classic-thing", Thing, "thing");
    assert(classic_thing);

    host_thing = define_disjoint_primitive("host-thing", Thing, "thing");
    assert(host_thing);

    _init = TRUE;

    number_thing = define_disjoint_primitive("number", HostThing, "type");
    assert(number_thing);

    string_thing = define_disjoint_primitive("string", HostThing, "type");
    assert(string_thing);

    integer = define_disjoint_primitive("integer", NumberThing, "type");
    assert(integer);

    floating = define_disjoint_primitive("float", NumberThing, "type");
    assert(floating);

#ifdef Debug
cout << "ObjectBase::ObjectBase() returns" << endl;
#endif
}


ObjectBase::~ObjectBase()
/*
    Call delete on each element of world.
 */
{
    for (int i = 0; i < world.size(); i++)
	if (world[i])
	    delete world[i];
}


Object::Object()
{
#ifdef Debug
cout << "Object::Object()" << endl;
#endif
    /*
	It's hard to decide whether an unnamed object should be inserted
	to the object base.

	I think it should not, the problem is all objects registered in 
	the object base will be deleted by the ObjectBase destructor.
	Now, if there are temporary objects like this:

	    f() {
		RoleRestriction x;
	    }

	Then x will be destroyed twice, once when f() returns, and once
	when the program ends.

	turns out the above situation has been taken care of in ~Object,
	but the problem is this will cause the Object id to increase 
	quite uncontrollably.  E.g., every instantiation of a new ind will
	cause the id to increase 200 because of a lot of temporary objects
	during classification and normalization (most notably, the use of
	RoleRestriction).
     */
#if 1
    _id = object_base.insert(this);
#else
    _id = -1;
#endif
#ifdef Debug
cout << "Object::Object() returns" << endl;
#endif
}


Object::Object(const String& name)
{
#ifdef Debug
cout << "Object::Object(const String& name)" << endl;
#endif
    assert(!object_base.exists(name));
    _id = object_base.insert(this);
    object_base.insert(name, _id);
#ifdef Debug
cout << "Object::Object(const String& name) returns" << endl;
#endif
}   


Object::~Object()
{
    if (id() != -1) {
        // cout << "Destroying Object " << id() << " from the world\n";
        if (!object_base.erase(id()))
            cout << id() << " is not in the world." << endl;
    }
}


Object* object(const String& name) 
{
    return OB.find(name);
}


void Object::print(ostream& out) const
/*
    Object::print and pretty_print are trickier than other prints.  We
    cannot rely on the "virtual-ity" of themselves because when the 
    system parses any ID, such as "name", it can be a concept or an
    individual or a role, etc.  It simply returns an Object* without
    knowing the actual type.  This may be curable but not sure.

    The quick and dirty solution is to do "switch" statement right 
    here.  Quite against the principle of OO, but OO can get to a very
    complicated point where the programmer simply just cannot follow
    whether a pointer has an associated virtual table or not, as in 
    this case.
 */
{
#if 0
    switch (type()) {

    case object:
    case function:
        out << "Id: " << id() << endl; 
        out << "Type: " << type_name() << endl;
	break;

    case concept:
        ((Concept*)this)->print(out); break; 

    case primitive_concept:
        ((PrimitiveConcept*)this)->print(out); break; 

    case disjoint_primitive_concept:
        ((DisjointPrimitiveConcept*)this)->print(out); break; 

    case individual:
	switch (((Individual*)this)->subtype()) {

	case Individual::integer_individual:
	    ((IntegerIndividual*)this)->print(out); break; 

	case Individual::float_individual:
	    ((FloatIndividual*)this)->print(out); break; 

	case Individual::string_individual:
	    ((StringIndividual*)this)->print(out); break; 

	case Individual::classic_individual:
	    ((ClassicIndividual*)this)->print(out); break; 

	default:
	    cout << "Unknown object.\n";
	}
	break;

    case role:
        ((Role*)this)->print(out); break; 

    case rule:
        ((Rule*)this)->print(out); break; 

    case role_restriction:
        ((RoleRestriction*)this)->print(out); break; 

    case procedure:
        ((ProceduralAttachment*)this)->print(out); break; 

    case test:
        ((Test*)this)->print(out); break; 

    case computed_concept:
        ((ComputedConcept*)this)->print(out); break; 

    case computed_fillers:
        ((ComputedFillers*)this)->print(out); break; 

    default:
	out << "Unknown object.\n";
    }
#else
        out << "Id: ";
	if (id() == -1) 
	    cout << "(unnamed)" << endl;
	else
	    cout << id() << endl; 
        out << "Type: " << type_name() << endl;
#endif
}


void Object::pretty_print(ostream& out) const
{
#if 1

out << "Here " << type() << endl;
    switch (type()) {

    case object:
    case function:
        out << "Object: " << id(); break;

    case concept:
        ((Concept*)this)->pretty_print(out); break; 

    case primitive_concept:
        ((PrimitiveConcept*)this)->pretty_print(out); break; 

    case disjoint_primitive_concept:
        ((DisjointPrimitiveConcept*)this)->pretty_print(out); break; 

    case individual:

	switch (((Individual*)this)->subtype()) {

	case Individual::integer_individual:
	    ((IntegerIndividual*)this)->pretty_print(out); break; 

	case Individual::float_individual:
	    ((FloatIndividual*)this)->pretty_print(out); break; 

	case Individual::string_individual:
	    ((StringIndividual*)this)->pretty_print(out); break; 

	case Individual::classic_individual:
	    ((ClassicIndividual*)this)->pretty_print(out); break; 

	default:
	    cout << "Unknown object.\n";
	}
	break;

    case role:
        ((Role*)this)->pretty_print(out); break; 

    case rule:
        ((Rule*)this)->pretty_print(out); break; 

    case role_restriction:
        ((RoleRestriction*)this)->pretty_print(out); break; 

    case procedure:
        ((ProceduralAttachment*)this)->pretty_print(out); break; 

    case test:
        ((Test*)this)->pretty_print(out); break; 

    case computed_concept:
        ((ComputedConcept*)this)->pretty_print(out); break; 

    case computed_fillers:
        ((ComputedFillers*)this)->pretty_print(out); break; 

    default:
	out << "Unknown object.\n";
    }
#else
        out << "Object: " << id();
#endif
}

