%{
#include <stdio.h>
#include <string.h>
#include "concept.h"
#include "individual.h"
#include "procedure.h"
#include "y.tab.h"

int keyword(const char* s);

#define MAX_INCLUDE_DEPTH 20

int lineno[MAX_INCLUDE_DEPTH] = {1};

YY_BUFFER_STATE include_stack[MAX_INCLUDE_DEPTH];
int include_stack_ptr = 0;
bool eval = FALSE;

extern void load(const char*);
extern void init_load(const char*);
extern int yyparse();
extern void yyrestart(FILE*);

%}

ws	[ \t\n]+
real	[-+]?([0-9]+\.[0-9]*)|(\.[0-9]+)
unsigned	\+?[0-9]+
integer	[-+]?[0-9]+
comment	;.*$
id	[a-zA-Z][-_!@#$%^&*+=:<>/\|'~`.?a-zA-Z0-9]*
string	\"[^\"\n]*[\"\n]
nl	\n

%%

{ws}		{ if (yytext[0] == '\n') 
                      lineno[include_stack_ptr]++; 
		}

{comment}	;

{real}	    { sscanf(yytext, "%f", &yylval.real); return REAL; }

{unsigned}  { sscanf(yytext, "%u", &yylval.integer); return UNSIGNED; }

{integer}   { sscanf(yytext, "%d", &yylval.integer); return INTEGER; }

{string}    { 
                if (yytext[strlen(yytext)-1] != '"')
                    cerr << "Unterminated string\n";
		else
		    yytext[strlen(yytext)-1] = 0;
  	        yylval.string = strdup(&yytext[1]); 
	        return STRING;
	    }

\<=	    {
                yylval.cmd = OP_LE;
		return yylval.cmd;
	    }

>=	    {
                yylval.cmd = OP_GE;
		return yylval.cmd;
	    }

{id}	    { 
		if ((yylval.cmd = keyword(yytext)) != 0)
		    return yylval.cmd;
		else {
		    yylval.string = strdup(yytext); 
		    return ID;
		}
	    }

.	    { return yytext[0]; }

<<EOF>>     {
	    /*

		This routine is very kludgy due to my lack of time to
		understanding the bison parser fully.  The major
		problem is with recursive calling of yyparse.  The
		evaluate() routine in expression.C calls yyparse
		recursively to parse an expression.  When the
		expression ends (it's stored in a tmp file), we
		arrive at this point.  It seems like if we do not
		call yyterminate(), the parser will go on to wait for
		user input instead of returning to evaluate().  If we
		call yyterminate(), then the parser will quit
		altogether after evaluate() returns.  So, we add an
		eval flag and put a loop in main() to call
		interpreter() repeatedly until stdin is eof.  Also,
		the prompt "Classic>" is also a kludge.  For some
		unknown reason, the prompt is printed twice if
		evaluate() is called.  I put \r in front of Classic>
		to make the prompt appear together!  The user can
		tell this in a typescript, though.

	     */

	        if (include_stack_ptr == 0)
	 	    yyterminate();
	        else {
		    include_stack_ptr--;
		    if (yyin != stdin)
			fclose(yyin);
		    yy_delete_buffer(YY_CURRENT_BUFFER);
		    yy_switch_to_buffer(include_stack[include_stack_ptr]);
		    if (eval) {
			eval = FALSE;
			yyterminate();
		    } else
			if (include_stack_ptr == 0 && yyin == stdin)
			    cout << "\rClassic> " << flush;
		}
	    }

%%

static struct keyword_t {
    char* name;
    int token;
} keywords[] = {
    "define-concept", CMD_DEF_CON,
    "define-con", CMD_DEF_CON,
    "def-concept", CMD_DEF_CON,
    "def-con", CMD_DEF_CON,

    "define-primitive", CMD_DEF_PRIM,
    "define-prim", CMD_DEF_PRIM,
    "def-primitive", CMD_DEF_PRIM,
    "def-prim", CMD_DEF_PRIM,

    "define-disjoint-primitive", CMD_DEF_DISJ_PRIM,
    "define-disjoint-prim", CMD_DEF_DISJ_PRIM,
    "define-disj-primitive", CMD_DEF_DISJ_PRIM,
    "def-disjoint-primitive", CMD_DEF_DISJ_PRIM,
    "define-disj-prim", CMD_DEF_DISJ_PRIM,
    "def-disj-primitive", CMD_DEF_DISJ_PRIM,
    "def-disjoint-prim", CMD_DEF_DISJ_PRIM,
    "def-disj-prim", CMD_DEF_DISJ_PRIM,

    "define-individual", CMD_DEF_IND,
    "define-ind", CMD_DEF_IND,
    "def-individual", CMD_DEF_IND,
    "def-ind", CMD_DEF_IND,

    "define-role", CMD_DEF_ROLE,
    "def-role", CMD_DEF_ROLE,

    "define-attribute", CMD_DEF_ATTR,
    "define-attr", CMD_DEF_ATTR,
    "def-attribute", CMD_DEF_ATTR,
    "def-attr", CMD_DEF_ATTR,

    "define-rule", CMD_DEF_RULE,
    "def-rule", CMD_DEF_RULE,

    "and", CMD_AND,
    "all", CMD_ALL,
    "fills", CMD_FILLS,
    "fill", CMD_FILLS,
    "one-of", CMD_ONE_OF,
    "at-least", CMD_AT_LEAST,
    "at-most", CMD_AT_MOST,
    "exactly", CMD_EXACTLY,
    "between", CMD_BETWEEN,
    "test-c", CMD_TEST_C,

    "computed-concept", CMD_COMPUTED_CONCEPT,
    "computed-con", CMD_COMPUTED_CONCEPT,

    "computed-fillers", CMD_COMPUTED_FILLERS,
    "computed-filler", CMD_COMPUTED_FILLERS,

    "individual-close-role", CMD_CLOSE_ROLE,
    "ind-close-role", CMD_CLOSE_ROLE,

    "individual-attribute-filler", CMD_IND_ATTR_FILLER,
    "individual-attr-filler", CMD_IND_ATTR_FILLER,
    "ind-attribute-filler", CMD_IND_ATTR_FILLER,
    "ind-attr-filler", CMD_IND_ATTR_FILLER,

    "individual-role-fillers", CMD_IND_ROLE_FILLERS,
    "ind-role-fillers", CMD_IND_ROLE_FILLERS,
    "individual-role-filler", CMD_IND_ROLE_FILLERS,
    "ind-role-filler", CMD_IND_ROLE_FILLERS,

    "individual-parents", CMD_IND_PARENTS,
    "ind-parents", CMD_IND_PARENTS,
    "concept-parents", CMD_CON_PARENTS,
    "con-parents", CMD_CON_PARENTS,

    "nth", CMD_NTH,

    "load", CMD_LOAD,
    "trace", CMD_TRACE,
    "untrace", CMD_UNTRACE,
    "rule", OP_RULE,
    "sub", OP_SUB,
    "subsumption", OP_SUB,
    "class", OP_CLASS,
    "classification", OP_CLASS,
    "test", OP_TEST,
    "on", OP_ON,
    "off", OP_OFF,

    "echo", CMD_ECHO
};

#define keyword_n (sizeof(keywords)/sizeof(keyword_t))

int keyword(const char* s)
{
    for (int i = 0; i < keyword_n; i++) {
	if (!strcmp(s, keywords[i].name))
	    return keywords[i].token;
    }
    return 0;
}


/*
    There are 3 different kinds of loading in the system.  The first is
    the (load file) command executed in the interpreter or in a loading
    file.  The second is the file arguments supplied in argv.  This is
    not the same as the first because in this case, the interpreter has
    not started running yet.  The last is the evalutate command in 
    expression.C.  It creates a temp file and loads it.
 */

void load(const char* s)
{
    if (include_stack_ptr >= MAX_INCLUDE_DEPTH) {
	cerr << "Included files nested too deeply.\n";
	return;
    }

    include_stack[include_stack_ptr++] = YY_CURRENT_BUFFER;
    yyin = fopen(s, "r");
    if (!yyin) {
	cerr << "Cannot open " << s << endl;
	include_stack_ptr--;
	return;
    }

    lineno[include_stack_ptr] = 1;
    yy_switch_to_buffer(yy_create_buffer(yyin, YY_BUF_SIZE));
}


void eval_load(const char* s)
{
    eval = TRUE;
    if (include_stack_ptr >= MAX_INCLUDE_DEPTH) {
	cerr << "Included files nested too deeply.\n";
	return;
    }

    include_stack[include_stack_ptr++] = YY_CURRENT_BUFFER;
    yyin = fopen(s, "r");
    if (!yyin) {
	cerr << "Cannot open " << s << endl;
	include_stack_ptr--;
	return;
    }

    lineno[include_stack_ptr] = 1;
    yy_switch_to_buffer(yy_create_buffer(yyin, YY_BUF_SIZE));

    /*
        Must not call yyrestart here, otherwise, the buffer for the 
	original file will be lost.
	
	// yyrestart(yyin);
    */

    yyparse();
}


void init_load(const char* s)
{
    /*
	No need to check max depth because this is from the init.
     */

    yyin = fopen(s, "r");
    if (!yyin) {
	cerr << "Cannot open " << s << endl;
	return;
    }

    cout << "Loading " << s << endl;
    lineno[include_stack_ptr] = 1;
    yyparse();
}
