%{
#define YYDEBUG 1

/*
    I don't know if this is a bug or what.  #include "trace.h" must be
    put in this position (first of all #include's).  If it is below
    iostream.h and stdlib.h then there will be a linking error.  

    Possibly some name clashes or typedef clashes in stdlib.h or
    iostream.h.
 */

#include "trace.h"
#include <iostream.h>
#include <stdlib.h>
#include "individual.h"
#include "concept.h"
#include "expression.h"
#include "classify.h"

extern void print_object(char*);
extern void load(const char*);
extern int yyerror(char* s);
extern int yylex();
extern int lineno[];
extern int include_stack_ptr;
extern bool eval;

/*
    When there are nested "and" statements like so:

    (and C1 
	 C2 
	 (and C3 ...))

    Then, we need to keep more than one list of concepts in 
    concept_expr_list.  The first list will consist of 

	C1 C2 <tmp>

    where <tmp> is the second list 

	C3 ...

    The more nested "and" statements, the more tmp lists needed.
    _____________________________________________________________

    Second thought.  Observe that

    (and C1 
	 C2 
	 (and C3 ...))

    is equivalent to 

    (and C1 C2 C3 ...)

    It is quite possible to expand the inner and in Yacc.  This 
    can be done by simply using a single list for all inner ands.
    The only problem is how do we know which and is the outer 
    most, so that we only return an "and" concept for the outer
    most and.  This can be done by a simple counter.
    _____________________________________________________________

    Third thought.  Wrong wrong wrong.  Consider this:

    (and C1 (all R (and X1 ...)))

    This is not equivalent to

    (and C1 X1 ...)
*/

#define MAX_AND 20
static List<Concept*> and_expr[MAX_AND];
static int and_n = 0;

static List<Individual*> fillers;
static List<Individual*> role_fillers;
static List<Concept*> concepts;
static List<Role*> roles;
static ObjectVector objects;

%}

%union {
    int integer;
    float real;
    char* string;
    int cmd;
    Concept* concept;
    Individual* individual;
    Object* object;
    ComputedConcept* computed_concept;
    ComputedFillers* computed_fillers;
    List<Individual*>* role_fillers;
}

%token <integer> 	INTEGER UNSIGNED
%token <real> 	REAL
%token <string> STRING ID
%token <cmd>	COMMAND
%token <cmd>    CMD_DEF_CON
%token <cmd>    CMD_DEF_PRIM
%token <cmd>    CMD_DEF_DISJ_PRIM
%token <cmd>    CMD_DEF_IND
%token <cmd>	CMD_DEF_ROLE
%token <cmd> 	CMD_DEF_ATTR
%token <cmd> 	CMD_DEF_RULE
%token <cmd> 	CMD_AND
%token <cmd> 	CMD_ALL
%token <cmd> 	CMD_FILLS
%token <cmd> 	CMD_ONE_OF
%token <cmd>    CMD_AT_LEAST
%token <cmd>    CMD_AT_MOST
%token <cmd>    CMD_EXACTLY
%token <cmd>    CMD_BETWEEN
%token <cmd>    CMD_TEST_C
%token <cmd> 	CMD_COMPUTED_CONCEPT
%token <cmd> 	CMD_COMPUTED_FILLERS
%token <cmd> 	CMD_CLOSE_ROLE
%token <cmd> 	CMD_IND_ATTR_FILLER
%token <cmd> 	CMD_IND_ROLE_FILLERS
%token <cmd> 	CMD_IND_PARENTS
%token <cmd> 	CMD_CON_PARENTS
%token <cmd> 	CMD_LOAD
%token <cmd> 	CMD_TRACE
%token <cmd> 	CMD_UNTRACE
%token <cmd> 	CMD_ECHO
%token <cmd> 	CMD_NTH
%token <cmd> 	OP_CLASS
%token <cmd> 	OP_SUB
%token <cmd> 	OP_RULE
%token <cmd> 	OP_TEST
%token <cmd> 	OP_ON
%token <cmd> 	OP_OFF
%token <cmd> 	OP_LE
%token <cmd> 	OP_GE

%type <concept> concept_expr
%type <concept> named_concept
%type <concept> unnamed_concept
%type <individual> host_individual
%type <individual> named_individual
%type <individual> unnamed_individual
%type <individual> individual
%type <role_fillers> role_fillers
%type <object> object
%type <computed_concept> computed_concept
%type <computed_fillers> computed_fillers
%type <string> disj_type

%start lines

%%

lines:  /* nothing */
	| lines line
        {
	    extern FILE* yyin;
	    if (yyin == stdin)
		cout << "\rClassic> " << flush;
	}
	| error ')'
	{ 
	    extern FILE* yyin;
	    extern bool eval;
	    if (yyin == stdin)
		cout << "\rClassic> " << flush;
	}
	;

line:   command
	;

command:  inspect
        | trace
        | load_file
	| define_role
	| define_attr
	| define_concept
	| define_primitive
	| define_disjoint_primitive
	| define_individual
	| define_rule
	| concept_list
	| close_role
	| echo
	;

inspect: ID
        {
	    print_object($1); free($1);
	}
	| host_individual
        { 
	    if (!eval) 
		cout << *($1) << endl;
	}
	| unnamed_individual
        {
	    if (!($1))
		cerr << "Null individual.\n";
	    else
		if (!eval) 
		    cout << *($1) << endl;
	}
	| unnamed_concept
	{
	    if (!($1))
		cerr << "Null concept.\n";
	    else
		if (!eval)
		    cout << *($1) << endl;
	}
	| role_fillers
	{
	    if ($1) {
		cout << "[ ";
		for (List<Individual*>::iterator i = $1->begin();
		    i != $1->end(); i++)
		    cout << **i << " ";
		cout << "]\n";
	    }
	}
	| '(' '<' ID named_concept ')'
	{
	    if (concept($3))
		cout << (*concept($3) < *$4? "yes": "no") << endl;
	    else
		if (individual($3))
		    cout << (*individual($3) < *$4? "yes": "no") << endl;
		else
		    cerr << $3 << " is not a concept or an individual.\n";
	}
	| '(' '>' named_concept ID ')'
	{
	    if (concept($4))
		cout << (*concept($4) < *$3? "yes": "no") << endl;
	    else
		if (individual($4))
		    cout << (*individual($4) < *$3? "yes": "no") << endl;
		else
		    cerr << $4 << " is not a concept or an individual.\n";
	}
	| '(' OP_LE ID named_concept ')'
	{
	    if (concept($3))
		cout << (*concept($3) <= *$4? "yes": "no") << endl;
	    else
		cerr << $3 << " is not a concept.\n";
	}
	| '(' OP_GE named_concept ID ')'
	{
	    if (concept($4))
		cout << (*concept($4) <= *$3? "yes": "no") << endl;
	    else
		cerr << $4 << " is not a concept.\n";
	}
	;

concept_list: '(' CMD_IND_PARENTS named_individual ')'
	{
	    if ($3) {
		const Set<Concept*>& P = $3->parents();
		cout << "[ ";
		for (Set<Concept*>::const_iterator i = P.begin();
		    i != P.end(); i++)
		    cout << **i << " ";
		cout << "]\n";
	    }
	}
	| '(' CMD_CON_PARENTS named_concept ')'
	{
	    if ($3) {
		const Set<Concept*>& P = $3->parents();
		cout << "[ ";
		for (Set<Concept*>::const_iterator i = P.begin();
		    i != P.end(); i++)
		    cout << **i << " ";
		cout << "]\n";
	    }
	}
	;

named_individual: ID
	{
	    $$ = individual($1);
	    if (!$$)
		cerr << $1 << " is not an individual.\n";
	    free($1);
	}
	;

host_individual: INTEGER
	{   
	    $$ = ind($1);
	}
	| UNSIGNED
	{
	    $$ = ind($1);
	}
	| REAL
	{
	    $$ = ind($1);
	}
	| STRING
	{
	    $$ = ind($1); free($1);
	}
	;

unnamed_individual: '(' CMD_IND_ATTR_FILLER individual ID ')'
       	{
	    if ($3)
		$$ = ((ClassicIndividual*)$3)->filler($4); 
	    else
		$$ = 0;
	    free($4);
	}
	| '(' CMD_NTH role_fillers UNSIGNED ')'
	{
	    if ($3->size() > $4) {
		List<Individual*>::iterator i = $3->begin();
		for (int j = 0; j < $4; j++)
		    i++;
		$$ = *i;
	    } else
		$$ = 0;
	}
	;

individual: named_individual
	| host_individual
	| unnamed_individual
	;

individual_list: individual
	{
	    if ($1)
		fillers.push_back($1);
	}
	| individual_list individual
	{
	    if ($2)
		fillers.push_back($2);
	}
	;

role_fillers: '(' CMD_IND_ROLE_FILLERS individual ID ')'
       	{
	    if ($3) {
		role_fillers = ((ClassicIndividual*)$3)->fillers($4); 
		$$ = &role_fillers;
	    } else
		$$ = 0;
	    free($4);
	}
	;

role_list: ID
	{
	    Role* R = role($1);
	    if (R)
		roles.push_back(R);
	    free($1);
	}
	| role_list ID
	{
	    Role* R = role($2);
	    if (R)
		roles.push_back(R);
	    free($2);
	}
	;

load_file: '(' CMD_LOAD ID ')'
        {
	    cout << "Loading " << $3 << endl;
	    load($3);
	    free($3);
	    /*
		yyparse is restarted after loading, therefore the 
		"lines" rule will not remember to print a prompt.
		Has to be handled very differently, and in parse.l
	     */
	}
	| '(' CMD_LOAD STRING ')'
        {
	    cout << "Loading " << $3 << endl;
	    load($3);
	}
	;

trace: '(' CMD_TRACE OP_ON ')'
	{
	    trace.resume();
	}
	| '(' CMD_TRACE OP_OFF ')'
	{
	    trace.suspend();
	}
	| '(' CMD_TRACE OP_CLASS ')'
	{
	    trace.set(Trace::classification);
	}
	| '(' CMD_TRACE OP_SUB ')'
	{
	    trace.set(Trace::subsumption);
	}
	| '(' CMD_TRACE OP_RULE ')'
	{
	    trace.set(Trace::rule);
	}
	| '(' CMD_TRACE OP_TEST ')'
	{
	    trace.set(Trace::test);
	}
	| '(' CMD_UNTRACE OP_CLASS ')'
	{
	    trace.unset(Trace::classification);
	}
	| '(' CMD_UNTRACE OP_RULE ')'
	{
	    trace.unset(Trace::rule);
	}
	| '(' CMD_UNTRACE OP_TEST ')'
	{
	    trace.unset(Trace::test);
	}
	| '(' CMD_UNTRACE OP_SUB ')'
	{
	    trace.unset(Trace::subsumption);
	}
	;

define_role: '(' CMD_DEF_ROLE ID ')'
	{
	    define_role($3);
	    if (!eval)
		cout << $3 << endl;
	    free($3);
	}
	;

define_attr: '(' CMD_DEF_ATTR ID ')'
	{
	    define_attribute($3);
	    if (!eval)
		cout << $3 << endl;
	    free($3);
	}
	;

define_concept: '(' CMD_DEF_CON ID concept_expr ')'
	{
	    define_concept($3, $4);
	    if (!eval)
		cout << $3 << endl;
	    free($3);
	}
	;

define_primitive: '(' CMD_DEF_PRIM ID concept_expr ')'
	{
	    define_primitive($3, $4);
	    if (!eval)
		cout << $3 << endl;
	    free($3);
	}
	;

disj_type: ID
        {
	    $$ = $1;
	}
        | UNSIGNED
	{
	    $$ = (char*)malloc(10);
	    sprintf($$, "%d", $1);
	}
	;

define_disjoint_primitive: '(' CMD_DEF_DISJ_PRIM ID concept_expr disj_type ')'
	{
	    define_disjoint_primitive($3, $4, $5);
	    if (!eval)
		cout << $3 << endl;
	    free($3);
	    free($5);
	}
	;

define_individual: '(' CMD_DEF_IND ID concept_expr ')'
	{
	    define_individual($3, $4);
	    if (!eval)
		cout << $3 << endl;
	    free($3);
	}
	;

define_rule: '(' CMD_DEF_RULE ID named_concept concept_expr ')'
        {
	    define_rule($3, $4, $5);
	    if (!eval)
		cout << $3 << endl;
	    free($3);
	}
	| '(' CMD_DEF_RULE ID named_concept computed_concept ')'
	{
	    define_rule($3, $4, $5);
	    if (!eval)
		cout << $3 << endl;
	    free($3);
	}
	| '(' CMD_DEF_RULE ID named_concept computed_fillers ')'
	{
	    define_rule($3, $4, $5);
	    if (!eval)
		cout << $3 << endl;
	    free($3);
	}
	;

close_role: '(' CMD_CLOSE_ROLE named_individual role_list ')'
	{
	    ClassicIndividual* I = (ClassicIndividual*)$3;
	    if (I) {
		for (List<Role*>::iterator r = roles.begin(); 
		    r != roles.end(); ) {
		    RoleRestriction* RR = I->role_restriction(*r);
		    if (!RR) {
			cout << "Undefined role " << **r << endl;
			r++;
		    } else
			/*
			    Here is an optimization.  RR->close is called for
			    all roles except the last.  For the last role, 
			    I->close_role() is called.  The former function
			    does not call classify().  The latter does.  And
			    classify() only need be called once.
			 */
			if (++r == roles.end())
			    I->close_role(RR->role());
			else
			    I->close_role_aux(RR->role());
		}
	    }
	    while (roles.size())
		roles.pop_front();
	}
	;

echo: '(' CMD_ECHO host_individual ')'
	{
	    cout << *$3 << endl;
	}
	;

concept_expr: named_concept
        {
	    $$ = $1;
	}
	| unnamed_concept
	{
	    $$ = $1;
	}
	;

named_concept: ID
	{
	    $$ = concept($1);
	    if (!$$)
		cerr << $1 << " is not a concept.\n";
	    free($1);
	}
	;

unnamed_concept: '(' CMD_AND 
	{
	    and_n++;
	    assert(and_n <= MAX_AND);
	}
	concept_expr_list ')'
	{
	    assert(and_n);
	    and_n--;
	    $$ = and(and_expr[and_n]);
	    while (and_expr[and_n].size())
		and_expr[and_n].pop_front();
	}
	| '(' CMD_ALL ID concept_expr ')'
	{
	    $$ = all(role($3), $4); free($3);
	}
	| '(' CMD_AT_MOST UNSIGNED ID ')'
	{
	    $$ = at_most($3, role($4)); free($4);
	}
	| '(' CMD_AT_LEAST UNSIGNED ID ')'
	{
	    $$ = at_least($3, role($4)); free($4);
	}
	| '(' CMD_EXACTLY UNSIGNED ID ')'
	{
	    $$ = exactly($3, role($4)); free($4);
	}
	| '(' CMD_BETWEEN UNSIGNED UNSIGNED ID ')'
	{
	    $$ = between($3, $4, role($5)); free($5);
	}
	| '(' CMD_FILLS ID individual_list ')'
	{
	    $$ = fills(role($3), fillers);
	    while (fillers.size())
		fillers.pop_front();
	    free($3);
	}
	| '(' CMD_ONE_OF individual_list ')'
	{
	    $$ = one_of(fillers);
	    while (fillers.size())
		fillers.pop_front();
	}
	| '(' CMD_TEST_C ID object_list ')'
	{
	    $$ = test_c($3, objects);
	    while (objects.size())
		objects.pop_back();
	    free($3);
	}
	;

concept_expr_list: concept_expr
	{
	    if ($1) 
		and_expr[and_n-1].push_back($1);
	}
	| concept_expr_list concept_expr
	{
	    if ($2) 
		and_expr[and_n-1].push_back($2);
	}

computed_concept: '(' CMD_COMPUTED_CONCEPT ID object_list ')'
	{
	    $$ = computed_concept($3, objects);
	    while (objects.size())
		objects.pop_back();
	    free($3);
	}
	;

computed_fillers: '(' CMD_COMPUTED_FILLERS ID ID object_list ')'
	{
	    $$ = computed_fillers($3, role($4), objects);
	    while (objects.size())
		objects.pop_back();
	    free($3); free($4);
	}
	;

object: ID
	{
	    Object* obj = object($1);
	    if (obj)
		$$ = obj;
	    else
		cerr << "no such object\n";
	    free($1);
	}
	| host_individual
	{
	    $$ = $1;
	}
	;

object_list: /* empty list */
	| object_list object
	{
	    if ($2)
		objects.push_back($2);
	}
	;

%%

int yyerror(char* s)
{
    cerr << s << " line " << lineno[include_stack_ptr] << endl;
    return 0;
}
