#if !defined(PROCEDURE_H)
#define PROCEDURE_H

#include "role.h"
#include "concept.h"
#include "func.h"

class ProceduralAttachment: public Object {

/*
    class ProceduralAttachment is a general base class for the
    three main kinds of procedural attachments in C-Classic:

      . Test
      . ComputedConcept
      . ComputedFillers

    It has a name but is not registered into the world map.  The name
    is used to find out the which C function from the world map.  see
    class Function.
 */

public:

    virtual ~ProceduralAttachment() {}
    
    Type type() const { return procedure; }

    String type_name() const { return "procedure"; }

    String name() const { return _name; }

    Object* func() const { return _func; }

    const ObjectVector& argument() const { return _argument; }

    void print(ostream& out = cout) const;

    void pretty_print(ostream& out = cout) const;

    friend ostream& operator<<(ostream& out, const ProceduralAttachment& T) 
    {
	T.pretty_print(out);
	return out;
    }

    bool operator==(const ProceduralAttachment& x) const
    {
	return this == &x || 
	       ( _name == x._name &&
		 _argument == x._argument );
    }

protected:

    ProceduralAttachment(const String& name, const ObjectVector& arg): 
	_name(name), _argument(arg)
    {
	_func = object_base.find(name);
	assert(_func);
	assert(_func->type() == function);
    }

private:

    String _name;

    ObjectVector _argument;

    Object* _func;
};


class Test: public ProceduralAttachment {

/*
    class Test is used as a test concept.

    It has a name but is not registered into the world map.  The name
    is used to find out the TestFunction from the world map.  see
    class Function.
 */

    friend Concept* test_c(const String&, const ObjectVector&);
    friend Concept* test_c(char*, ...);

public:

    virtual ~Test() {}
    
    Type type() const { return test; }

    String type_name() const { return "test"; }

    TestFunction func() const { return _func; }

    bool host() const { return _host; }

    bool classic() const { return !_host; }

    TestResult execute(Individual* I);

    TestResult operator()(Individual* I) { return execute(I); }

protected:

    /*
	It is interesting to note that even though Function is a named
	object, Test is not.  The reason is, the same test function, say
	"some" can be used in more than one way, depending on what 
	arguments are being passed to it.  Therefore, tests are dis-
	tinguished by the test name as well as the arguments.  The name
	alone is not enough to distinguish tests.  As a result, Test is
	not a named object.
     */

    Test(const String& name, const ObjectVector& arg, bool host = FALSE): 
	ProceduralAttachment(name, arg), _host(host)
    {
	_func = ((Function<TestFunction>*)ProceduralAttachment::func())->func();
    }

private:

    TestFunction _func;  // most important feature, the procedural test.

    bool _host;
};


class ComputedConcept: public ProceduralAttachment {

/*
 */

    friend ComputedConcept* computed_concept(const String&, 
				const ObjectVector&);

    friend ComputedConcept* computed_concept(char*, ...);

    /*
	Interesting observation, if computed concept and computed
	fillers are both defined as computed, then ambiguity will
	happen:

	friend ComputedConcept* computed(char*, ...);
	friend ComputedFillers* computed(char*, Role*, ...);

	what if we have computed("abc", 0)?  To a human eye, it 
	looks like the former.  But the compiler will use the
	latter causing a core dump.
     */

public:

    virtual ~ComputedConcept() {}
    
    Type type() const { return computed_concept; }

    String type_name() const { return "computed concept"; }

    ComputedConceptFunction func() const { return _func; }

    Concept* execute(Individual* I) const;

    Concept* operator()(Individual* I) const { return execute(I); }

protected:

    ComputedConcept(const String& name, const ObjectVector& arg): 
	ProceduralAttachment(name, arg)
    {
	_func = ((Function<ComputedConceptFunction>*)
		    ProceduralAttachment::func())->func();
    }

private:

    ComputedConceptFunction _func;
};


class ComputedFillers: public ProceduralAttachment {

/*
 */

    friend ComputedFillers* computed_fillers(const String&, Role*, 
				const ObjectVector&);

    friend ComputedFillers* computed_fillers(char*, Role*, ...);

public:

    virtual ~ComputedFillers() {}
    
    Type type() const { return computed_fillers; }

    String type_name() const { return "computed fillers"; }

    ComputedFillersFunction func() const { return _func; }

    Role* role() const { return _role; }

    ObjectVector execute(Individual* I) const;

    ObjectVector operator()(Individual* I) const { return execute(I); }

protected:

    ComputedFillers(const String& name, Role* role, 
	const ObjectVector& arg): 
	ProceduralAttachment(name, arg)
    {
	_role = role;
	_func = ((Function<ComputedFillersFunction>*)
		    ProceduralAttachment::func())->func();
    }

private:

    ComputedFillersFunction _func;

    Role* _role;
};

#endif
