#include "trace.h"
#include "classify.h"
#include "role_restr.h"
#include "expression.h"
#include <limits.h>


RoleRestriction::RoleRestriction(const RoleRestriction& rr)
/*
    Copy Contructor for RoleRestriction.
 */
{
    _role = rr._role;
    _nr = rr._nr;
    _fillers = rr._fillers;
    _closed = rr._closed;

    /*
	When copying a RoleRestriction, make sure the value restriction (a
	concept) is copied into a newly allocated space.  See comments in
	ii.h class RoleRestriction.

	_vr = new Concept;
	if (rr._vr->classified) {
	    _vr->add_parent(rr._vr);
	    _vr->_null = rr._vr->_null;
	} else
	    _vr = rr._vr;

     */

    /*
	When allocating a new RR, we used to create a new concept and
	add a concept C1 to it as parent.  This is no longer in use 
	because we do not want to see such a beast: 

	    (and string)

	This will not be equivalent to string when using operator==
	unless we call normalize.  But if the above is not constructed
	by calling the and() function, normalize is not called.  

	So, we simply do 

	    _vr = rr._vr	// i.e., pointer assignment.

	the only catch is, we have to change merge_role_restriction(),
	so that when another concept C2 is added to it, we then 
	allocate space and do add_parents (possibly normalize).
     */

	_vr = rr._vr;
}


Concept* common_ancestor(const List<Individual*>& ind_list)
/*
    Find the concept that is the most specific common ancestor of the
    list of individuals.  This is used in RoleRestriction::close().

    Notice that each i in the list may have more than one parents.  
    This means the resulting concept may also be a conjunct of several
    concepts.

    It is not as difficult as first expected.  The algorithm goes as
    follows:

    1. for the first individual i , take its ancestors and put that in
				 1
       a list, anc.

    2. for each concept, C, in anc, if C subsumes every one of 
       i  ... i , then insert C into result.
	2      n

    3. Return and(result).  result should have at least "Thing" in it.
 */
{
    assert(!ind_list.empty());

    const Set<Concept*>& anc = (*ind_list.begin())->ancestors();
    Set<Concept*> result;

    for (Set<Concept*>::const_iterator C = anc.begin(); C != anc.end(); C++) {
	bool subsume_all = TRUE;
	for (List<Individual*>::const_iterator i = ind_list.begin(); 
	    i++, i != ind_list.end(); )
	    if (!(**C > **i)) {
		subsume_all = FALSE;
		break;
	    }
	if (subsume_all) {
	/*
	    Now *C subsumes all individuals, prepare to add *C to result
	    but first check if any of *C's descendants or ancestors is
	    already there.  If a descendant is there, ignore *C.  If 
	    ancestor(s) are there, delete them.
	 */
	    bool forget = FALSE;
	    Set<Concept*>::iterator j;
	    for (j = result.begin(); j != result.end(); j++) {
		if (**C > **j) {
		    forget = TRUE;
		    break;
		}
	    }
            if (!forget) {
		for (j = result.begin(); j != result.end(); j++)
		    if (**j > **C)
			result.erase(j);
		result.insert(*C);
	    }
	}
    }

    /*
	Due to a proliferation of objects, we should do a simple optimization
	here.  and() is called only if result has more than one base concept.
     */

    Concept* A;
    if (result.size() == 1) {
	A = *result.begin();
    } else { 
	List<Concept*> tmp;
	copy(result.begin(), result.end(), back_inserter(tmp));
        A = and(tmp);
    }

    return A;
}


void RoleRestriction::close()
/*
    A simple operation that closes the role restriction.  It simply
    alter the number restrictions, so that they have the exact value
    of the no. of fillers.  This operation is used only once in 
    instance_of_aux.

    A more correct implementation should also find the most specific
    concept that subsumes all of the fillers.  A lazy way (but more
    efficient) is to simply tell the classifier to be aware of this
    situation, so that (all options string) will subsume (fills
    options "-r") if the latter is closed.  Notice if it is not
    closed, the former does not subsume the latter, because another
    filler, say (fills options 1), can be added any time.

    The correct implementation turns out to be not as difficult.  
    See common_ancestor() above.

    Now we have one-of concepts, we should also make a one-of concept
    for the _vr.  e.g.,

	(fills name "a" "b")

    should infer the vr for name as 
       
	(and string         	; this has been done before 
	     (one-of "a" "b"))
 */
{
    _nr.first = _nr.second = _fillers.size();
    if (!_fillers.empty()) {
	Concept* C = one_of(_fillers);
	if (*C >= *_vr)
	   ; // no need to do anything
	else 
	if (*_vr >= *C)
	    _vr = C;
	else {
	/*
	    I think we should never be here... but not sure.  This will
	    certainly work if it ever gets here, though.
	 */
	    Concept* D = and(_vr, C, 0);
	    _vr = D;
	}
    }
    _closed = TRUE;
}


Role* define_role(const String& name)
{
    if (OB.exists(name)) {
	cout << "Symbol \"" << name << "\" exists already." << endl;
	return 0;
    }

    return new Role(name);
}


Role* define_attribute(const String& name)
{
    if (OB.exists(name)) {
	cout << "Symbol \"" << name << "\" exists already." << endl;
	return 0;
    }

    return new Role(name, TRUE);
}


void Role::print(ostream& cout) const 
{
    Object::print(cout);
    cout << "Name: " << _name << endl;
    cout << "attribute? " << (_attribute? "yes": "no") << endl;
}


ostream& operator<<(ostream& cout, const RoleRestriction& RR)
{
    cout << RR.name() << ": ";
    cout << '[';
    if (RR.low() >= 0)
	cout << RR.low();
    else
	cout << "?";
    cout << ".." ;
    if (RR.high() >= 0)
	cout << RR.high();
    else
	cout << "?";
    cout << ']';

    if (RR.closed())
	cout << "*";

    if (RR.vr() && RR.vr() != Thing) 
	cout << " " << *RR.vr();

    if (!RR._fillers.empty()) {
	cout << " fills(";
	for (List<Individual*>::const_iterator i = RR._fillers.begin();
	    i != RR._fillers.end(); ) {
	/*
	    Print each individual in the filler list.  Notice that, each 
	    filler can be a host individual or a classic individual.  We
	    cannot use 

	    cout << **i;

	    because friend operator<<(ostream&, const Individual&) is not
	    defined.  And I doubt it can be defined because it has to be
	    virtual to work, but it is only a friend function, not a 
	    member function, therefore, cannot be defined as virtual.

	    Pretty_print(), a virtual function, is used.
	 */
	    (*i)->pretty_print(cout);
	    if (++i != RR._fillers.end())
		cout << " ";
	}
	cout << ")";
    }

    return cout;
}


void RoleRestriction::add_filler(Individual* I)
/*
    Insert the individual I into the fillers list.  This is not too
    certain... should we insert duplicates?  Tentatively not.

    The Classic definition of fillers is a set of individuals.  This
    is undesirable in our system.  e.g., the 'arguments' role of a
    UNIX command is an ordered list of tokens.  (notice that even
    though we do not allow duplicates, that does not mean we cannot
    represent 'cat a a'.  This is because each 'a' is represented in
    a unique individual, say Identifier0011 and Identifier0012.)
 */
{
    /*
	'I' already in the filler list?  Cannot use 'find' because we
	must not only do pointer comparisons of individuals.  e.g.,
	
	       fills(options, ind("-r"), 0);
	and    fills(options, ind("-r"), 0);

	should be the same, but they are not the same if pointer
	comparisons are used.
     */
    for (List<Individual*>::iterator i = _fillers.begin();
        i != _fillers.end(); i++)
	if (**i == *I)
	    return;

    _fillers.push_back(I);
}


bool operator==(const RoleRestriction& Ferrari, const RoleRestriction& Porche)
/*
    Test if two RRs are equivalent.  An RR consists of 

      . a role

      . a number restriction

      . a value restriction

      . a list of fillers (a set will be much easier, but then we need 
	the ordering in some case, so we really need to figure out a way
	to compare two lists of fillers.  Currently treat them as sets.
	It is a better way if we define two different kinds of fillers:
	unordered_fillers and ordered_fillers) ###
 */
{
    if (&Ferrari == &Porche)
	return TRUE;

    Set<Individual*> city;
    Set<Individual*> country;

    copy(Ferrari.fillers().begin(), Ferrari.fillers().end(),
	inserter(city));
    copy(Porche.fillers().begin(), Porche.fillers().end(),
	inserter(country));

    return Ferrari.role() == Porche.role() 
	       &&
	   Ferrari.low() == Porche.low() 
	       &&
	   Ferrari.high() == Porche.high() 
	       &&
		   /*
		       For value restrictions, either pointer equality or
		       actual Concept equivalence. (be careful, don't de-
		       reference null concepts... probably never happens
		       because a RR is init with Thing as its vr)
		    */
	   ((Ferrari.vr() == Porche.vr()) || 
            (Ferrari.vr() && Porche.vr() && *Ferrari.vr() == *Porche.vr())) 
	       &&
	   Ferrari.fillers().size() == Porche.fillers().size() 
	       &&
	   city == country;
}


bool subsumes_nr(pair<int, int> n1, pair<int, int> n2)
/*
    Given number restrictions n1, n2, decide whether n1 <= n2.
 */
{
    int low1 = n1.first;
    int high1 = n1.second;

    int low2 = n2.first;
    int high2 = n2.second;

    /*
	It is very problematic if we keep using -1 to indicate
	untold upper bounds because -1 is a small number.  This
	causes a lot of exceptions and is error-prone.  Thus, we
	adjust the highs to INT_MAX thusly:
     */

    if (high1 == -1)
	high1 = INT_MAX;

    if (high2 == -1)
	high2 = INT_MAX;

/*
if (trace.on(Trace::classification))
cout << "DEBUG here (" << low1 << "," << high1 << ")  (" <<
    low2 << "," << high2 << ")\n";
*/

    /*
	With the upper bounds stratified, we can now compare them to
	see if n1 is within range of n2.
     */

    if (low1 >= low2 && high1 <= high2)
	return TRUE;
    else {
	if (trace.on(Trace::classification))
	    cout << "subsumes_nr returns false.\n";
	return FALSE;
    }
}


bool subsumes_vr(const RoleRestriction& RR1, const RoleRestriction& RR2)
{
    const Concept* D = RR1.vr();
    const Concept* C = RR2.vr();
    assert(C);
    assert(D);

    /*
	We must be very careful when the input RR1 is closed.  A closed
	RR means the input is coming from an Individual.  Now consider
	this individual:

	    options: [0..0]* string 

	If we compare the vr of the above to the vr of

	    options: [?..?] string one-of("a" "b")

        then, we are comparing 

	    string <= string one-of("a" "b")

	The result is FALSE while the correct result is TRUE because the
	role options is closed and without a filler.  
     */

    /*
	So, here is the solution.  closed() and no fillers is considered
	subsumed by anything.
     */

    if (RR1.closed() && RR1.number() == 0)
	return TRUE;

    if (*D <= *C)
	return TRUE;
    else {
	if (trace.on(Trace::classification))
	    cout << "subsumes_vr returns false.\n";
	return FALSE;
    }
}


bool subsumes_fillers(const Set<Individual*>& F1, const Set<Individual*>& F2)
{
    if (set_subset(F2, F1, individual_comp))
	return TRUE;
    else {
	if (trace.on(Trace::classification))
	    cout << "subsumes_fillers returns false.\n";
	return FALSE;
    }
}


bool operator<=(const RoleRestriction& HongKong, const RoleRestriction& China)
/*
    Compare to see if the RR 'HongKong' is subsumed by the RR
    'China'.  That is if the number restriction has a smaller or
    equal range, and its value restriction is subsumed by China's and
    its fillers is a subset of China's.
 */
{
    if (HongKong.role() != China.role())
	return FALSE;

    if (&HongKong == &China)
	return TRUE;

    Set<Individual*> city;
    Set<Individual*> country;

    copy(HongKong.fillers().begin(), HongKong.fillers().end(),
	inserter(city));
    copy(China.fillers().begin(), China.fillers().end(),
	inserter(country));

    return subsumes_nr(HongKong.nr(), China.nr()) 
	       && 
	   subsumes_vr(HongKong, China)
	       &&
	   subsumes_fillers(city, country);
#if 0
    return
    /*
	bug alert.  this returns true for

	    HongKong [1..?] <= China [0..0]

	while it should be false.
     */
	HongKong.low() >= China.low()
	    &&
       (HongKong.high() <= China.high() || China.high() == -1)
	    &&
        (!China.vr() || 
	 (HongKong.vr() && *HongKong.vr() <= *China.vr()))
	    &&
    /*
	This is a subtle bug here.  It turns out the following is
	the reverse of what should be.

	    city <= country;

	The reason is the *fewer* fillers subsumes the more.  So,
	for country to subsume city, country has to be a subset.
     */
	country <= city
#endif
#if 0
    /*
	It is interesting to note that even if R2's vr subsumes R1's,
	it does not guarantee R1's vr subsumes all of R2's fillers.

	e.g.,  R1 = (and (all name string) (all name (one-of "a" "b" "c")))
	       R2 = (and (all name string) (fills name "d"))

	In this case, R1's vr subsumes R2's vr (both are 'string'), but 
	R1 does not subsume R2 because of filler problem.  The converse is
	also a problem:

	e.g.,  R1 = (and (all name string) (all name (one-of "a" "b" "c")))
	       R2 = (and (all name string) (fills name "a"))
     */
#endif
}


bool operator<(const RoleRestriction& x, const RoleRestriction& y)
{
    return (x <= y) && (x != y);
}


bool operator>(const RoleRestriction& x, const RoleRestriction& y)
{
    return y < x;
}


bool operator>=(const RoleRestriction& x, const RoleRestriction& y)
{
    return y <= x;
}


bool RoleRestriction::contradict() const
/*
    The algorithm to check for contradiction in role restrictions call
    for a few other helper functions:

      . contradict_nr (number restriction)
      . contradict_vr (value restriction)
      . contradict_fillers
*/
{
#if 1
if (trace.on(Trace::classification)) {
cout << "DEBUG here " << *this << endl;
if (contradict_nr()) cout << "contradict_nr\n"; else cout << "nr ok\n";
if (contradict_vr()) cout << "contradict_vr\n"; else cout << "vr ok\n";
if (contradict_fillers())
    cout << "contradict_fillers\n"; 
else 
    cout << "fillers ok\n";
}
#endif

    return contradict_nr() || contradict_vr() || contradict_fillers();
}


bool RoleRestriction::contradict_nr() const
/*
    Number Restriction [min..max] is contradicting if 

      . min > max != -1     or
      . is an attribute and (min/max != 0 or 1)
 */
{
    if (low() > high() && high() != -1)
	return TRUE;
		 
    if (role()->attribute())
	return low() > 1 || high() > 1; 

    return FALSE;
}


bool RoleRestriction::contradict_vr() const
/*
     Value Restriction is contradicting if

       . it is a null concept.
 */
{
    assert(vr());
    return vr()->null();
}


bool RoleRestriction::contradict_fillers() const
/*
    Fillers are contradicting if 

      . no. of fillers > max and max != -1     or
      . some filler f is not an instance of the vr 
 */
{
    if (_fillers.size() > high() && high() != -1)
	return TRUE;

    for (List<Individual*>::const_iterator i = _fillers.begin();
	i != _fillers.end(); i++) {
	if (!(**i < *vr()))
	    return TRUE;
    }

    return FALSE;
}


Role* role(const String& name)
{
    Object* obj = object(name);
    if (obj && obj->type() == Object::role) 
	return (Role*)obj;
    else
	return 0;
}

bool role_subsumes(const RoleRestriction& HongKong, const RoleRestriction& China)
/*
    Compare to see if the RR 'HongKong' is subsumed by the RR
    'China'.  That is if the number restriction has a smaller or
    equal range, and its value restriction is subsumed by China's and
    its fillers is a subset of China's.
 */
{
    if (HongKong.role() != China.role()) {
cout << "here the roles are different.  Returning FALSE = " << FALSE << endl;
	return FALSE;
    }

    if (&HongKong == &China)
	return TRUE;

    Set<Individual*> city;
    Set<Individual*> country;

    copy(HongKong.fillers().begin(), HongKong.fillers().end(),
	inserter(city));
    copy(China.fillers().begin(), China.fillers().end(),
	inserter(country));

    return subsumes_nr(HongKong.nr(), China.nr()) 
	       && 
	   subsumes_vr(HongKong, China)
	       &&
	   subsumes_fillers(city, country);
#if 0
    return
    /*
	bug alert.  this returns true for

	    HongKong [1..?] <= China [0..0]

	while it should be false.
     */
	HongKong.low() >= China.low()
	    &&
       (HongKong.high() <= China.high() || China.high() == -1)
	    &&
        (!China.vr() || 
	 (HongKong.vr() && *HongKong.vr() <= *China.vr()))
	    &&
    /*
	This is a subtle bug here.  It turns out the following is
	the reverse of what should be.

	    city <= country;

	The reason is the *fewer* fillers subsumes the more.  So,
	for country to subsume city, country has to be a subset.
     */
	country <= city
#endif
#if 0
    /*
	It is interesting to note that even if R2's vr subsumes R1's,
	it does not guarantee R1's vr subsumes all of R2's fillers.

	e.g.,  R1 = (and (all name string) (all name (one-of "a" "b" "c")))
	       R2 = (and (all name string) (fills name "d"))

	In this case, R1's vr subsumes R2's vr (both are 'string'), but 
	R1 does not subsume R2 because of filler problem.  The converse is
	also a problem:

	e.g.,  R1 = (and (all name string) (all name (one-of "a" "b" "c")))
	       R2 = (and (all name string) (fills name "a"))
     */
#endif
}

