#include "trace.h"
#include "individual.h"
#include "rule_base.h"


void RuleBase::insert(Rule* R)
{
    _rules.push_back(R);
    _antecedent_map.insert(pair<Concept* const, Rule*>(R->antecedent(), R));
    _consequent_map.insert(pair<Object* const, Rule*>(R->consequent(), R));
}


List<Rule*> RuleBase::forward_rules(Concept* C) const
{
    List<Rule*> result;

    MultiMap<Concept*, Rule*>::const_iterator low = 
	_antecedent_map.lower_bound(C);
    MultiMap<Concept*, Rule*>::const_iterator high =
	_antecedent_map.upper_bound(C);
    for ( ; low != high; low++)
	result.push_back((*low).second);

    return result;
}


List<Rule*> RuleBase::indirect_rules(Concept* C) const
{
    List<Rule*> result;
    const Set<Concept*>& anc = C->ancestors();

    for (Set<Concept*>::iterator i = anc.begin(); i != anc.end(); i++) {
	MultiMap<Concept*, Rule*>::const_iterator low = 
	    _antecedent_map.lower_bound(*i);
	MultiMap<Concept*, Rule*>::const_iterator high = 
	    _antecedent_map.upper_bound(*i);
	for ( ; low != high; low++)
	    result.push_back((*low).second);
    }

    return result;
}


List<Rule*> RuleBase::backward_rules(Object* C) const
{
    List<Rule*> result;

    MultiMap<Object*, Rule*>::const_iterator low =
	_consequent_map.lower_bound(C);
    MultiMap<Object*, Rule*>::const_iterator high =
	_consequent_map.upper_bound(C);
    for ( ; low != high; low++)
	result.push_back((*low).second);

    return result;
}


void RuleBase::fire(Concept* C, Individual* I)
/*
    Fire all the rules that have C or C's ancestors as antecedents.
    Classify I under their consequents.
 */
{
    Set<Concept*> anc = C->ancestors();
    anc.insert(C);

    List<Rule*> enable;

    /*
	For each ancestor of C (including C), find the rules that
	have it as antecedents.  Gather them into the "enable" set.
     */

    for (Set<Concept*>::iterator i = anc.begin(); i != anc.end(); i++) {
	MultiMap<Concept*, Rule*>::iterator low = 
	    _antecedent_map.lower_bound(*i);
	MultiMap<Concept*, Rule*>::iterator high = 
	    _antecedent_map.upper_bound(*i);
	for ( ; low != high; low++)
	    enable.push_back((*low).second);
    }

    /*
	Fire each of the enabled rules.
     */

    List<Rule*>::iterator r;

    for (r = enable.begin(); r != enable.end(); r++) {
	if ((*r)->enabled() == 50) {
	    cout << "Possible circular rule: " << **r << endl;
	} else {
	    (*r)->enable();
	    (*r)->fire(I);
	}
    }

    /*
	Disable each rule.
     */

    for (r = enable.begin(); r != enable.end(); r++) 
	(*r)->disable();
}


void Rule::init()
/*
    When a new rule is being defined, we need to fire it against
    every instance of the antecdent concept.
 */
{
    const Set<Individual*>& ins = _antecedent->instances();
    for (Set<Individual*>::iterator i = ins.begin(); i != ins.end(); i++) {
	assert((*i)->classic());
	/* 
	    The following fire(*i) is a virtual function, so the actual 
	    fire() under either Rule, ConceptRule or FillersRule will
	    be executed.
	 */
	fire(*i); 
    }
}


bool SimpleRule::fire(Individual* I)
/*
    When a simple rule is fired, the target individual is being
    classified under the consequent concept.
 */
{
    assert(I->classic());

    if (trace.on(Trace::rule))
	cout << "Firing Rule " << *this << endl;

    if (((ClassicIndividual*)I)->classify(consequent())) {
	I->add_rule(this);
//      cout << "debug 1 added " << *this << " to " << *I << endl;
	return TRUE;
    } else
	return FALSE;
}


bool ConceptRule::fire(Individual* I)
/*
    When a concept rule is fired, the target individual is being
    classified under the concept returned by an attached procedure.
 */
{
    assert(I->classic());

    Concept* conseq = (*consequent())(I);
    assert(conseq);
    if (((ClassicIndividual*)I)->classify(conseq)) {
	I->add_rule(this);
	return TRUE;
    } else
	return FALSE;
}


bool FillersRule::fire(Individual* I)
/*
    When a fillers rule is fired, the target individual will be
    modified by adding the values returned by the computed-fillers
    as fillers to the designated role.
 */
{
    assert(I->classic());

    ObjectVector fillers = (*consequent())(I);
    Role* R = consequent()->role();

    if (!fillers.empty()) {
	List<Individual*> F;
	for (int i = 0; i < fillers.size(); i++) 
	    F.push_back((Individual*)fillers[i]);
#if 0
	if (((ClassicIndividual*)I)->classify(fills(R, F))) {
	    I->add_rule(this);
	    return TRUE;
	} else
	    return FALSE;
#else
        if (((ClassicIndividual*)I)->add_fillers(R, F)) {
	    I->add_rule(this);
	    return TRUE;
	} else
	    return FALSE;
#endif
    }

    return TRUE;
}


void Rule::print(ostream& out) const
{
    Object::print(out);
    out << "Name: " << _name << endl;
    out << "Antecedent: " << *_antecedent << endl;
    out << "Consequent: " << *_consequent << endl;
}


void Rule::pretty_print(ostream& out) const 
{ 
    out << _name << ": " << *_antecedent << " --> " << *_consequent;
}

