#if !defined(RULE_H)
#define RULE_H

#include "rule_base.h"
#include "procedure.h"

class Rule: public Object {

/*
    Since there are three forms of rules in Classic:
     
	Rule:        Antecedent --> Concept
	ConceptRule: Antecedent --> ComputedConcept
	FillersRule: Antecedent --> ComputedFillers

    this class is made to store and process the operations common
    to all three.  Basically, the common data is the antecedent,
    i.e., half a rule.
 */

    enum SubType { simple_rule, concept_rule, fillers_rule };

protected:

    Rule(const String& name, Concept* ant, Concept* con): Object(name),
	_name(name), _antecedent(ant), _consequent(con), _enabled(0),
	_subtype(simple_rule)
    {
	RB.insert(this);
    }

    Rule(const String& name, Concept* ant, ComputedConcept* cc): Object(name),
	_name(name), _antecedent(ant), _consequent(cc), _enabled(0),
	_subtype(concept_rule)
    {
	RB.insert(this);
    }

    Rule(const String& name, Concept* ant, ComputedFillers* cf): Object(name),
	_name(name), _antecedent(ant), _consequent(cf), _enabled(0),
	_subtype(fillers_rule)
    {
	RB.insert(this);
    }

    void init();

public:

    virtual ~Rule() {}

    Type type() const { return rule; }

    String type_name() const { return "rule"; }

    String name() const { return _name; }

    virtual SubType subtype() const { return _subtype; }

    Concept* antecedent() const { return _antecedent; }

    /*
	This consequent() function is not to be used by users since 
	it returns the consequent as pure Object*.  We cannot
	define it as pure virtual either, because the consequent()
	functions of all its derived classes have have different 
	returned values (see explanation in the above doc).
     */

    Object* consequent() const { return _consequent; }

    void print(ostream& out = cout) const;

    void pretty_print(ostream& out = cout) const;

    friend ostream& operator<<(ostream& out, const Rule& R)
    {
	R.pretty_print(out);
	return out;
    }

    int enabled() const { return _enabled; }

    bool disabled() const { return !_enabled; }

    void enable() { _enabled++; }    

    void disable() { if (_enabled) _enabled--; }

    virtual bool fire(Individual*) = 0;

private:

    Concept* _antecedent;

    Object* _consequent;

    String _name;

    int _enabled;

    SubType _subtype;
};


class SimpleRule: public Rule {

    friend SimpleRule* define_rule(const String& name, Concept* a, Concept* c) 
    {
	return new SimpleRule(name, a, c);
    }

protected:

    SimpleRule(const String& name, Concept* ant, Concept* con): 
	Rule(name, ant, con)
    {
	/*
	    Each derived class of Rule has to call Rule::init() in its
	    constructor.  init() cannot be put in the ctor of Rule 
	    since it calls a virtual function (fire).
	 */
	init();
    }

public:

    virtual ~SimpleRule() {}

    Concept* consequent() const
    {
	return (Concept*)Rule::consequent(); 
    }

    bool fire(Individual*);
};


class ConceptRule: public Rule {

    friend ConceptRule* define_rule(const String& name, Concept* ant, 
	ComputedConcept* cc)
    {
	return new ConceptRule(name, ant, cc);
    }

protected:

    ConceptRule(const String& name, Concept* ant, ComputedConcept* cc): 
	Rule(name, ant, cc)
    {
	/*
	    Each derived class of Rule has to call Rule::init() in its
	    constructor.  init() cannot be put in the ctor of Rule 
	    since it calls a virtual function (fire).
	 */
	init();
    }

public:

    virtual ~ConceptRule() {}

    ComputedConcept* consequent() const
    {
	return (ComputedConcept*)Rule::consequent(); 
    }

    bool fire(Individual*);
};


class FillersRule: public Rule {

    friend FillersRule* define_rule(const String& name, Concept* ant, 
	ComputedFillers* cf)
    {
	return new FillersRule(name, ant, cf);
    }

protected:

    FillersRule(const String& name, Concept* ant, ComputedFillers* cf): 
	Rule(name, ant, cf)
    {
	/*
	    Each derived class of Rule has to call Rule::init() in its
	    constructor.  init() cannot be put in the ctor of Rule 
	    since it calls a virtual function (fire).
	 */
	init();
    }

public:

    virtual ~FillersRule() {}

    ComputedFillers* consequent() const
    {
	return (ComputedFillers*)Rule::consequent(); 
    }

    bool fire(Individual*);
};

#endif
