#if !defined(STL_H)
#define STL_H

#include <algo.h>
#include <list.h>
#include <vector.h>
#include <map.h>
#include <multimap.h>
#include <set.h>
#include <assert.h>

/*
    The map structure is used for the following purposes:

    - maintaining names of objects (i.e., String to Object map)

    Since the interface to the STL map class is not too standard yet
    (e.g., many compilers do not allow the use of default template
     arguments, therefore, the Map class is used to wrap it up)
 */
template <class Key, class Value>
class Map: public map<Key, Value, less<Key> > {
/*
    I want my map to have simply two arguments: key and value.  I also
    want the [] operator to signal whenever the input parameter is 
    invalid.  The parent version (g++ stl) simply returns the first
    value of the map, which is undesirable.
 */
public:
    /*
        This [] operator has same semantics as the parent except
        it asserts to make sure the input key is valid.
     */
    Value& operator[](const Key& s)
    {
        assert(find(s) != end());

        /*
            Simply call the parent's [] operator.  It may be more 
            efficient to use the result of the above find()... but
            I am more confident with calling the parent's [].
         */
        return map<Key, Value, less<Key> >::operator[](s);
    }
};


template <class Key, class Value>
class MultiMap: public multimap<Key, Value, less<Key> > {
};


template <class Type>
class Vector: public vector<Type> {

public:

    int operator==(const Vector<Type>& x) const
    {
	if (size() != x.size())
	    return 0;
	for (int i = 0; i < size(); i++)
	    if ((*this)[i] != x[i])
		return 0;
	return 1;
    }

#if 0
    Vector<Type>& operator=(const List<Type>& L)
    /*
	Because of computed fillers, it is very useful to have a function
	that converts a list of inds into an ObjectVector.
     */
    {
	int j = 0;
	for (List<Type>::const_iterator i = L.begin(); i != L.end(); i++)
	    (*this)[j++] = *i;
	return *this;
    }
#endif
};


template <class Type>
class Set: public set<Type, less<Type> > {

public:

    /*
	Subset/proper subset relationships.  Superset relationships are 
	still defined by standard stl operators (although I made a 
	slight change in >=.
     */

    bool operator<=(const Set<Type>& s) const
    {
	return includes(s.begin(), s.end(), begin(), end());
    }

    bool operator<(const Set<Type>& s) const
    {
	return size() < s.size() && *this <= s;
    }

    bool operator==(const Set<Type>& s) const
    {
	return size() == s.size() && equal(begin(), end(), s.begin());
    }

    Set<Type> operator+(const Set<Type>& s) const
    {
	Set<Type> result = *this;

        for (Set<Type>::iterator i = s.begin() ; i != s.end(); i++) 
	    result.insert(*i);

	return result;
    }
};


template <class T, class Compare>
inline int set_in(const T& x, const Set<T>& s, Compare comp)
{
    for (Set<T>::const_iterator i = s.begin(); i != s.end(); i++)
	if (!comp(*i, x))
	    return 1;

    return 0;
}


template <class T, class Compare>
inline int set_subset(const Set<T>& s1, const Set<T>& s2, Compare comp)
{
    for (Set<T>::const_iterator i = s1.begin(); i != s1.end(); i++)
	if (!set_in(*i, s2, comp)) 
	    return 0;

    return 1;
}


template <class T, class Compare>
inline int set_equal(const Set<T>& s1, const Set<T>& s2, Compare comp)
{
    if (s1.size() != s2.size())
	return 0;

    return set_subset(s1, s2, comp);
}


template <class Type>
class List: public list<Type> 
{
};


template <class Container>
inline insert_iterator<Container> inserter(Container& x)
{
    return insert_iterator<Container>(x, Container::iterator(x.end()));
}

#endif
