#include "trace.h"
#include <unistd.h>
#include <dirent.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <pwd.h>
#include <limits.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "concept.h"
#include "individual.h"
#include "procedure.h"
#include "expression.h"
#include "classify.h"

/*
    Writing a cf.

    . declare an ObjectVector result
    . check arg size
    . check I->filler(R) exists, if so, return it.
      . be very careful about this.  I->filler() has an assert statement
	checking whether I has R.  Also, even if I has R, it does not
	mean R has any fillers.
 */

String newname(const String& C)
/*
    When a new individual is needed, it is automatically created using
    the convention: CONCEPT00xx, where 00xx is a number.  This routine
    creates a new name for a new individual.  'C' is the concept name.
 */
{
    assert(concept(C));

    char name[100];
    int x = concept(C)->individuals().size() + 1;
    sprintf(name, "%s%04d", (const char*)C, x);
      
    return name;
} 


/*
    An arg element: arg[i] is an object pointer.  It can be casted to
    Role*, Concept* etc.
 */


inline bool valid(const ClassicIndividual* I, const Object* R)
{
    return I->role_restriction((Role*)R) && I->number((Role*)R);
}


TestResult test_email(const ClassicIndividual* I, const ObjectVector& arg)
/*
   Test to see if an identifier looks like an email address.
   To simplify matter, an email address is defined to be string@string
   where each string does not have any space or : or /.
 */
{ 
    if (!valid(I, arg[0]))
	return MAYBE;

    String name = *I->filler((Role*)arg[0]);

    if (!name.contains(Regex(" :/")) &&
        name.firstchar() != '@' &&
        name.lastchar() != '@' &&
        name.freq('@') == 1)
        return YES;
    else
        return NO;
}


TestResult test_username(const ClassicIndividual* I, const ObjectVector& arg)
/*
    Test to see if an identifier is an ordinary user name in the site.
 */
{ 
    if (!valid(I, arg[0]))
	return MAYBE;

    String name = *I->filler((Role*)arg[0]);
    return getpwnam(name)? YES: NO;
}


Concept* cc_filename_type(const ClassicIndividual* I, const ObjectVector& arg)
/*
    This is a simple procedural attachment implementing a computed
    concept.

    An individual has been classified under FILE-TYPE.  Its real
    file-type should be inferred by calling the "file" program.
    But before we do that, we tried to guess its type from its
    pathname first.  ind is guaranteed to have a "name" filler.
    It is then partitioned to "base" and "suffix".  Basically,
    "suffix" is used to infer the filename_type.
 */
{
    assert(I->filler("name"));

    if (!I->role_restriction("base") || !I->role_restriction("suffix"))
	return concept("FILENAME");

    String name = *I->filler("name");
    String base = *I->filler("base");
    String suffix = *I->filler("suffix");

    if (name.lastchar() == '/')
        return concept("DIRECTORY");

    else if (suffix == ".txt" || downcase(base) == "readme")
        return concept("ENGLISH");
    
    else if (suffix == ".c")
        return concept("C");

    else if (suffix == ".C" || suffix == ".c++" ||
             suffix == ".cc" || suffix == ".cxx")
        return concept("C++");

    else if (suffix == ".h")
        return concept("HEADER");

    else if (suffix == ".p")
        return concept("PASCAL");

    else if (suffix == ".f")
        return concept("FORTRAN");

    else if (suffix == ".s")
        return concept("ASSEMBLER");

    else if (suffix == ".lisp" || suffix == ".lsp")
        return concept("LISP");

    else if (suffix == ".ps")
        return concept("POSTSCRIPT");

    else if (suffix == ".ms")
        return concept("MS");

    else if (suffix == ".me")
        return concept("ME");

    else if (suffix == ".tex")
        return concept("TEX");

    else if (suffix == ".grn")
        return concept("GRN");

    else if (suffix.length() == 2 && isdigit(suffix[1]))
        return concept("MAN");

    else if (name == "mbox")
        return concept("MAILBOX");

    else if (name == "a.out")
        return concept("EXEC");

    else if (suffix == ".o")
        return concept("OBJECT");

    else if (suffix == ".Z")
        return concept("COMPRESSED");

    else if (suffix == ".gz")
        return concept("GZIP");

    else if (suffix == ".tar")
        return concept("TARRED");

    else if (suffix == ".uu")
        return concept("UUENCODED");

    else if (suffix == ".cl")
        return concept("CLASSIC");

    else if (suffix == ".a")
        return concept("LIBRARY");

    else if (suffix == ".csh" || name == ".login" ||
            name == ".logout" || name == ".cshrc")
        return concept("C-SHELL");

    else if (suffix == ".sh" || name == ".profile")
        return concept("BOURNE-SHELL");

    else if (name == ".ksh")
        return concept("KORN-SHELL");

    else if (name == "core")
        return concept("CORE");

    else
        return ClassicThing;
}


ObjectVector cf_filename_suffix(const ClassicIndividual* I, const Role* R,
    const ObjectVector& arg)
{
    assert(arg.size() == 1);
    ObjectVector result;

    if (valid(I, R))
        result.push_back(I->filler(R));
    else {
	assert(valid(I, arg[0]));
	String name = *I->filler((Role*)arg[0]);
	name.reverse();
	String suffix = name.before('.');
	suffix += '.';
	suffix.reverse();
	result.push_back(ind(suffix));
    }

    return result;
}


ObjectVector cf_filename_base(const ClassicIndividual* I, const Role* R,
    const ObjectVector& arg)
/*
   Compute the base name from a given filename.  "name" is guaranteed
   to have a filler, otherwise, it won't pass the filename? test.
 */
{
    assert(arg.size() == 1);
    ObjectVector result;

    if (valid(I, R))
        result.push_back(I->filler(R));
    else {
	assert(valid(I, arg[0]));
	String name = *I->filler((Role*)arg[0]);
	if (name.contains('.'))
	    result.push_back(ind(name.before('.')));
	else
	    result.push_back(ind(name));
    }

    return result;
}


String real_path(const String& name)
{
    String resolved;

    resolved = name;

    return resolved;
}


ObjectVector cf_fname_realpath(const ClassicIndividual* I, const Role* R,
    const ObjectVector& arg)
/*
    Compute the canonical form of any pathname.  Sun OS has a realpath
    system function.  
 */
{
    ObjectVector result;

    if (valid(I, R))
	result.push_back(I->filler(R));
    else {
	assert(valid(I, arg[0]));
        String name = *I->filler((Role*)arg[0]);
        result.push_back(ind(real_path(name)));
    }

    return result;
}


#include <astro.h>

TestResult test_legal_date(const ClassicIndividual* I, const ObjectVector& arg)
{
    assert(valid(I, arg[0]));
    assert(valid(I, arg[1]));
    assert(valid(I, arg[2]));

    int year = *I->filler((Role*)arg[0]);
    int month = *I->filler((Role*)arg[1]);
    int day = *I->filler((Role*)arg[2]);

    if (year == -1 || month == -1 || day == -1)
        return MAYBE;

    return legaldate(year, month, day)? YES: NO;
}


ObjectVector cf_dow(const ClassicIndividual* I, const Role* R,
    const ObjectVector& arg)
{
    assert(valid(I, arg[0]));
    assert(valid(I, arg[1]));
    assert(valid(I, arg[2]));

    int year = *I->filler((Role*)arg[0]);
    int month = *I->filler((Role*)arg[1]);
    int day = *I->filler((Role*)arg[2]);
    int dow = epoch(year, month, day) % 7;

    ObjectVector result;
    result.push_back(ind(dow));

    return result;
}


ObjectVector cf_getcwd(const ClassicIndividual* I, const Role* R,
    const ObjectVector& arg)
{
    char dir[MAXPATHLEN];
    String s = getcwd(dir, MAXPATHLEN)? dir: "";

    ObjectVector result;
    result.push_back(ind(s));

    return result;
}


TestResult test_filename(ClassicIndividual* I, const ObjectVector& arg)
/*
   For simplicity, a filename is defined as a string that is less
   than 60 characters in length and has no / in it. 
 */
{
    if (!valid(I, arg[0]))
	return MAYBE;

    String name = *I->filler((Role*)arg[0]);
    return (name.length() > 0 && name.length() < 60 && 
            !name.contains('/'))?  YES: NO;
}


TestResult test_relname(ClassicIndividual* I, const ObjectVector& arg)
/*   
   A relname is defined to be a string with length < 128 characters
   and has '/' in the middle (but not at the ends).
 */  
{    
    if (!valid(I, arg[0]))
        return MAYBE;

    String name = *I->filler((Role*)arg[0]);

    TestResult i = (name.length() > 0 && name.length() < 128 && 
	      name.firstchar() != '/' && name.lastchar() != '/' && 
	      name.index('/', 1) != -1)? YES: NO;

    return i;
}


TestResult test_pathname(ClassicIndividual* I, const ObjectVector& arg)
/*
   A pathname is defined to be a string with length < 128 characters  
   and has '/' as the first character.  The theoretical limit is
   MAXPATHLEN, but practically 128 is more realistic.
 */
{ 
    if (!valid(I, arg[0]))
        return MAYBE;

    String name = *I->filler((Role*)arg[0]);
    return (name.length() > 0 && name.length() < 128 &&
            name[0] == '/')? YES: NO;
}


TestResult test_prefix(ClassicIndividual* I, const ObjectVector& arg)
{    
    assert(valid(I, arg[0]));

    String s, p;
     
    s = *I->filler((Role*)arg[0]);
    p = *arg[1];

    return s.matches(p)? YES: NO;
}


TestResult test_exist(ClassicIndividual* I, const ObjectVector& arg)
{
    if (!valid(I, arg[0]))
	return MAYBE;

    String name = *I->filler((Role*)arg[0]);
    struct stat buf;
    return stat(name, &buf) != -1? YES: NO;
}


TestResult test_not_exist(ClassicIndividual* I, const ObjectVector& arg)
{
    if (!valid(I, arg[0]))
	return MAYBE;

    String name = *I->filler((Role*)arg[0]);
    struct stat buf;
    return stat(name, &buf) == -1? YES: NO;
}


ObjectVector cf_filename_file(const ClassicIndividual* I, const Role* R,
    const ObjectVector& arg)
/*   
   If a pathname is found to be EXIST, then we should fetch the
   file associated w/ that pathname.  We also put a back pointer
   into that file to point back to pathname.

   We have to check if that file already exists under FILE, if so
   simply return it.  Otherwise, create a new one.
 */
{ 
    ObjectVector result;
    char expr[BUFSIZ];

/* printf("DEBUG in filename-file for %s\n", cl_ind_name(ind)); */

    assert(valid(I, arg[0]));
    String name = *I->filler((Role*)arg[0]);

    struct stat buf;
    if (stat(name, &buf) == -1)  {
        cout << '"' << name << "\" does not exist (in filename-file)\n";
        return result;
    }

    /*
        Since a file is a unique object, if the same filename (of the
	same file) is passed in this routine more than once, the first
	time, we must assert a new FILE individual and return it.  The
	second and subsequent times, we must return the same individual
        that we created in the first time.
     */

    /*
	See if there are individuals of the form

	    (and FILE (fills st_ino <inode-number>))

	The inode number (form buf.st_ino) will uniquely identify a file.
	A more computationally efficient way is just to look through each
	individual of FILE.  I'll take the latter approach.
     */

    assert(concept("FILE"));
    const Set<Individual*>& files = concept("FILE")->individuals();
    for (Set<Individual*>::iterator i = files.begin(); i != files.end(); i++) {
	ClassicIndividual* I = (ClassicIndividual*)(*i);
	Individual* ino = I->filler("st_ino");
	if (ino && *((IntegerIndividual*)ino) == buf.st_ino) {
            result.push_back(*i);
	    return result;
	}
    }

    String id = newname("FILE");
    sprintf(expr, "(define-individual %s (and FILE \
	   (fills st_dev %d)\
	   (fills st_ino %d)\
	   (fills st_mode %d)\
	   (fills st_nlink %d)\
	   (fills st_uid %d)\
	   (fills st_gid %d)\
	   (fills st_rdev %d)\
	   (fills st_size %d)\
	   (fills st_atime %d)\
	   (fills st_mtime %d)\
	   (fills st_ctime %d)\
	   (fills st_blksize %d)\
	   (fills st_blocks %d)\
	   (at-most %d fname)\
	   (fills fname %s)))",
	   (const char*)id,
	   buf.st_dev,
	   buf.st_ino,
	   buf.st_mode,
	   buf.st_nlink,
	   buf.st_uid,
	   buf.st_gid,
	   buf.st_rdev,
	   buf.st_size,
	   buf.st_atime,
	   buf.st_mtime, 
	   buf.st_ctime, 
	   buf.st_blksize,
	   buf.st_blocks, 
	   buf.st_nlink,
	   (const char*)(I->name())
    );   
    evaluate(expr);
    result.push_back(individual(id));

    return result;
}


ObjectVector cf_file_mode(const ClassicIndividual* I, const Role* R,
    const ObjectVector& arg)
/*
   Return a vector that is the modes a file has.  This is a simplified
   version in which there are only three modes r, w, x instead of rwx
   for owner/group/others.
 */
{
    ObjectVector result;

    if (valid(I, R)) {
	const List<Individual*>& tmp = I->fillers(R);
	/*
	    No simple way to assign a List<Individual*> to Vector<Object*>.
	    It is easy if rhs was List<Object*>.  

	    result = tmp;
	*/
	copy(tmp.begin(), tmp.end(), back_inserter(result));
	return result;
    }

    if (!valid(I, arg[0])) 
	return result;

    int st_mode = *I->filler((Role*)arg[0]);

    if ((st_mode & S_IRUSR) || (st_mode & S_IROTH))
        result.push_back(ind("r"));

    if ((st_mode & S_IWUSR) || (st_mode & S_IWOTH))
        result.push_back(ind("w"));

    if ((st_mode & S_IXUSR) || (st_mode & S_IXOTH))
        result.push_back(ind("x"));
    
    return result;
}


Concept* cc_file_type(const ClassicIndividual* I, const ObjectVector& arg)
/*
   Given a file, try to find its file type by calling "file(1)".
   Not ready for symbolic links yet.
 */
{
    if (!valid(I, arg[0]))
	return ClassicThing;

    const List<Individual*>& pathnames = I->fillers((Role*)arg[0]);
    if (pathnames.empty())
        return ClassicThing;

    ClassicIndividual* path = (ClassicIndividual*)*pathnames.begin();

    if (!valid(path, role("realpath")))
	return ClassicThing;
    String name = *path->filler(role("realpath"));

    char expr[100];
    sprintf(expr, "myfile %s", (const char*)name);
    FILE* in = popen(expr, "r");
    if (!in)
        cerr << "Error: popen " << expr << " in file-type\n";

    fgets(expr, BUFSIZ, in);
    if (strchr(expr, '\n')) 
	*strchr(expr, '\n') = 0;
    pclose(in);

    if (strstr(expr, "No such file"))
        return ClassicThing;
    else
        return concept(expr);
}


ObjectVector cf_find_first_argument(const ClassicIndividual* I, const Role* R,
    const ObjectVector& arg)
{
    assert(valid(I, arg[0]));
    const List<Individual*>& arguments = I->fillers((Role*)arg[0]);

    ObjectVector result;
    result.push_back(arguments.front());

    return result;
}


ObjectVector cf_find_last_argument(const ClassicIndividual* I, const Role* R,
    const ObjectVector& arg)
{
    assert(valid(I, arg[0]));
    const List<Individual*>& arguments = I->fillers((Role*)arg[0]);

    ObjectVector result;
    result.push_back(arguments.back());

    return result;
}


TestResult test_some(const ClassicIndividual* I, const ObjectVector& arg)
/*
   Syntax: (test-c some role CONCEPT [min [max]])

   Test to see if the fillers of the role of this individual have
   at least min and at most max CONCEPT.  Min defaults to 1 and
   max defaults to max integer.
 */
{
    assert(arg.size() >= 2);
    assert(arg[0]->type() == Object::role);
    assert(arg[1]->type() == Object::concept ||
	   arg[1]->type() == Object::primitive_concept ||
	   arg[1]->type() == Object::disjoint_primitive_concept);

    int min = 0, max = INT_MAX, n = arg.size();

    if (n >= 3)
         min = *arg[2];
    if (n >= 4)
         max = *arg[3];

    Role* R = (Role*)arg[0];
    Concept* C = (Concept*)arg[1];

    if (!valid(I, R))
         return NO;

    int count = 0;
    const List<Individual*>& fillers = I->fillers(R);
    for (List<Individual*>::const_iterator i = fillers.begin();
        i != fillers.end(); i++)  {

	/*
	    cout << "DEBUG here\n";
	    cout << "    " << *C << endl;
	    cout << "    " << **i << endl;  // works ok

	    if (**i < *C)	// doesn't work
		count++;

	    Seems to be a bug of g++.  **i < *C should work.  The same
	    thing has been used in many other places!
            ___________________________________________________________

	    Whoops!  Sorry, forgot to include "classify.h".  My fault.
	    Don't understand why the compiler does not trap it as 
	    compile-time error though.
	 */

	if (**i < *C)	// doesn't work
	    count++;
    }

    if (min <= count && count <= max)
         return YES;
    else
         return NO;
}


ObjectVector cf_get_option(const ClassicIndividual* I, const Role* R,
    const ObjectVector& arg)
/*   
   This is a computed-filler used to get a particular option argument 
   from "optargs" and assert it into the corresponding role. 
     
   Usage: (computed-filler get-option role "-flag")
 
   It is assumed that optargs is filled with the necessary filler.
   An empty vector is returned if there is no such option.
 */
{ 
    assert(arg.size() > 0);

    ObjectVector result;

    if (valid(I, R)) {
	const List<Individual*>& tmp = I->fillers(R);
	copy(tmp.begin(), tmp.end(), back_inserter(result));
    } else {

	String flag = *arg[0];
	const List<Individual*>& optargs = I->fillers("optargs");
   
	for (List<Individual*>::const_iterator i = optargs.begin();
	    i != optargs.end(); i++) {
	    String option = **i;
	    if (option.index(flag) == 0) {
		/*
		    Found! Create a new individual, tentatively under
		    IDENTIFIER, for this option.
		 */
		String argument = option.after(flag);
#if 0
    /*
	This is from the old code, I don't see any reason to create a new
	IDENTIFIER ... just use string is fine.
     */
		Individual* id = define_individual(
				     newname("IDENTIFIER"), 
				     and(concept("IDENTIFIER"),
					 fills(role("name"), ind(argument), 0),
					 0
				     )
				 );
		assert(id);
		result.push_back(id);
#else
                result.push_back(ind(argument));
#endif
		break;
	    }
	}
    }

    return result;
}


ObjectVector cf_get_option_integer(const ClassicIndividual* I, const Role* R,
    const ObjectVector& arg)
/*
    optarg is a filler that stores option pairs.  such as -b 8, -font courier
    etc.  Supposed we have -b and wanted to find 8, we should us this
    procedure passing "-b" as arg[0].  Then 8 will be returned.
 */
{
    assert(arg.size() > 0);

    ObjectVector result;

    if (valid(I, R)) {
	const List<Individual*>& tmp = I->fillers(R);
	copy(tmp.begin(), tmp.end(), back_inserter(result));
    } else {

	String flag = *arg[0];
	const List<Individual*>& optargs = I->fillers("optargs");
   
	for (List<Individual*>::const_iterator i = optargs.begin();
	    i != optargs.end(); i++) {
	    String option = **i;
	    if (option.index(flag) == 0) {
		result.push_back(ind(atoi(option.after(flag).chars())));
		break;
	    }
	}
    }

    return result;
}


void declare()
{
    define_function("filename-suffix", cf_filename_suffix);
    define_function("filename-base", cf_filename_base);
    define_function("fname-realpath", cf_fname_realpath);
    define_function("filename-type", cc_filename_type);
    define_function("legal-date?", test_legal_date);
    define_function("find-dow", cf_dow);
    define_function("filename?", test_filename);
    define_function("relname?", test_relname);
    define_function("pathname?", test_pathname);
    define_function("strprefix", test_prefix);
    define_function("getcwd", cf_getcwd);
    define_function("exist?", test_exist);
    define_function("not-exist?", test_not_exist);
    define_function("filename-file", cf_filename_file);
    define_function("file-mode", cf_file_mode);
    define_function("file-type", cc_file_type);
    define_function("email?", test_email);
    define_function("username?", test_username);
    define_function("find-first-argument", cf_find_first_argument);
    define_function("find-last-argument", cf_find_last_argument);
    define_function("some", test_some);
    define_function("get-option", cf_get_option);
    define_function("get-option-integer", cf_get_option_integer);
}
